

# Extract data from Exxon-311
# The velocity model for isotropic migration is Vnmo
# Vnmo = Vv*sqrt(1+2*delta)

dataloc = /data/exxon-311/


exxon311b.H:
	Headermath < EXXON311.HH Headermath key1=s_x eqn1="(sx-497634.3106)*0.725-0.689*(sy-3183172.481)" type1=scalar_float key2=s_y eqn2="-(sx-497634.3106)*0.689-0.725*(sy-3183172.481)" type2=scalar_float key3=g_x eqn3="(gx-497634.3106)*0.725-0.689*(gy-3183172.481)" type3=scalar_float key4=g_y eqn4="-(gx-497634.3106)*0.689-0.725*(gy-3183172.481)" type4=scalar_float key5=off_x eqn5="(gx-497634.3106)*0.725-0.689*(gy-3183172.481)-(sx-497634.3106)*0.725+0.689*(sy-3183172.481)" type5=scalar_float maxsize=100 > $@


# Location 1
Data/dat-extract1.H: exxon311b.H
	Window_key < exxon311b.H key1=s_y maxk1=1670. mink1=1530. key2=g_y maxk2=1670. mink2=1530. > tmp1.H
	Sort3d < tmp1.H nkeys=2 key1=off_x ng1=144 dg1=25. og1=125. key2=s_x ng2=611  dg2=25. og2=1125. > tmp2.H
	Stack3d < tmp2.H normalize=y axes=2 | Infill3d normalize=y maxsize=1000 > Data/dat-extract1.H
	echo "label1=t(s) label2=offx(m) label3=sx(m)" >> $@


Data/dat311-250m.H: Data/dat-extract1.H
	pycode/pyRemove.py tmpdat*.H
	Window3d < Data/dat-extract1.H n3=31 o3=0 j3=10 > tmpdat1.H
	Bandpass < tmpdat1.H fhi=25. flo=3. > tmpdat2.H
	Pad < tmpdat2.H beg1=25 > tmpdat3.H
	Tpow < tmpdat3.H tpow=0.7 panel=a > tmpdat4.H
	Window3d < tmpdat4.H f1=30 | Pad beg1=30 > tmpdat5.H
	Interp < tmpdat5.H d1out=0.002 n1out=2250 > $@
	echo "o1=0.0 label1=t(s) label2=offx(m) label3=sx(m)" >> $@
	pycode/pyRemove.py tmpdat*.H  

# Throw out the bad traces	
Data/dat311-250m-clean.H Data/sp-250m-clean.H: bin/RemoveEmptyShot.x Data/dat311-250m.H
	bin/RemoveEmptyShot.x < Data/dat311-250m.H shotposfile=Data/sp-250m-clean.H >$@ 


################################################################################################
##### Velocity model 
################################################################################################

Data/vel311-1.H:
	Window3d < ${dataloc}/Vel/exxon-velocity.H n2=500 d2=25. o2=1000. > tmpvel1.H
	Window3d < tmpvel1.H n3=1 f3=52 | Interp d1out=25. > $@
	echo "label1=z(m) label2=x(m)" >> $@
	pycode/pyRemove.py tmpvel1.H


################################################################################################
##### Data Weighting
################################################################################################

DatWgt/DWTpow250m.H: DatWgt/DW-250m-clean.H 
	Tpow tpow=1 < DatWgt/DW-250m-clean.H  > $@

DatWgt/DW-250m-clean.H: bin/DataWeight2D.x 
	pycode/pyRemove.py tmp1.H
	bin/DataWeight2D.x < Data/dat311-250m-clean.H > tmp1.H	
	Mute < tmp1.H tmute=0.5 vmute=1400. > $@
	pycode/pyRemove.py tmp1.H

