program Dottest

  use sep
  use prop_mod1D

  use hycdsolver_smp_mod
  use hycdstep_mod
  use identity_mod

  implicit none

  integer                                      ::  niter         ! number of iterations
  real                                         :: qntd, qntm     ! quantile
  real                                         :: thresh_d,thresh_m  ! threshold

  logical(4) :: success
  logical                               :: wfldmovie, VDer, DW, MW
  real                                  :: dx, dz, oz, ost, dst, ot_data, dt_data, dt_prop
  real                                  :: zsrc, zrec,vmax,tdelay,tapeco
  real                                  :: dot1, dot2
  integer                               :: nnz, nz, nst, nt_data, nop, npad
  integer            			:: iz , it, orderT, bout
  real, allocatable, dimension(:)     :: vel, mwarr, dat, DWarray
  real, allocatable, dimension(:)     :: image, resdArr
  real, allocatable, dimension(:)       ::wavelet, mwarr1D
  real, allocatable, dimension(:)       ::x1, x2, y1, y2
  character(len=128)                    :: souWfile_, imgWfile_, recWfile_
  character(len=128)                    :: outfile_

  call sep_init()

  ! velocity model info
  nz = 350
  dz = 12.5
  oz = 0.

 ! shot gather (data) info
  nt_data = 500
  dt_data = 0.001
  ot_data = 0.

  ! source and receiver location
  zsrc = 12.5
  zrec = zsrc

  ! Read various switches and popagation specs
  call from_param("VDer",VDer,.false.)
  call from_param("DW",DW,.false.);   call from_param("MW",MW,.false.)
  call from_param("tdelay",tdelay,0.002)
  call from_param("dt_prop",dt_prop,0.001)
  call from_param("tapeco",tapeco,0.00002);
  call from_param("orderT",orderT,2);
  npad = 30

  nst = 1000.    ! sourcce wavelet nt
  dst = dt_data  ! source wavelet dt
  ost = 0.       ! source wavelet ot
  tdelay = 0.002  ! Delay in source wavelet


  write(0,*) 'nt_data, dt_data, ot_data', nt_data, dt_data, ot_data
  write(0,*) 'dt_prop', dt_prop  


  if (dst .ne. dt_data) call erexit('dst should equal to dt_data')
  if (dt_prop .ne. dt_data) call erexit('expect dt_prop .eq. dt_data for now')

  nnz=nz
  nz = nnz+2*npad

  ! Allocate memory
  allocate(vel(nz),image(nnz),wavelet(nt_data))
  allocate(dat(nt_data))

  ! Define velocity model
  vel= 2500. 
  vmax = 2500.

  ! Read source wavelet
  do it=1,nst
     wavelet(it) = sin(real(it))
  end do 
 

  write(0,*) '******** Dottest ***********'
  write(0,*) 'vel, oz, dz, dst', sum(vel), oz, dz, dst
  write(0,*) 'dt_prop nt_data dt_data ot_data', dt_prop, nt_data, dt_data, ot_data
  write(0,*) 'tdelay, vmax, npad, tapeco', tdelay, vmax, npad, tapeco
  write(0,*) 'orderT, VDer, wfldmovie', orderT, VDer, wfldmovie
  write(0,*) 'wavelet, zsrc, zrec, tdeleay', sum(wavelet), zsrc, zrec, tdelay
  write(0,*) '***************************'



  ! Initiate propagation
  call init_propagate(vel,oz,dz,dst,dt_prop,nt_data,dt_data,ot_data, tdelay,vmax,npad,tapeco,orderT, wfldmovie,souWfile_, recWfile_, imgWfile_) 
  ! Initiate source parameters
  call sou_rec_init(wavelet,zsrc,zrec,tdelay)


! Dot product test
!   y1 = L x1
!   x2 = Ladj y2
!   x1*x2 = y1*y2

  allocate(x1(nnz), x2(nnz), y1(nt_data), y2(nt_data))
  do iz=1,nz 
      x1(iz) = cos(real(iz))
      x2(iz) = 0.
  end do
  do it=1,nt_data
       y2(it) = sin(real(it))
       y1(it) = 0.
  end do

  bout = born2way_op(.false.,.false.,x1,y1)
  bout = born2way_op(.true.,.false.,x2,y2)

  do iz=1,nnz
     dot1 = dot1 + x1(iz)*x2(iz)
  end do
  do it=1,nt_data
     dot2 = dot2 + y1(it)*y2(it)
  end do

  write(0,*) 'dot test result', dot1, dot2


!  deallocate (image, dat,wavelet,vel)




end program
