program DottestDP2D

  use sep
  use propDP_mod2DTX

  implicit none

  integer                                      ::  niter         ! number of iterations
  real*8                                         :: qntd, qntm     ! quantile
  real*8                                         :: thresh_d,thresh_m  ! threshold

  logical(4) :: success
  logical                               :: wfldmovie, VDer, DW, MW
  real*8                                  :: dx, dz, ox, oz, ost, dst, ot_data, dt_data, dt_prop
  real*8                                  :: drx, orx
  real*8                                  :: zsrc, xsrc, zrec,vmax,tdelay,tapeco
  real*8                                  :: dot1, dot2
  integer                               :: nnz, nz, nnx, nx, nst, nt_data, nop, npad, nrx
  integer            			:: iz , ix, ii,it, bout
  real*8, allocatable, dimension(:,:)     :: vel, mwarr, dat, DWarray
  real*8, allocatable, dimension(:,:)     :: image
  real*8, allocatable, dimension(:)       ::wavelet
  real*8, allocatable, dimension(:)       ::x1, x2, y1, y2
  character(len=128)                    :: souWfile_, imgWfile_, recWfile_
  character(len=128)                    :: outfile_

  call sep_init()

  ! velocity model info
  nz = 60
  nx = 60
  dz = 12.5;   dx=12.5
  oz = 0.;     ox=0.

 ! shot gather (data) info
  nt_data = 300
  dt_data = 0.001
  ot_data = 0.

  ! source and receiver location
  zsrc = 12.5
  xsrc = 25.
  zrec = zsrc
  nrx = 10
  drx=12.5
  orx=xsrc

  ! Read various switches and popagation specs
  VDer=.false.
  DW=.false.
  MW=.false.
  tdelay=0.002
  dt_prop=0.001
  tapeco=0.00002

  npad = 30

  nst = 1000.    ! sourcce wavelet nt
  dst = dt_data  ! source wavelet dt
  ost = 0.       ! source wavelet ot
  tdelay = 0.002  ! Delay in source wavelet


  write(0,*) 'nt_data, dt_data, ot_data', nt_data, dt_data, ot_data
  write(0,*) 'dt_prop', dt_prop  


  if (dst .ne. dt_data) call erexit('dst should equal to dt_data')
  if (dt_prop .ne. dt_data) call erexit('expect dt_prop .eq. dt_data for now')

  nnz=nz
  nz = nnz+2*npad
  nnx=nx
  nx = nnx+2*npad

  ! Allocate memory
  allocate(vel(nz,nx),image(nnz,nnx),wavelet(nt_data))
  allocate(dat(nt_data,nrx))

  ! Define velocity model
  vel= 2500. 
  vmax = 2500.

  ! Read source wavelet
  do it=1,nst
     wavelet(it) = sin(real(it))
  end do 
 

  write(0,*) '******** Dottest ***********'
  write(0,*) 'vel, oz, dz, dst', sum(vel), oz, dz, dst
  write(0,*) 'dt_prop nt_data dt_data ot_data', dt_prop, nt_data, dt_data, ot_data
  write(0,*) 'tdelay, vmax, npad, tapeco', tdelay, vmax, npad, tapeco
  write(0,*) ' VDer, wfldmovie', VDer, wfldmovie
  write(0,*) 'wavelet, zsrc, zrec, tdeleay', sum(wavelet), zsrc, zrec, tdelay
  write(0,*) '***************************'



  ! Initiate propagation
  call init_propagate(vel,oz,dz,ox, dx,dst,dt_prop,nt_data,dt_data,ot_data, tdelay,vmax,npad,tapeco, wfldmovie,souWfile_, recWfile_, imgWfile_) 
  ! Initiate source parameters
  call sou_rec_init(wavelet,zsrc,xsrc,zrec,nrx,orx,drx,tdelay)


! Dot product test
!   y1 = L x1
!   x2 = Ladj y2
!   x1*x2 = y1*y2

  write(0,*) 'before allocate x1', nnz, nnx, nt_data, nrx
  allocate(x1(nnz*nnx), x2(nnz*nnx), y1(nt_data*nrx), y2(nt_data*nrx))
  do ii=1,nnz*nnx
      x1(ii) = cos(real(ii))
      x2(ii) = 0.
  end do
  do ii=1,nt_data*nrx
       y2(ii)= sin(real(ii))
       y1(ii) = 0.
  end do

  bout = born2way_op(.false.,.false.,x1,y1)
  bout = born2way_op(.true.,.false.,x2,y2)

  do ii=1,nnz*nnx
     dot1 = dot1 + x1(ii)*x2(ii)
  end do
  do ii=1,nt_data*nrx
     dot2 = dot2 + y1(ii)*y2(ii)
  end do

  write(0,*) 'dot test result', dot1, dot2


!  deallocate (image, dat,wavelet,vel)




end program
