program RTM2D_Aper

  use sep
  use prop_mod2DAper


  implicit none

  logical(4) :: success
  logical                               :: wfldmovie, DW, MW, sder
  real                                  :: dz, oz, ost, dst, ot_data, dt_data, dt_prop
  real                                  :: zsrc, zrec,vmax,tdelay,tapeco
  real                                  :: drx, orx, dsx, osx
  real					:: ox, dx, tswmax
  integer                               :: nnz, nz, nst, nt_data, nop, npad
  integer				:: nrx, nsx, nnx, ix, nx
  integer            			:: iz , it, bout, record_sx
  real, allocatable, dimension(:,:)     :: vel
  real, allocatable, dimension(:)     :: image, dat, shotposArr
  real, allocatable, dimension(:)       ::wavelet, waveletVD
  character(len=128)                    :: souWfile_, imgWfile_, recWfile_
  character(len=128)                    :: outfile_, outmov_

  call sep_init()

  ! read in velocity model
  call from_history("n1", nz);    call from_history("n2", nx)
  call from_history("o1", oz);    call from_history("o2", ox)
  call from_history("d1", dz);    call from_history("d2", dx)

  ! read shot gather (data) file info
  call from_aux("datafile", "n1", nt_data);   call from_aux("datafile", "n2", nrx);  call from_aux("datafile","n3",nsx);
  call from_aux("datafile", "o1", ot_data);   call from_aux("datafile", "o2", orx);  call from_aux("datafile","o3",osx);
  call from_aux("datafile", "d1", dt_data);   call from_aux("datafile", "d2", drx);  call from_aux("datafile","d3",dsx);

  ! read source wavelet information
  call from_aux("aux", "n1", nst)
  call from_aux("aux", "o1", ost)
  call from_aux("aux", "d1", dst)
  call from_param("tswmax",tswmax,(nt_data-1)*dt_data+ot_data)

  ! read source location
  call from_param("zsrc",zsrc)
  call from_param("npad",npad,30)
  call from_param("zrec",zrec,zsrc)
  
  ! Read wavefield movie switch
  call from_param("wfldout",wfldmovie,.false.)
  call from_param("DW",DW,.false.);   call from_param("MW",MW,.false.)
  call from_param("souwfld",souWfile_,"souwfld.H")
  call from_param("imgwfld",imgWfile_,"imgwfld.H")  
  call from_param("recwfld",recWfile_,"recwfld.H")
  call from_param("outfile",outfile_)
  call from_param("tdelay",tdelay)
  call from_param("dt_prop",dt_prop,0.001)
  call from_param("tapeco",tapeco,0.00002);
  call from_param("sder", sder, .false.);


  write(0,*) 'nt_data, dt_data, ot_data', nt_data, dt_data, ot_data
  write(0,*) 'dt_prop tswmax', dt_prop , tswmax
  
  if (dst .ne. dt_data) call erexit('dst should equal to dt_data')
  if (dt_prop .ne. dt_data) call erexit('expect dt_prop .eq. dt_data for now')


  nnz=nz
  nz = nnz+2*npad
  nnx=nx
  nx = nnx+2*npad
  record_sx=1.

  write(0,*) 'pass 1'

  ! Allocate memory
  allocate(vel(nz,nx),image(nnz*nnx),wavelet(nt_data),waveletVD(nt_data))
  allocate(dat(nt_data*nrx*nsx))
  allocate(shotposArr(nsx))

  write(0,*) 'pass 2'

  ! Read velocity model
  call sep_read(vel(npad+1:nz-npad,npad+1:nx-npad));  call from_param("vmax",vmax);

  write(0,*) 'pass 3'

  ! padding  
  do iz=1,npad
     vel(iz,:) = vel(npad+1,:)
     vel(nz-iz+1,:) = vel(nz-npad-1,:)
  end do
  do ix=1,npad
     vel(:,ix) = vel(:,npad+1)
     vel(:,nx-ix+1) = vel(:,nx-npad-1)
  end do

  write(0,*) 'pass 4'

  ! Read source wavelet
  do it=1,nt_data
     wavelet(it) = 0.
     waveletVD(it) = 0.
  end do
  
  write(0,*) 'pass 5'  
  
  call sep_read(wavelet,"aux")
  call Ver2ndDer(.true.,.false.,wavelet,waveletVD,nt_data)
  
  write(0,*) 'pass 6'  
  
  ! Read data file
  call sep_read(dat,'datafile')
  call sep_read(shotposArr,'shotposfile')

  write(0,*) 'pass 7'

  ! Read output data
  call from_param("outfile",outfile_,'outfile.H')

  write(0,*) 'pass 8'

  ! Initiate propagation
  call init_propagate(vel,oz,dz,ox,dx,dst,dt_prop,nt_data,dt_data,ot_data, tdelay,vmax,npad,tapeco, wfldmovie,record_sx,souWfile_, recWfile_, imgWfile_) 
  ! Initiate source parameters
  if (sder) then
     call sou_rec_init(waveletVD,zsrc,nsx,dsx,osx,zrec,nrx,orx,drx,tdelay,tswmax,shotposArr)
  else
    call sou_rec_init(wavelet,zsrc,nsx,dsx,osx,zrec,nrx,orx,drx,tdelay,tswmax,shotposArr)
  end if

  bout = born2way_op(.true.,.true.,image,dat)



   call close_propagate()
   call sou_rec_close()


    call sep_write(image,outfile_)
  
    call auxputch("n1",'i',nnz,outfile_)
    call auxputch("n2",'i',nnx,outfile_)    
    call auxputch("d1",'f',dz,outfile_)
    call auxputch("d2",'f',dx,outfile_)    
    call auxputch("o1",'f',oz,outfile_)
    call auxputch("o2",'f',ox,outfile_)    
    call auxputch("label1",'s','z(m)',outfile_) 
    call auxputch("label2",'s','x(m)',outfile_)   


  deallocate (image, dat)




end program
