program born1D

  use sep
  use prop_mod1D


  implicit none

  integer                                      ::  niter         ! number of iterations
  logical(4) :: success
  logical                               :: wfldmovie, DW, MW, sder
  real                                  :: dx, dz, oz, ost, dst, ot_data, dt_data, dt_prop
  real                                  :: zsrc, zrec,vmax,tdelay,tapeco
  real                                  :: oz1, dz1
  integer                               :: nnz, nz, nz1, nst, nt_data, nop, npad
  integer            			:: iz , it, orderT, bout
  real, allocatable, dimension(:)     :: vel, mwarr, dat, DWarray
  real, allocatable, dimension(:)     :: image, resdArr
  real, allocatable, dimension(:)       ::wavelet, mwarr1D, waveletVD
  real, allocatable, dimension(:)       ::x1, x2, y1, y2
real, allocatable, dimension(:,:)       :: mmov
  character(len=128)                    :: souWfile_, imgWfile_, recWfile_
  character(len=128)                    :: outfile_, outmov_

  call sep_init()

  ! read in velocity model
  call from_history("n1", nz)
  call from_history("o1", oz)
  call from_history("d1", dz)

  ! read source wavelet information
  call from_aux("aux", "n1", nst)
  call from_aux("aux", "o1", ost)
  call from_aux("aux", "d1", dst)

  ! read shot gather (data) file info
  call from_aux("reflfile", "n1", nz1)
  call from_aux("reflfile", "o1", oz1)
  call from_aux("reflfile", "d1", dz1)

  ! Read data parmeter
  call from_param('ntrec',nt_data);
  call from_param('dtrec',dt_data);
  call from_param('otrec',ot_data);

  ! read source location
  call from_param("zsrc",zsrc)
  call from_param("npad",npad,30)
  call from_param("zrec",zrec,zsrc)
  
  ! Read wavefield movie switch
  call from_param("wfldout",wfldmovie,.false.)
  call from_param("DW",DW,.false.);   call from_param("MW",MW,.false.)
  call from_param("souwfld",souWfile_,"souwfld.H")
  call from_param("imgwfld",imgWfile_,"imgwfld.H")  
  call from_param("recwfld",recWfile_,"recwfld.H")
  call from_param("outfile",outfile_)
  call from_param("tdelay",tdelay)
  call from_param("dt_prop",dt_prop,0.001)
  call from_param("tapeco",tapeco,0.00002);
  call from_param("orderT",orderT,2);
  call from_param("sder", sder, .false.);

  write(0,*) 'nt_data, dt_data, ot_data', nt_data, dt_data, ot_data
  write(0,*) 'dt_prop', dt_prop  


  if (dst .ne. dt_data) call erexit('dst should equal to dt_data')
  if (dt_prop .ne. dt_data) call erexit('expect dt_prop .eq. dt_data for now')
  if ((nz1 .ne. nz) .or. (dz1 .ne. dz) .or. (oz1 .ne. oz) ) call erexit('velocity and refl-image does not match')

  nnz=nz
  nz = nnz+2*npad

  ! Allocate memory
  allocate(vel(nz),image(nnz),wavelet(nt_data),waveletVD(nt_data))
  allocate(dat(nt_data),resdArr(nt_data))

  ! Read velocity model
  call sep_read(vel(npad+1:nz-npad));  call from_param("vmax",vmax);
  ! padding
  do iz=1,npad
     vel(iz) = vel(npad+1)
     vel(nz-iz+1) = vel(nz-npad-1)
  end do

  ! Read source wavelet
  do it=1,nt_data
     wavelet(it) = 0.
     waveletVD(it) = 0.
  end do
  
  call sep_read(wavelet,"aux")
  call Ver2ndDer(.true.,.false.,wavelet,waveletVD,nt_data)
  
  ! Read image file
  call sep_read(image,'reflfile')

  ! Read output data
  call from_param("outfile",outfile_,'outfile.H')

  ! Initiate propagation
  call init_propagate(vel,oz,dz,dst,dt_prop,nt_data,dt_data,ot_data, tdelay,vmax,npad,tapeco,orderT, wfldmovie,souWfile_, recWfile_, imgWfile_) 
  ! Initiate source parameters
  if (sder) then
     call sou_rec_init(waveletVD,zsrc,zrec,tdelay)
  else
    call sou_rec_init(wavelet,zsrc,zrec,tdelay)
  end if

  bout = born2way_op(.false.,.true.,image,dat)

!@
!  call Ver2ndDer(.true.,.false.,dat,waveletVD,nt_data)

   call close_propagate()
   call sou_rec_close()

   call to_history("n1",nt_data,outfile_)
    call to_history("o1",ot_data,outfile_)
    call to_history("d1",dt_data,outfile_)
    call to_history  ("label1","t(s)",outfile_)
    call sep_write(dat,outfile_)

!@
!    call sep_write(waveletVD,outfile_)

  deallocate (image, dat)




end program
