program born2D

  use sep
  use prop_mod2DTX


  implicit none

  logical(4) :: success
  logical                               :: wfldmovie, DW, MW, sder
  real                                  :: dz, oz, ost, dst, ot_data, dt_data, dt_prop
  real                                  :: zsrc, zrec,vmax,tdelay,tapeco
  real                                  :: oz1, dz1, ox1, dx1, drx, orx, dsx, osx
  real					:: ox, dx, tswmax
  integer                               :: nnz, nz, nz1, nst, nt_data, nop, npad
  integer				:: nrx, nsx, nnx, ix, nx, nx1
  integer            			:: iz , it, bout, record_sx
  real, allocatable, dimension(:,:)     :: vel
  real, allocatable, dimension(:)     :: image, dat, shotposArr
  real, allocatable, dimension(:)       ::wavelet, waveletVD
  character(len=128)                    :: souWfile_, imgWfile_, recWfile_
  character(len=128)                    :: outfile_, outmov_

  call sep_init()

  ! read in velocity model
  call from_history("n1", nz);    call from_history("n2", nx)
  call from_history("o1", oz);    call from_history("o2", ox)
  call from_history("d1", dz);    call from_history("d2", dx)

  ! read source wavelet information
  call from_aux("aux", "n1", nst)
  call from_aux("aux", "o1", ost)
  call from_aux("aux", "d1", dst)
  call from_param("tswmax",tswmax,(nt_data-1)*dt_data+ot_data)

  ! read shot gather (data) file info
  call from_aux("reflfile", "n1", nz1);   call from_aux("reflfile", "n2", nx1)
  call from_aux("reflfile", "o1", oz1);   call from_aux("reflfile", "o2", ox1)
  call from_aux("reflfile", "d1", dz1);   call from_aux("reflfile", "d2", dx1)

  ! Read data parmeter
  call from_param('ntrec',nt_data);
  call from_param('dtrec',dt_data);
  call from_param('otrec',ot_data);

  call from_param('nrx',nrx);
  call from_param('drx',drx);
  call from_param('orx',orx);
  call from_param('nsx',nsx);
  call from_param('dsx',dsx);
  call from_param('osx',osx);

  ! convert from offset to absolute x
  orx=orx+osx


  ! read source location
  call from_param("zsrc",zsrc)
  call from_param("npad",npad,30)
  call from_param("zrec",zrec,zsrc)
  
  ! Read wavefield movie switch
  call from_param("wfldout",wfldmovie,.false.)
  call from_param("DW",DW,.false.);   call from_param("MW",MW,.false.)
  call from_param("souwfld",souWfile_,"souwfld.H")
  call from_param("imgwfld",imgWfile_,"imgwfld.H")  
  call from_param("recwfld",recWfile_,"recwfld.H")
  call from_param("outfile",outfile_)
  call from_param("tdelay",tdelay)
  call from_param("dt_prop",dt_prop,0.001)
  call from_param("tapeco",tapeco,0.00002);
  call from_param("sder", sder, .false.);

  write(0,*) 'nt_data, dt_data, ot_data', nt_data, dt_data, ot_data
  write(0,*) 'dt_prop tswmax', dt_prop , tswmax


  if (dst .ne. dt_data) call erexit('dst should equal to dt_data')
  if (dt_prop .ne. dt_data) call erexit('expect dt_prop .eq. dt_data for now')
  if ((nz1 .ne. nz) .or. (dz1 .ne. dz) .or. (oz1 .ne. oz) ) call erexit('velocity and refl-image does not match')

  nnz=nz
  nz = nnz+2*npad
  nnx=nx
  nx = nnx+2*npad
  record_sx=1.

  ! Allocate memory
  allocate(vel(nz,nx),image(nnz*nnx),wavelet(nt_data),waveletVD(nt_data))
  allocate(dat(nt_data*nrx*nsx))
  allocate(shotposArr(nsx))
  
  
  ! Read velocity model
  call sep_read(vel(npad+1:nz-npad,npad+1:nx-npad));  call from_param("vmax",vmax);
  ! padding
  do iz=1,npad
     vel(iz,:) = vel(npad+1,:)
     vel(nz-iz+1,:) = vel(nz-npad-1,:)
  end do
  do ix=1,npad
     vel(:,ix) = vel(:,npad+1)
     vel(:,nx-ix+1) = vel(:,nx-npad-1)
  end do
  
  ! Read source wavelet
  do it=1,nt_data
     wavelet(it) = 0.
     waveletVD(it) = 0.
  end do
  
  call sep_read(wavelet,"aux")
  call Ver2ndDer(.true.,.false.,wavelet,waveletVD,nt_data)
  
  ! Read image file
  call sep_read(image,'reflfile')
  call sep_read(shotposArr,'shotposfile')
  
  ! Read output data
  call from_param("outfile",outfile_,'outfile.H')

  write(0,*) '****'
  write(0,*) oz, dz, ox, dx, dst, dt_prop, nt_data, dt_data, ot_data, tdelay, vmax
  write(0,*) npad, tapeco, wfldmovie, record_sx, souWfile_, recWfile_, imgWfile_
  write(0,*) '****'
  write(0,*) zsrc, nsx, dsx, osx, zrec, nrx, orx, drx, tdelay, tswmax
  write(0,*) '****'
  call erexit('born2D.x exit now')

  ! Initiate propagation
  call init_propagate(vel,oz,dz,ox,dx,dst,dt_prop,nt_data,dt_data,ot_data, tdelay,vmax,npad,tapeco, wfldmovie,record_sx,souWfile_, recWfile_, imgWfile_) 
  ! Initiate source parameters
  if (sder) then
     call sou_rec_init(waveletVD,zsrc,nsx,dsx,osx,zrec,nrx,orx,drx,tdelay,tswmax,shotposArr)
  else
    call sou_rec_init(wavelet,zsrc,nsx,dsx,osx,zrec,nrx,orx,drx,tdelay,tswmax,shotposArr)
  end if

  bout = born2way_op(.false.,.true.,image,dat)


   call close_propagate()
   call sou_rec_close()

   call to_history("n1",nt_data,outfile_)
    call to_history("o1",ot_data,outfile_)
    call to_history("d1",dt_data,outfile_)
    call to_history  ("label1","t(s)",outfile_)
   call to_history("n2",nrx,outfile_)
    call to_history("o2",orx,outfile_)
    call to_history("d2",drx,outfile_)
    call to_history  ("label2","x(m)",outfile_)
   call to_history("n3",nsx,outfile_)
    call to_history("o3",osx,outfile_)
    call to_history("d3",dsx,outfile_)
    call to_history  ("label1","shot(m)",outfile_)



    call sep_write(dat,outfile_)

!@
!    call sep_write(waveletVD,outfile_)

  deallocate (image, dat)




end program
