!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! Wavefield propagation module
!! Written by Noha in 09-12, modified by Mandy in 11-12 for BEI
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

module pointer_mod
  use sep
  use smooth_mod
  implicit none 


  logical, private       :: wfldout                                    ! Switch for output wavefield movie
  real, private :: dt, dt_data, dt_prop,ot_data, velconst,tdelay                                   ! Propagation time step size
  real, private :: oz, dz, zsrc,zrec, okz, dkz, pi=4.*atan(1.)
  real          :: ztapco
  integer,private :: orderT
  integer,private :: jt, jst                                    ! Ratio between dt and dt_out
  integer,private :: itdelay                               ! itdelay
  integer,private :: nt_data, nt_prop, ntwin, nz, nnz,nkz             ! Size of data
  integer,private :: nop, npad                             ! order of Laplacian operator
  real, private,pointer :: v(:), v2(:)                          ! Velocity 
  real*8, private         :: a1z, a2z, a3z, a4z, a0        ! FD coef 8th order
  real*8, private         :: b1z, b2z, b3z, b4z, b5z, b6z, b7z, b8z, b0        ! FD coef 12th order
  real, private         ::d2t  ,d2t12                         ! coef
  real, private, allocatable :: src_func(:,:),rec_func(:,:)                 ! src wavelet, receiver data
  real, allocatable, dimension(:,:) :: source_array
  integer,private, allocatable  :: src_loc(:,:), rec_loc(:,:)               ! source and receiver location
  real, private, pointer :: ztaper(:)                     ! Taper
  real :: ovel(2),dvel(2)                                  ! Sampling of velocity model
  character(len=128), private    :: wfldfile_
  character(len=128)              :: wfldsou_, wfldimg_, wfldrec_
  contains


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
!! Wrapper function:  born2way_op
!! input:    add ([not] erase output)
!!           adj  [T] model=Ladj data; [F] data=L model
!!           model = 1D array model space vector
!!           data = 1D array data space vector
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
  integer function born2way_op(adj,add,model_in,data_in)

    logical, intent(in)          :: add, adj
    integer                      :: nz, nx, nt, ii 
    integer                      :: iz, ix, it
    real, dimension(:)          :: model_in, data_in
    real, allocatable, dimension(:,:)  :: model2D, data2D
    
        
    nz=10
    nx=10
    nt=8
    
    allocate(model2D(nz,nx),data2D(nz,nx))    
    
   model2D = reshape(model_in,(/nz,nx/))
   data2D = reshape(data_in,(/nt,nx/))
   
   do ix=1,nx
     do iz=1,nz
         ii = (ix-1)*nz+iz
         write(0,*) 'iz', iz, ix, ii, model2D(iz,ix)- model_in(ii)

     end do
  end do


  ! Now is resize the data_array
   do ix=1,nx
     do it=1,nt 
        data2D(it,ix) = cos(real(it+ix))
     end do
  end do


  write(0,*) 'is data_in affected', sum(data2D), sum(data_in)
  
  do ix=1,nx
     do it=1,nt
        ii = (ix-1)*nt+it
        data_in(ii) = data2D(it,ix)
     end do
 end do
   
     write(0,*) 'how about now', sum(data2D), sum(data_in)

   
   born2way_op=1
   

  end function born2way_op



end module
