!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! Wavefield propagation module
!! Written by Noha in 09-12, modified by Mandy in 11-12 for BEI
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

module prop_mod2DAper
  use sep
  use smooth_mod
  implicit none 


  logical, private       :: wfldout                                 ! Switch for output wavefield movie
  real, private :: dt, dt_data, dt_prop,ot_data, velconst,tdelay                                   ! Propagation time step size
  real, private :: oz, dz, ox, dx, zsrc, xsrc, zrec, orx, drx, okz, dkz, okx, dkx, pi=4.*atan(1.)
  real, private :: dsrc, osrc, ohx
  integer, private :: nsrc, record_isx
  integer,private :: jt, jst                                    ! Ratio between dt and dt_out
  integer,private :: itdelay                               ! itdelay
  integer,private :: nt_data, nt_prop, nt_sw, ntwin
  integer,private :: nx, nnx, nkx, nz, nnz, nkz               ! Size of data
  integer,private :: nop, npad, nrx                                ! Order of Laplacian operator
  real, private, allocatable ::  v2(:,:)                    ! Velocity 
  real, private         ::d2t  ,d2t12                         ! coef
  real, private         :: ztapco
  real*8, private         :: a1x, a2x, a3x, a4x, a0          ! FD coef  8th order
  real*8, private         :: a1z, a2z, a3z, a4z              ! FD coef  8th order
  real*8, dimension(:), allocatable :: bz, bx				      ! FD coef 16th order
  real, private, allocatable :: src_func(:,:)                     ! src wavelet, receiver data
    real, allocatable, dimension(:,:,:) :: source_array         ! source wavefield array
  integer,private, allocatable  :: src_loc(:,:)                 ! source location
    integer,private, allocatable  :: rec_loc(:,:,:)     ! receiver location
  real, private, allocatable :: zxtaper(:),ctaper(:,:)       ! Taper
  real, private, allocatable :: prev(:,:), next(:,:), cur(:,:)
  character(len=128)              :: wfldsou_, wfldimg_, wfldrec_
 
 
  contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Wrapper function:  born2way_op
!! input:    add ([not] erase output)
!!           adj  [T] model=Ladj data; [F] data=L model
!!           model = 1D array model space vector
!!           data = 1D array data space vector
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  integer function born2way_op(adj,add,model_in,data_in)

    integer                      :: i1, i2, ii, ix, iz, it, irx, isx
    logical, intent(in)          :: add, adj
    real, dimension(:)    :: model_in, data_in
    real, allocatable, dimension(:,:)  :: model2D
    real, allocatable, dimension(:,:,:)  :: data2D        

    if (adj) then
       if (.not. add) then
          do i1=1,size(model_in) 
             model_in(i1)=0.
          end do
       end if
    else
       if (.not. add) then
          do i1=1,size(data_in,1) 
             data_in(i1)=0.
          end do
       end if
    end if
    
    allocate(model2D(nnz,nnx),data2D(nt_data,nrx,nsrc))    
       
    do ix=1,nnx
      do iz=1,nnz
          model2D(iz,ix) = 0.
      end do
    end do

    write(0,*) 'nt_data,nrx,nsrc', nt_data,nrx,nsrc
    
    model2D = reshape(model_in,(/nnz,nnx/))
    data2D = reshape(data_in,(/nt_data,nrx,nsrc/))
   
   
   
    
    if (adj) then
       write(0,*) 'Beg Migration', adj, add, sum(model_in), sum(data_in)
          call migrate_all_shot(add,model2D,data2D)
    else
       write(0,*) 'Beg born', adj, add, sum(model_in), sum(data_in)
          call born_model_all(add,model2D,data2D)
    end if
    
    do isx=1,nsrc
    do irx=1,nrx
      do it=1,nt_data
        ii = (isx-1)*nt_data*nrx+(irx-1)*nt_data+it
        data_in(ii) =  data2D(it,irx,isx)
      end do
    end do
    end do
    do ix=1,nnx
      do iz=1,nnz
        ii = (ix-1)*nnz+iz
        model_in(ii) = model2D(iz,ix)
      end do
    end do

    if (adj) then
       write(0,*) 'End Migration',  sum(model_in), sum(data_in)
    else
       write(0,*) 'End born', adj, add, sum(model_in), sum(data_in)
    end if

        born2way_op = 0

  end function born2way_op


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
!! SUBROUTINE #1: INITIALIZE PROPAGATION
!! This subroutine is to prepare for propagation by calculating necessary parameters and coeffients
!! 1) Figure out a dt that is an integer ratio of dt_out for propagation and passes stability
!! 2) Calculate coefficients 
!! 3) Set velocity sampling, it will just be the same as in the input velocity model. 
!! 4) Other definitions
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
  
  subroutine init_propagate(vel,oz_in,dz_in,ox_in,dx_in,dst,dt_prop_in,nt_data_in,dt_data_in,ot_data_in,tdelay_in,vmax, npad_in,ztapco_in,wfldout_in,record_isx_in,wfldsou_in,wfldrec_in,wfldimg_in) 
    logical          :: wfldout_in
    integer          :: record_isx_in
    real, target     :: vel(:,:) !Input velocity model
    real             :: dz_in, dx_in !Sampling of input field
    real             :: ztapco_in
    real*8             :: idx, idz    
    real             :: dst, dt_data_in,ot_data_in, dt_prop_in,dt_stable, val, tdelay_in  !Sampling of the data
    real             :: oz_in, ox_in   !Origin of velocity model
    real             :: vmax    !the maximum velocity contained in the velocity model
    integer          :: iz, ix,nt_data_in,k,p, npad_in !integral ratios between dt_data and dt, origins
  character(len=128)    :: wfldsou_in,wfldrec_in,wfldimg_in
    
    write(0,*) 'Running: init_propagate'

    ! (1) The stability condition to calculate the new dt from dt_data:
 !   vmax=MAXVAL(vel)
    nt_data=nt_data_in
    dt_data=dt_data_in 
    ot_data=ot_data_in
    dt_prop=dt_prop_in
    nt_prop = ((nt_data-1)*dt_data)/dt_prop+1.5
    tdelay=tdelay_in
    dz=dz_in; oz=oz_in
    dx=dx_in; ox=ox_in;
    wfldout=wfldout_in
   wfldsou_ = wfldsou_in
   wfldrec_ = wfldrec_in
   wfldimg_ = wfldimg_in
   record_isx = record_isx_in


	k=ceiling(dt_data/(0.8*dz/vmax))+1	!using ceiling to get the largest possible integer,
	!!since we need vmax*dt to be less than 0.707 (2D Courant number)
	dt_stable=dt_data/k
        if (dt_prop .gt. dt_stable) write(0,*) 'Warning, dt_prop is greater than dt_stable', dt_stable, dt_prop
        dt=dt_prop
	jt=dt_data/dt_prop
        jst=dst/dt
        write(0,*) 'init_', vmax, nt_data, k, dt_data, dt,jt, dz
	if (jt .ne. 1) call erexit('jt is not equal to 1')

     write(0,*) 'nt_data nt_prop', nt_data, nt_prop
     write(0,*) 'dt_data dt_prop', dt_data, dt_prop


    


    !(2) Coefficient calculation (8th order finite diff)
    d2t=dt*dt	;   d2t12 = 1./12.
    !(2) Coefficient calculation (8th order finite diff)
    idx=DBLE(dt*dt/(dx*dx))
    idz=DBLE(dt*dt/(dz*dz))	
    a0 =  DBLE(-2.84722222222)*(idx+idz)
    a1x =  DBLE(1.6000000*idx);       a1z= DBLE(1.6000000*idz)       	
    a2x =  DBLE(-0.2000000*idx);      a2z=DBLE(-0.2000000*idz)       	
    a3x =  DBLE(0.0253968254*idx);    a3z=DBLE(0.0253968254*idz)       	
    a4x =  DBLE(-0.00178571428*idx);  a4z=DBLE(-0.00178571428*idz)     
 

 
    !(2b) Coefficient calculation (16th order finite diff)
       allocate(bz(0:8),bx(0:8))
    bz(0)  = DBLE(-3.054844104292959 *(idx+idz)); bx(0) = bz(0)  
    bz(1) = DBLE( 1.777777777756123 *idz);  bx(1) = DBLE( 1.777777777756123 *idx)
    bz(2) = DBLE( -0.311111111087384 *idz);     bx(2) = DBLE( -0.311111111087384 *idx)
    bz(3) = DBLE( 0.075420875408331 *idz);      bx(3) = DBLE( 0.075420875408331 *idx)
    bz(4) = DBLE(-0.017676767673411*idz);       bx(4) = DBLE(-0.017676767673411*idx);
    bz(5) = DBLE( 0.003480963480268 *idz);      bx(5) = DBLE( 0.003480963480268 *idx);
    bz(6) = DBLE( -5.180005178943935e-04*idz);  bx(6) = DBLE( -5.180005178943935e-04*idx);
    bz(7) = DBLE( 5.074290787522322e-05 *idz);  bx(7) = DBLE( 5.074290787522322e-05 *idx)
    bz(8) = DBLE(  -2.428127427618616e-06*idz); bx(8) = DBLE(  -2.428127427618616e-06*idx)
    

 
    !(3) Find the velocity parameters

   !(4) Other defn
   nop=8;   ! Using 16th order Laplacian



   !(5) Array size definition
   npad=npad_in; ztapco=ztapco_in
   nz=size(vel,1);
   nnz=nz-2*npad
   nx=size(vel,2);
   nnx=nx-2*npad

   !(6) Setup damping boundary
   call init_taperCos(npad)   
   
   ! (7) square the velocity
   allocate(v2(nz,nx))
   do ix=1,nx
      do iz=1,nz
         v2(iz,ix) = vel(iz,ix)*vel(iz,ix)
!	 if (mod(iz,50) .eq. 0) write(0,*) 'iz, v2', iz, v2(iz)
      end do
   end do


  end subroutine
  
    subroutine close_propagate()

     deallocate(v2) 
     deallocate(zxtaper)
     deallocate(prev,next,cur)

  end subroutine
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #2: INITIALIZE THE SOURCE
!! This subroutine define the parameters necessary to propagate source wavefields 
!! 1) Allocate src_loc(1,1): the size of this array is 1 row by 1 column
!!    Allocate src_func(nt_data*jst,1) - store the source-wavelet in time
!! 2) Set src_loc location
!! 3) Interpolate source to src_func to fit the number of steps we want to propagate
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  subroutine sou_rec_init(source, zsrc_in, nsrc_in,dsrc_in,osrc_in, zrec_in, nrx_in, orx_in, drx_in, tdelay, tswmax_in,shotposArr_in)
    real, allocatable, dimension(:) :: source !Source function 
    real :: dsrc_in, osrc_in, xsrc_in, zsrc_in !Source location 
    integer :: nsrc_in                   ! number of shots
    real :: zrec_in          ! receiver depth
    integer :: nrx_in        ! receiver x position param
    real    :: orx_in, drx_in  ! receiver x position param
    integer :: i1,i0, ir, irx, isx, it, ix, iz !number of sources
    real :: fract, tdelay, xr, xs, tswmax_in
    real, dimension(:)  :: shotposArr_in


    write(0,*) 'Running: source_init'

   !(0) Itdelay
    itdelay = (tdelay)/dt + 0.5

    nt_sw = (tswmax_in-ot_data)/dt+0.5

    !(1) Allocations:
    zsrc=zsrc_in
    nsrc=nsrc_in; dsrc=dsrc_in; osrc=osrc_in
    allocate(src_loc(2,nsrc), src_func(nt_data*jt+itdelay,1))

    !(2) Set src_loc
    src_loc(1,:)=(zsrc_in-oz)/dz+.5 +max(nop,npad)
    do isx=1,nsrc
         xs = shotposArr_in(isx)
!        xs = (isx-1)*dsrc+osrc
        src_loc(2,isx)=(xs-ox)/dx+0.5+max(nop,npad)
   end do    
   
   do isx=1,nsrc
     write(0,*) 'shotnum xs index', isx, shotposArr_in(isx), src_loc(2,isx)
   end do
   
 !  call erexit('exit now')
 
    ! error check to make sure inside grid
    if ((src_loc(1,1) .gt. nz) .and. (src_loc(1,1) .lt. 1)) call erexit('src z location out of bound')
    do isx=1,nsrc
       if ((src_loc(2,isx) .gt. nx) .and. (src_loc(2,isx) .lt. 1)) call erexit('src x location out of bound')
    end do

    !(3) Interpolate source to src_func to fit the number of steps in propagation   
    src_func=0
    ! linear interpolation
    do i0=1,size(source)
       do i1=1,jst
          fract=real(i1-1)/real(jst)
          src_func(i1+(i0-1)*jst,1)=(1.-fract)*source(i0)+fract*source(min(i0+1,size(source)))
       end do
    end do

    ! (4) Propagate and save source wavefield


    ! (5) Set receiver position
    zrec=zrec_in    
    nrx = nrx_in
    ohx = orx_in
    drx = drx_in

   write(0,*) 'receiver_init', ohx
   
   !(5.1) Allocation
   allocate (rec_loc(2,nrx,nsrc))   

   !(5.2) Set rec_loc 
   do isx = 1,nsrc
!     xsrc = osrc + (isx-1)*dsrc
     xsrc = shotposArr_in(isx)
     orx = ohx+ xsrc
     write(0,*) 'xsrc, ohx, orx', xsrc, ohx, orx
     do irx=1,nrx
       xr = (irx-1)*drx+orx
       rec_loc(2,irx,isx)=(xr-ox)/dx+0.5+max(nop,npad)
     end do   
     rec_loc(1,:,isx)=(zrec-oz)/dz+0.5 + max(nop,npad)
   end do
   
   
   !check to make sure inside grid
   do isx=1,nsrc
     if ( (rec_loc(1,1,isx) .lt. 1) .or. (rec_loc(1,1,isx) .gt. nz)) call erexit('receiver z location out of bound')
     do irx=1,nrx
        ir = rec_loc(2,irx,isx)
        if ( (ir .lt. 1) .or. (ir .gt. nx)) call erexit('receiver x location out of bound')
     end do   
   end do

   ! Allocation for the 3 slices of wavefields
   allocate(prev(nz,nx),cur(nz,nx),next(nz,nx))
   write(0,*) 'allocate prev,cur,next', size(prev,1), size(prev,2), nz, nx

  end subroutine


  subroutine sou_rec_close()

     deallocate(src_loc,src_func)
     if (allocated(rec_loc)) deallocate(rec_loc)
 
  end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #3: WAVE-FIELD PROPAGATION
!! Input: prev,cur,next =  u(t+1), u(t), u(t-1)
!! Input: rec_f = time-index marker on where to inject data
!! Input: nt = number of time step to propagate
!! output: next (represent the wavefield after propagating nt time steps)
!! (1) Apply Laplacian and time-marching (in reverse time)
!!         u(t+/-1) = - u(t-/+1) + 2*u(t) + (v^2)*(dt^2)*(Laplacian)u(t)
!! Note that propagate forward and reverse in time is symmetric
!! (2) data/source injection       
!! (3) Time stepping (update the 3 time slices)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


 subroutine propagate_wavefield(prev,cur,next,time_marker,nt)
    real, dimension(:,:)    :: prev,cur,next !Previous,current, and next wavefields
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt,ir,irx, j, sh
    real*8 :: val,val2

 ! if (mod(time_marker,200) == 0) write(0,*) 'Running: propagate_wavefield', time_marker ,'out of', nt_data*jt , sum(prev), sum(cur), sum(next)

        
    do it=1,nt
       val = 0.
       
!       write(0,*) 'pro-wave 1', time_marker, sum(prev),  sum(cur), sum(next)
       
        do ix=1+nop,nx-nop
           do iz=1+nop,nz-nop
             val = DBLE(v2(iz,ix))*(bz(0)*DBLE(cur(iz,ix)) + & 
                   bz(1)*(DBLE(cur(iz+1,ix))+DBLE(cur(iz-1,ix))) + &
                   bz(2)*(DBLE(cur(iz+2,ix))+DBLE(cur(iz-2,ix))) + &
                   bz(3)*(DBLE(cur(iz+3,ix))+DBLE(cur(iz-3,ix))) + &
                   bz(4)*(DBLE(cur(iz+4,ix))+DBLE(cur(iz-4,ix))) + &
                   bz(5)*(DBLE(cur(iz+5,ix))+DBLE(cur(iz-5,ix))) + &
                   bz(6)*(DBLE(cur(iz+6,ix))+DBLE(cur(iz-6,ix))) + &
                   bz(7)*(DBLE(cur(iz+7,ix))+DBLE(cur(iz-7,ix))) + &
                   bz(8)*(DBLE(cur(iz+8,ix))+DBLE(cur(iz-8,ix))) + &                                                                                                                                     
		   bx(1)*(DBLE(cur(iz,ix+1))+DBLE(cur(iz,ix-1))) + &
		   bx(2)*(DBLE(cur(iz,ix+2))+DBLE(cur(iz,ix-2))) + &
		   bx(3)*(DBLE(cur(iz,ix+3))+DBLE(cur(iz,ix-3))) + &
		   bx(4)*(DBLE(cur(iz,ix+4))+DBLE(cur(iz,ix-4))) + &
		   bx(5)*(DBLE(cur(iz,ix+5))+DBLE(cur(iz,ix-5))) + &
		   bx(6)*(DBLE(cur(iz,ix+6))+DBLE(cur(iz,ix-6))) + &
		   bx(7)*(DBLE(cur(iz,ix+7))+DBLE(cur(iz,ix-7))) + &
		   bx(8)*(DBLE(cur(iz,ix+8))+DBLE(cur(iz,ix-8))))		   		   		   		   		   		   		   
	     next(iz,ix) = REAL(val)
           end do
       end do       
       
       
       do ix=1,nx
         do iz=1,nz
            next(iz,ix) = next(iz,ix) - prev(iz,ix) + 2*cur(iz,ix)
!            next(iz,ix) = next(iz,ix) + 0.02*cur(iz,ix)            
         end do
       end do     

!       write(0,*) 'pro-wave 2', time_marker, sum(prev),  sum(cur), sum(next)
     

	         

	! Time-stepping
        do ix=1,nx
           do iz=1,nz
		prev(iz,ix) = cur(iz,ix)
		cur(iz,ix) = next(iz,ix)
	   end do
	end do


        ! Applying damping boundary
        call taperingCos(prev,cur,next)
        
        
   end do   
end subroutine



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #4: WAVE-FIELD PROPAGATION Adjoint
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine propagate_wavefield_adj(prev,cur,next,time_marker,nt)
    real, dimension(:,:)    :: prev,cur,next !Previous,current, and next wavefields
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt, j, sh
    real*8 :: val, val2

! write(0,*) 'Running: propagate_wavefield', time_marker ,'out of', nt_data*jt

    do it=1,nt
       val = 0.
     
	!(1) Tapering
	call taperingCos(prev,cur,next)

        do ix=1+nop,nx-nop
           do iz=1+nop,nz-nop
             val = bz(0)*DBLE(v2(iz,ix))*DBLE(cur(iz,ix)) + &
		bz(1)*(DBLE(v2(iz+1,ix))*DBLE(cur(iz+1,ix))+ DBLE(v2(iz-1,ix))*DBLE(cur(iz-1,ix))) + &
		bz(2)*(DBLE(v2(iz+2,ix))*DBLE(cur(iz+2,ix))+ DBLE(v2(iz-2,ix))*DBLE(cur(iz-2,ix))) + &
		bz(3)*(DBLE(v2(iz+3,ix))*DBLE(cur(iz+3,ix))+ DBLE(v2(iz-3,ix))*DBLE(cur(iz-3,ix))) + &
		bz(4)*(DBLE(v2(iz+4,ix))*DBLE(cur(iz+4,ix))+ DBLE(v2(iz-4,ix))*DBLE(cur(iz-4,ix))) + &
		bz(5)*(DBLE(v2(iz+5,ix))*DBLE(cur(iz+5,ix))+ DBLE(v2(iz-5,ix))*DBLE(cur(iz-5,ix))) + &
		bz(6)*(DBLE(v2(iz+6,ix))*DBLE(cur(iz+6,ix))+ DBLE(v2(iz-6,ix))*DBLE(cur(iz-6,ix))) + &
		bz(7)*(DBLE(v2(iz+7,ix))*DBLE(cur(iz+7,ix))+ DBLE(v2(iz-7,ix))*DBLE(cur(iz-7,ix))) + &
	        bz(8)*(DBLE(v2(iz+8,ix))*DBLE(cur(iz+8,ix))+ DBLE(v2(iz-8,ix))*DBLE(cur(iz-8,ix))) + &
	        bx(1)*(DBLE(v2(iz,ix+1))*DBLE(cur(iz,ix+1))+ DBLE(v2(iz,ix-1))*DBLE(cur(iz,ix-1))) + &
	        bx(2)*(DBLE(v2(iz,ix+2))*DBLE(cur(iz,ix+2))+ DBLE(v2(iz,ix-2))*DBLE(cur(iz,ix-2))) + &
	        bx(3)*(DBLE(v2(iz,ix+3))*DBLE(cur(iz,ix+3))+ DBLE(v2(iz,ix-3))*DBLE(cur(iz,ix-3))) + &
	        bx(4)*(DBLE(v2(iz,ix+4))*DBLE(cur(iz,ix+4))+ DBLE(v2(iz,ix-4))*DBLE(cur(iz,ix-4))) + &
	        bx(5)*(DBLE(v2(iz,ix+5))*DBLE(cur(iz,ix+5))+ DBLE(v2(iz,ix-5))*DBLE(cur(iz,ix-5))) + &
	        bx(6)*(DBLE(v2(iz,ix+6))*DBLE(cur(iz,ix+6))+ DBLE(v2(iz,ix-6))*DBLE(cur(iz,ix-6))) + &
	        bx(7)*(DBLE(v2(iz,ix+7))*DBLE(cur(iz,ix+7))+ DBLE(v2(iz,ix-7))*DBLE(cur(iz,ix-7))) + &
	        bx(8)*(DBLE(v2(iz,ix+8))*DBLE(cur(iz,ix+8))+ DBLE(v2(iz,ix-8))*DBLE(cur(iz,ix-8)))        	        	        	      	     
	     next(iz,ix) = REAL(val)             	
           end do
       end do


       do ix=1,nx
         do iz=1,nz
           next(iz,ix) = next(iz,ix) - prev(iz,ix) + 2*cur(iz,ix)
!           next(iz,ix) = next(iz,ix) + 0.02*cur(iz,ix)                        
         end do
       end do
	 

        !(3) Time stepping
        do ix=1,nx
           do iz=1,nz
		prev(iz,ix) = cur(iz,ix)
		cur(iz,ix) = next(iz,ix)
	   end do
	end do        
        
   end do   
end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #6: IMAGING
!! (1) apply zero-lag cross correlation imaging condition
!! (2) Output to wavefield movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 


 subroutine simple_imaging_condition(source0, receiver,image)
    real, dimension(:,:) :: image,  source0, receiver
    integer              ::  iz,ix

    !(1) Apply imaging condition
    do ix=1,nnx
       do iz=1,nnz
          image(iz,ix) = image(iz,ix) + source0(iz,ix)*receiver(iz+npad,ix+npad)          
       end do
    end do
    

 end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #7: Forward Modeling Born Injection
!! (1) Apply Born source injection to receiver wavefield
!! (2) Output wavefield movies  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 


 subroutine simple_born_inject(source0,receiver,image)

    real, dimension(:,:) :: image,  source0, receiver
    integer              :: iz, ix

    !(1) Apply Born source injection to receiver wavefield
    do ix=1,nnx
       do iz=1,nnz
          receiver(iz+npad,ix+npad) = receiver(iz+npad,ix+npad) + source0(iz,ix)*image(iz,ix)
       end do
    end do

 end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #8: Propagate and save source wavefield
!! (1) Allocation
!! (2) propagate source wavefield
!! (3) Save to array at every jt time-step
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 


  subroutine create_source_array(array,isx)
  real, allocatable, dimension(:,:,:)  :: array !Array of nz,nt_data
  integer :: i, it, iz, it_cur, id, is, ix, iid, isx


!  write(0,*) 'Running: create_source_array', size(array,1), size(array,2), size(array,3), nnz, nnx, nt_data


  ! (1)Initiate  
  do ix=1,nx
   do iz=1,nz
       prev(iz,ix) = 0.; 
       cur(iz,ix) = 0.;
       next(iz,ix) = 0.;
    end do
  end do

  do it=1,size(array,3)
    do ix=1,nnx
    do iz=1,nnz
       array(iz,ix,it) = 0.
    end do
    end do
  end do
  
  do i=1,nt_sw+itdelay
  

  
      it_cur = i
      !(1) source/receiver injection
      cur(src_loc(1,isx),src_loc(2,isx))=cur(src_loc(1,isx),src_loc(2,isx))+src_func(it_cur,1)

      !(2) Apply Laplacian
       call propagate_wavefield(prev,cur,next,it_cur,1)

     ! (3) Save a time-slice of sou wavefield to the array at every jt time-step
       id = (i-itdelay)
       if ( (id .ge. 1) .and. (id .le. nt_prop)) then 
            do ix=1,nnx  
            do iz=1,nnz
              array(iz,ix,id)=next(iz+npad,ix+npad)
            end do
            end do
       end if

  end do
      
      
!      do id=1,nt_prop
!         if (mod(id,10) .eq. 0) call sep_write(array(:,:,id),"swfd.H",n1out=nz,n2out=nx,esize=4)
!      end do
      
      
   


end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #10: Propagate the Born receiver wavefield
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine propagate_born(image,dat,source_array,isx)
  real, dimension(:,:,:) :: source_array  
  real, dimension(:,:) :: image, dat
  integer :: i, id, ix, it,iz, irx, io, iid,isx  !Internal time variable


!  write(0,*) 'Running: propagate_born'  


  ! (1)Initiate  
  do ix=1,nx
   do iz=1,nz
       prev(iz,ix) = 0.; 
       cur(iz,ix) = 0.;
       next(iz,ix) = 0.;
    end do
  end do
  
  it=0

!  write(0,*) 'pass 1 @@'

  do i=1,nt_prop

! write(0,*) 'pass 2 @@', i

     ! (2) propagate rec wavefield 
         call propagate_wavefield_adj(prev,cur,next,i,1)
!  write(0,*) 'pass 3 @@', i


     ! (3) Inject Born Source at every time-step    
       if ((i .ge. 1) .and. (i .le. nt_sw)) then
       	  call simple_born_inject(source_array(:,:,i),cur,image)
       end if       	  

!  write(0,*) 'pass 4 @@', i

 !    !(4) Output wavefield movies  
 !      if ( (wfldout) .and. (mod(i,jt) .eq. 0)) then
 !        call sep_write(image,wfldimg_)
 !        call sep_write(source_array(:,:,id),wfldsou_)
 !        call sep_write(cur,wfldrec_)
 !      end if

!  write(0,*) 'pass 5 @@', i
    ! (5) collect data
       do irx=1,size(dat,2)
         dat(i,irx) = dat(i,irx) +  cur(rec_loc(1,irx,isx),rec_loc(2,irx,isx))  
       end do
!  write(0,*) 'pass 6 @@', i

  end do

!  write(0,*) 'propagate_born ends'

end subroutine 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #9: Propagate receiver wavefield and apply imaging condition
!! (1) Allocation
!! (2) propagate receiver wavefield
!! (3) apply imaging condition
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine propagate_receivers(image,source_array,recs,isx)
  real, dimension(:,:,:) :: source_array  
  real, dimension(:,:) :: image,recs
  integer :: i,io,id,irx, ir,ix,iz,it,iid,isx


!  write(0,*) 'Running: propagate_receivers'  

  ! (1)Initiate  
  do ix=1,nx
   do iz=1,nz
       prev(iz,ix) = 0.; 
       cur(iz,ix) = 0.;
       next(iz,ix) = 0.;
    end do
  end do


  do i=nt_prop,1,-1

!  write(0,*) 'pr 1', i

   ! (3) propagate rec wavefield for jt time-step (note: it is reverse in time)
        call propagate_wavefield(prev,cur,next,i,1)
 


!  write(0,*) 'pr 2', i
  
    !(4) source/receiver injection
     do ir=1,size(rec_loc,2)
        cur(rec_loc(1,ir,isx),rec_loc(2,ir,isx))=cur(rec_loc(1,ir,isx),rec_loc(2,ir,isx))+recs(i,ir)
     end do  

 !     if (mod(i,200) .eq. 0) write(0,*) 'pass 2', i, sum(cur), sum(source_array(:,:,i))     
  
       if ((i .ge. 1) .and. (i .le. nt_sw)) then
          call simple_imaging_condition(source_array(:,:,i),cur,image)
      end if

    !(6) Output wavefield movies  
    if ((mod(i,50) .eq. 0) .and. (isx .eq. 6) .and. .false. ) then    
!    if ((wfldout) .and. (mod(i,jt) .eq. 0) ) then
	call sep_write(image,"image-movie.H")
	call sep_write(source_array(:,:,i),"source-movie.H")
	call sep_write(prev,"rec-movie.H")
	    call auxputch("n1",'i',nnz,"image-movie.H")
            call auxputch("n2",'i',nnx,"image-movie.H")
    	    call auxputch("n3",'i',nt_prop/50,"image-movie.H")
	    call auxputch("d1",'f',dz,"image-movie.H")
	    call auxputch("d2",'f',dx,"image-movie.H")
	    call auxputch("d3",'f',dt_prop*50,"image-movie.H")
	    call auxputch("o1",'f',oz,"image-movie.H")
	    call auxputch("o2",'f',ox,"image-movie.H")
	    call auxputch("o3",'f',0.,"image-movie.H")
	    call auxputch("label1",'s','z(m)',"image-movie.H") 
	    call auxputch("label2",'s','x(m)',"image-movie.H") 
	    call auxputch("label3",'s','tiem(s)',"image-movie.H")
	    
	    call auxputch("n1",'i',nnz,"source-movie.H")
            call auxputch("n2",'i',nnx,"source-movie.H")
    	    call auxputch("n3",'i',nt_prop/50,"source-movie.H")
	    call auxputch("d1",'f',dz,"source-movie.H")
	    call auxputch("d2",'f',dx,"source-movie.H")
	    call auxputch("d3",'f',dt_prop*50,"source-movie.H")
	    call auxputch("o1",'f',oz,"source-movie.H")
	    call auxputch("o2",'f',ox,"source-movie.H")
	    call auxputch("o3",'f',0.,"source-movie.H")
	    call auxputch("label1",'s','z(m)',"source-movie.H") 
	    call auxputch("label2",'s','x(m)',"source-movie.H") 
	    call auxputch("label3",'s','tiem(s)',"source-movie.H")	     

	    call auxputch("n1",'i',nz,"rec-movie.H")
            call auxputch("n2",'i',nx,"rec-movie.H")
    	    call auxputch("n3",'i',nt_prop/50,"rec-movie.H")
	    call auxputch("d1",'f',dz,"rec-movie.H")
	    call auxputch("d2",'f',dx,"rec-movie.H")
	    call auxputch("d3",'f',dt_prop*50,"rec-movie.H")
	    call auxputch("o1",'f',oz,"rec-movie.H")
	    call auxputch("o2",'f',ox,"rec-movie.H")
	    call auxputch("o3",'f',0.,"rec-movie.H")
	    call auxputch("label1",'s','z(m)',"rec-movie.H") 
	    call auxputch("label2",'s','x(m)',"rec-movie.H") 
	    call auxputch("label3",'s','tiem(s)',"rec-movie.H")	     

    		
!       call sep_write(image,wfldimg_)
!       call sep_write(source_array(:,:,i),wfldsou_)
!       call sep_write(prev,wfldrec_)
    end if

  end do

! write(0,*) 'Endding: propagate_receivers'  

end subroutine


   
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #11: MIGRATE A SHOT
!! (1) Allocation
!! (2) Calculate source wavefield
!! (3) Calculate rec wavefield and apply imaging condition
!! (4) Write wavefield movie header (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine migrate_all_shot(add,image,dat)

  logical              :: add
  integer              :: isx,it, iz, ix
  real, dimension(:,:) :: image !Image 
  real, dimension(:,:,:) :: dat ! time, x, shot
  
  
  do isx=1,nsrc
    allocate(source_array(nnz,nnx,nt_sw))  
    do it=1,nt_sw
     do ix=1,nnx
     do iz=1,nnz
       source_array(iz,ix,it) = 0.
     end do
     end do
    end do
    
      call create_source_array(source_array,isx)

      write(0,*) 'migrate_one_shot beg', isx, sum(image),sum(dat(:,:,isx)), sum(source_array)
            
      call migrate_one_shot(.true.,image,dat(:,:,isx),isx)
      write(0,*) 'migrate_one_shot end', isx, sum(image),sum(dat(:,:,isx)), sum(source_array)


!      call sep_write(image,"image-each-shots.H")
!    call auxputch("n1",'i',nnz,"image-each-shots.H")
!    call auxputch("n2",'i',nnx,"image-each-shots.H")
!    call auxputch("n3",'i',nsrc,"image-each-shots.H")
!    call auxputch("d1",'f',dz,"image-each-shots.H")
!    call auxputch("d2",'f',dx,"image-each-shots.H")
!    call auxputch("d3",'f',1.,"image-each-shots.H")
!    call auxputch("o1",'f',oz,"image-each-shots.H")
!    call auxputch("o2",'f',ox,"image-each-shots.H")
!    call auxputch("o3",'f',0.,"image-each-shots.H")
!    call auxputch("label1",'s','z(m)',"image-each-shots.H") 
!    call auxputch("label2",'s','x(m)',"image-each-shots.H") 
!    call auxputch("label3",'s','shot#',"image-each-shots.H") 
  
        
      
    deallocate(source_array) 
  end do
  
  
  
end subroutine


subroutine migrate_one_shot(add,image,dat,isx)

  integer              :: isx
  logical              :: add
  real, dimension(:,:) :: image,dat !Image and data


  ! (1) Calculate receiver wavefield and apply imaging condition 
    call propagate_receivers(image,source_array,dat,isx)

    

  ! (2) Write wavefield movies header 
!  if (wfldout .and. (isx .eq. record_isx)) then
!       call write_movie_header1(wfldimg_)
!       call write_movie_header1(wfldsou_)
!       call write_movie_header2(wfldrec_)
!   end if

end subroutine 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #13: BORN Model
!! (1) Allocation
!! (2) Propagate and save source wavefield
!! (3) Collect data
!! (4) Output source movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


subroutine born_model_all(add,image,dat) 

  logical              :: add
    integer              :: isx, it, iz ,ix
  real, dimension(:,:) :: image
    real, dimension(:,:,:) :: dat

  ! (1) born modeling
  do isx=1,nsrc
    allocate(source_array(nnz,nnx,nt_sw))  
    do it=1,nt_sw
     do ix=1,nnx
     do iz=1,nnz
       source_array(iz,ix,it) = 0.
     end do
     end do
    end do
    
    call create_source_array(source_array,isx)
    write(0,*) 'born_one beg', isx, sum(image), sum(dat(:,:,isx))
    call born_model_one(add,image,dat(:,:,isx),isx) 
     write(0,*) 'born_one end', isx,sum(image), sum(dat(:,:,isx))
     
     deallocate(source_array)
  end do

end subroutine  


subroutine born_model_one(add,image,dat,isx) 

  logical              :: add
  integer		:: isx
  real, dimension(:,:) :: image, dat

  ! (1) born modeling
  call propagate_born(image,dat,source_array,isx)

  ! (2) Write wavefield movies header 
!  if (wfldout .and. (isx .eq. record_isx)) then
!       call write_movie_header1(wfldsou_)
!       call write_movie_header2(wfldrec_)
!       call write_movie_header1(wfldimg_)
!   end if
 

end subroutine  


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #12: MODEL A SHOT
!! (1) Allocation
!! (2) Propagate and save source wavefield
!! (3) Collect data
!! (4) Output source movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine model_shot(dat,order,outname_) 

  integer              :: itrec, iz,ix, it, icur,order, irx,isx
  real                 :: rz,trec
  real, dimension(:,:,:) :: dat !Shot gather
  real, allocatable, dimension(:,:,:) :: source_array
  character(len=128)   :: outname_

  write(0,*) 'Running: model_shot'
  

  do isx=1,nsrc

  ! (2) Propagate and save source wavefield
      call create_source_array(source_array,isx)


  if (wfldout .and. (isx .eq. record_isx)) then
    call sep_write(source_array,wfldsou_)
    call auxputch("n1",'i',nz,wfldsou_)
    call auxputch("n2",'i',nx,wfldsou_)
    call auxputch("n3",'i',nt_data,wfldsou_)
    call auxputch("d1",'f',dz,wfldsou_)
    call auxputch("d2",'f',dx,wfldsou_)
    call auxputch("d3",'f',dt_data,wfldsou_)
    call auxputch("o1",'f',oz,wfldsou_)
    call auxputch("o2",'f',ox,wfldsou_)
    call auxputch("o3",'f',ot_data,wfldsou_)
    call auxputch("label1",'s','z(m)',wfldsou_) 
    call auxputch("label2",'s','x(m)',wfldsou_) 
    call auxputch("label3",'s','t(s)',wfldsou_) 
  end if 

 !  write(0,*) 'zsrc/vel', zsrc/velconst, tdelay      

  ! (3) collect data
  do irx=1,size(dat,2)
     dat(:,irx,isx) = source_array(rec_loc(1,irx,isx),rec_loc(2,irx,isx),:)  
  end do
  
  end do

end subroutine  


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #14: CLEANUP! 
!! Deallocate memory specific to a shot
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

 subroutine clean_up_array()
     write(0,*) 'dealloc src_loc, src_func, rec_loc'
     deallocate (src_loc,src_func,rec_loc) 
     write(0,*) 'dealloc zxtaper ctaper'
     deallocate(zxtaper,ctaper)	
 end subroutine
 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #15: Write wavefield movie header file
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
  subroutine write_movie_header1(filename)

    character(len=128)  :: filename

    call auxputch("n1",'i',nnz,filename)
    call auxputch("n2",'i',nnx,filename)
    call auxputch("n3",'i',nt_data-1,filename)
    call auxputch("d1",'f',dz,filename)
    call auxputch("d2",'f',dx,filename)    
    call auxputch("d3",'f',dt*jt,filename)
    call auxputch("o1",'f',oz,filename)
    call auxputch("o2",'f',ox,filename)    
    call auxputch("o3",'f',0.,filename)
    call auxputch("label1",'s','z(m)',filename) 
    call auxputch("label2",'s','x(m)',filename)     
    call auxputch("label3",'s','t(s)',filename) 

   end subroutine

  subroutine write_movie_header2(filename)

    character(len=128)  :: filename

    call auxputch("n1",'i',nz,filename)
    call auxputch("n2",'i',nx,filename)    
    call auxputch("n3",'i',nt_data-1,filename)
    call auxputch("d1",'f',dz,filename)
    call auxputch("d2",'f',dx,filename)    
    call auxputch("d3",'f',dt*jt,filename)
    call auxputch("o1",'f',oz,filename)
    call auxputch("o2",'f',ox,filename)    
    call auxputch("o3",'f',0.,filename)
    call auxputch("label1",'s','z(m)',filename) 
    call auxputch("label2",'s','x(m)',filename)     
    call auxputch("label3",'s','t(s)',filename) 

   end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE # 16: Apply tapering
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine taperingCos(prev,cur,next)

    integer  :: iz, ix
    real, dimension(:,:) :: prev,cur,next !Previous,current, and next wavefields

       ! Corner
       do ix=1,npad
          do iz=1,npad

              prev(iz,ix) = prev(iz,ix)*ctaper(iz,ix);                   ! top left
              prev(iz,nx-ix+1) = prev(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
              prev(nz-iz+1,nx-ix+1) = prev(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              prev(nz-iz+1,ix) = prev(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left

              cur(iz,ix) = cur(iz,ix)*ctaper(iz,ix);                   ! top left
              cur(iz,nx-ix+1) = cur(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
              cur(nz-iz+1,nx-ix+1) = cur(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              cur(nz-iz+1,ix) = cur(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left

              next(iz,ix) = next(iz,ix)*ctaper(iz,ix);                   ! top left
              next(iz,nx-ix+1) = next(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
              next(nz-iz+1,nx-ix+1) = next(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              next(nz-iz+1,ix) = next(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left
          end do
       end do

       ! Left Right
       do iz=npad+1,nz-npad
          prev(iz, 1:npad) = prev(iz, 1:npad) * zxtaper(1:npad:+1); 
          prev(iz,nx-npad+1:nx) = prev(iz,nx-npad+1:nx) * zxtaper(npad:1:-1)
          cur(iz, 1:npad) = cur(iz, 1:npad) * zxtaper(1:npad:+1);
          cur(iz,nx-npad+1:nx) = cur(iz,nx-npad+1:nx) * zxtaper(npad:1:-1)
          next(iz, 1:npad) = next(iz, 1:npad) * zxtaper(1:npad:+1);
          next(iz,nx-npad+1:nx) = next(iz,nx-npad+1:nx) * zxtaper(npad:1:-1)
       end do

       ! Top and Bottom
       do ix=npad+1,nx-npad
          prev(1:npad,ix) = prev(1:npad,ix) * zxtaper(1:npad:+1);
          prev(nz-npad+1:nz,ix) = prev(nz-npad+1:nz,ix) * zxtaper(npad:1:-1)
          cur(1:npad,ix) = cur(1:npad,ix) * zxtaper(1:npad:+1);
          cur(nz-npad+1:nz,ix) = cur(nz-npad+1:nz,ix) * zxtaper(npad:1:-1)
          next(1:npad,ix) = next(1:npad,ix) * zxtaper(1:npad:+1);
          next(nz-npad+1:nz,ix) = next(nz-npad+1:nz,ix) * zxtaper(npad:1:-1)
       end do



  end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 17: Tapering Setup
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
    subroutine init_taperCos(npad)

    integer                     :: iz, ix,ii, npad
    real                       :: dist
    real                        :: pi=3.14159265359


    allocate(zxtaper(npad), ctaper(npad,npad))

       ! all four edges
       do ii=1,npad
          zxtaper(ii) = 0.985 + 0.015*cos(pi*(npad-ii)/npad)
       end do

       ! Corner
       do ix=1,npad
          do iz=1,npad
              dist = sqrt(1.001*(npad-ix)*(npad-ix)/npad/npad + 0.999999*(npad-iz)*(npad-iz)/npad/npad)
              ctaper(iz,ix) =  0.985 + 0.015*cos(pi*dist)
          end do
       end do

  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 21: Take derivative for 1D array
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

  subroutine FirstDer(adj,souIn,souOut,nt,dt)

    logical :: adj
    integer :: i, nt
    real    :: dt, rdt
    real, dimension(:) :: souIn, souOut
    real, dimension(:), allocatable :: temp

   allocate(temp(nt))
   temp(1:nt) = souIn(1:nt)
    
    souOut(1:nt) = 0.
    
    
    if (.not. adj) then
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1)+ temp(i)
         souOut(i)   = souOut(i) - temp(i)
      end do
      
    else
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1) + temp(i+1)-temp(i)
      end do
         
    end if    

    deallocate(temp)
  end subroutine 



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 23: Take Vertical Derivative of image
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
  subroutine Ver2ndDer(adj,add,m_in,m_out,nz)
  ! Take the second derivative vertically and its adjoint

  logical                   :: adj,add
  integer                   :: nz,iz
  real                      :: fac
  real, dimension(:), allocatable :: temp
  real, dimension(:)      :: m_in, m_out


  write(0,*) 'inside Ver2ndDer', adj, add

  allocate(temp(nz))
  do iz=1,nz
     temp(iz) = m_in(iz)
     m_out(iz) = 0.
  end do
  if (.not. add) m_out=0.
 
   fac = 0.5

         if (adj) then
            do iz=2,nz-1
	            m_out(iz) = m_out(iz) + fac*(2.*m_in(iz) - m_in(iz+1) - m_in(iz-1))
	    end do
         else
	    do iz=2,nz-1
	            m_out(iz)   = m_out(iz) + fac*2*m_in(iz)
        	    m_out(iz+1) = m_out(iz+1) - fac*m_in(iz)
        	    m_out(iz-1) = m_out(iz-1) - fac*m_in(iz)
	    end do
         end if


  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE # XX: Apply Laplacian 
!! 1) Fourier transform
!! 2) multiply by -kz**2
!! 3) Inverse fourier transform
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine ApplyLaplac2D(inArr,outArr)

  integer               :: iz, ix
  real, dimension(:,:)    :: inArr, outArr
    real*8             :: val
  !write(0,*) 'Inside Apply Laplac'
  !write(0,*) 'nz, dz', nz, dz


       ! (1) Apply Laplacian 
       do ix=1+nop,nx-nop   
          do iz=nop+1,nz-nop
             val = a0*inArr(iz,ix)
             val = val + a1x*DBLE(inArr(iz,ix+1))
	     val = val + a1x*DBLE(inArr(iz,ix-1))
             val = val + a2x*DBLE(inArr(iz,ix+2))
	     val = val + a2x*DBLE(inArr(iz,ix-2))
             val = val + a3x*DBLE(inArr(iz,ix+3))
	     val = val + a3x*DBLE(inArr(iz,ix-3))
             val = val + a4x*DBLE(inArr(iz,ix+4))
	     val = val + a4x*DBLE(inArr(iz,ix-4))
             val = val + a1z*DBLE(inArr(iz+1,ix))
	     val = val + a1z*DBLE(inArr(iz-1,ix))
             val = val + a2z*DBLE(inArr(iz+2,ix))
	     val = val + a2z*DBLE(inArr(iz-2,ix))
             val = val + a3z*DBLE(inArr(iz+3,ix))
	     val = val + a3z*DBLE(inArr(iz-3,ix))
             val = val + a4z*DBLE(inArr(iz+4,ix))
	     val = val + a4z*DBLE(inArr(iz-4,ix))
	     outArr(iz,ix) = REAL(val)
           end do
        end do



end subroutine


end module


