!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! Wavefield propagation module
!! Written by Noha in 09-12, modified by Mandy in 11-12 for BEI
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

module prop_mod2DTX_jt
  use sep
  use smooth_mod
  implicit none 


  logical, private       :: wfldout                                 ! Switch for output wavefield movie
  real, private :: dt, dt_data, dt_prop,ot_data, velconst,tdelay                                   ! Propagation time step size
  real, private :: oz, dz, ox, dx, zsrc, xsrc, zrec, orx, drx, okz, dkz, okx, dkx, pi=4.*atan(1.)
  integer,private :: jt, jst                                    ! Ratio between dt and dt_out
  integer,private :: itdelay                               ! itdelay
  integer,private :: nt_data, nt_prop, ntwin
  integer,private :: nx, nnx, nkx, nz, nnz, nkz               ! Size of data
  integer,private :: nop, npad, nrx                                ! Order of Laplacian operator
  real, private,pointer :: v(:,:), v2(:,:)                    ! Velocity 
  real, private         ::d2t  ,d2t12                         ! coef
  real, private         :: ztapco
  real*8, private         :: a1x, a2x, a3x, a4x, a0          ! FD coef  8th order
  real*8, private         :: a1z, a2z, a3z, a4z              ! FD coef  8th order
  real*8, dimension(:), allocatable :: bz, bx				      ! FD coef 16th order
  real, private, pointer :: src_func(:,:)                     ! src wavelet, receiver data
    real, allocatable, dimension(:,:,:) :: source_array         ! source wavefield array
  integer,private, pointer  :: src_loc(:,:), rec_loc(:,:)     ! source and receiver location
  real, private, pointer :: zxtaper(:),ctaper(:,:)       ! Taper
  complex, dimension(:), pointer    :: in1, out1, in2, out2
  complex, dimension(:,:), pointer  ::  cdatapad
  character(len=128)              :: wfldsou_, wfldimg_, wfldrec_
 
 
  contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Wrapper function:  born2way_op
!! input:    add ([not] erase output)
!!           adj  [T] model=Ladj data; [F] data=L model
!!           model = 1D array model space vector
!!           data = 1D array data space vector
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  integer function born2way_op(adj,add,model_in,data_in)

    integer                      :: i1, i2, ii, ix, iz, it, irx
    logical, intent(in)          :: add, adj
    real, dimension(:)    :: model_in, data_in
    real, allocatable, dimension(:,:)  :: model2D, data2D
    

    if (adj) then
       if (.not. add) then
          do i1=1,size(model_in) 
             model_in(i1)=0.
          end do
       end if
    else
       if (.not. add) then
          do i1=1,size(data_in,1) 
             data_in(i1)=0.
          end do
       end if
    end if
    

    
    allocate(model2D(nnz,nnx),data2D(nt_data,nrx))    
    

    
    do ix=1,nnx
      do iz=1,nnz
          model2D(iz,ix) = 0.
      end do
    end do


    
    model2D = reshape(model_in,(/nnz,nnx/))
    data2D = reshape(data_in,(/nt_data,nrx/))
   

    
    if (adj) then
       write(0,*) 'Beg Migration', adj, add, sum(model_in), sum(data_in)
          call migrate_shot(add,model2D,data2D)
    else
       write(0,*) 'Beg born', adj, add, sum(model_in), sum(data_in)
          call born_model(add,model2D,data2D)
    end if

    do irx=1,nrx
      do it=1,nt_data
        ii = (irx-1)*nt_data+it
        data_in(ii) =  data2D(it,irx)
      end do
    end do
    do ix=1,nnx
      do iz=1,nnz
        ii = (ix-1)*nnz+iz
        model_in(ii) = model2D(iz,ix)
      end do
    end do

    if (adj) then
       write(0,*) 'End Migration',  sum(model_in), sum(data_in)
    else
       write(0,*) 'End born', adj, add, sum(model_in), sum(data_in)
    end if

        born2way_op = 0

  end function born2way_op


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
!! SUBROUTINE #1: INITIALIZE PROPAGATION
!! This subroutine is to prepare for propagation by calculating necessary parameters and coeffients
!! 1) Figure out a dt that is an integer ratio of dt_out for propagation and passes stability
!! 2) Calculate coefficients 
!! 3) Set velocity sampling, it will just be the same as in the input velocity model. 
!! 4) Other definitions
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
  
  subroutine init_propagate(vel,oz_in,dz_in,ox_in,dx_in,dst,dt_prop_in,nt_data_in,dt_data_in,ot_data_in,tdelay_in,vmax, npad_in,ztapco_in,wfldout_in,wfldsou_in,wfldrec_in,wfldimg_in) 
    logical          :: wfldout_in
    real, target     :: vel(:,:) !Input velocity model
    real             :: dz_in, dx_in !Sampling of input field
    real             :: ztapco_in
    real*8             :: idx, idz    
    real             :: dst, dt_data_in,ot_data_in, dt_prop_in,dt_stable, val, tdelay_in  !Sampling of the data
    real             :: oz_in, ox_in   !Origin of velocity model
    real             :: vmax    !the maximum velocity contained in the velocity model
    integer          :: iz, ix,nt_data_in,k,p, npad_in !integral ratios between dt_data and dt, origins
  character(len=128)    :: wfldsou_in,wfldrec_in,wfldimg_in
    
    write(0,*) 'Running: init_propagate'

    ! (1) The stability condition to calculate the new dt from dt_data:
 !   vmax=MAXVAL(vel)
    nt_data=nt_data_in
    dt_data=dt_data_in 
    ot_data=ot_data_in
    dt_prop=dt_prop_in
    nt_prop = ((nt_data-1)*dt_data)/dt_prop+2.5
    tdelay=tdelay_in
    dz=dz_in; oz=oz_in
    dx=dx_in; ox=ox_in;
    wfldout=wfldout_in
   wfldsou_ = wfldsou_in
   wfldrec_ = wfldrec_in
   wfldimg_ = wfldimg_in


	k=ceiling(dt_data/(0.8*dz/vmax))+1	!using ceiling to get the largest possible integer,
	!!since we need vmax*dt to be less than 0.707 (2D Courant number)
	dt_stable=dt_data/k
        if (dt_prop .gt. dt_stable) write(0,*) 'Warning, dt_prop is greater than dt_stable', dt_stable, dt_prop
        dt=dt_prop
	jt=dt_data/dt_prop
        jst=dst/dt
        write(0,*) 'init_', vmax, nt_data, k, dt_data, dt,jt, dz
!	if (jt .ne. 1) call erexit('jt is not equal to 1')

     write(0,*) 'nt_data nt_prop', nt_data, nt_prop
     write(0,*) 'dt_data dt_prop', dt_data, dt_prop

    write(0,*) 'oass 1', jt
    


    !(2) Coefficient calculation (8th order finite diff)
    d2t=dt*dt	;   d2t12 = 1./12.
    !(2) Coefficient calculation (8th order finite diff)
    idx=DBLE(dt*dt/(dx*dx))
    idz=DBLE(dt*dt/(dz*dz))	
    a0 =  DBLE(-2.84722222222)*(idx+idz)
    a1x =  DBLE(1.6000000*idx);       a1z= DBLE(1.6000000*idz)       	
    a2x =  DBLE(-0.2000000*idx);      a2z=DBLE(-0.2000000*idz)       	
    a3x =  DBLE(0.0253968254*idx);    a3z=DBLE(0.0253968254*idz)       	
    a4x =  DBLE(-0.00178571428*idx);  a4z=DBLE(-0.00178571428*idz)     
 
     write(0,*) 'oass 2'
 
    !(2b) Coefficient calculation (16th order finite diff)
       allocate(bz(0:8),bx(0:8))
    bz(0)  = DBLE(-3.054844104292959 *(idx+idz)); bx(0) = bz(0)  
    bz(1) = DBLE( 1.777777777756123 *idz);  bx(1) = DBLE( 1.777777777756123 *idx)
    bz(2) = DBLE( -0.311111111087384 *idz);     bx(2) = DBLE( -0.311111111087384 *idx)
    bz(3) = DBLE( 0.075420875408331 *idz);      bx(3) = DBLE( 0.075420875408331 *idx)
    bz(4) = DBLE(-0.017676767673411*idz);       bx(4) = DBLE(-0.017676767673411*idx);
    bz(5) = DBLE( 0.003480963480268 *idz);      bx(5) = DBLE( 0.003480963480268 *idx);
    bz(6) = DBLE( -5.180005178943935e-04*idz);  bx(6) = DBLE( -5.180005178943935e-04*idx);
    bz(7) = DBLE( 5.074290787522322e-05 *idz);  bx(7) = DBLE( 5.074290787522322e-05 *idx)
    bz(8) = DBLE(  -2.428127427618616e-06*idz); bx(8) = DBLE(  -2.428127427618616e-06*idx)
    
    write(0,*) 'oass 3'        
 
    !(3) Find the velocity parameters
    v=>vel      

   !(4) Other defn
   nop=8;   ! Using 16th order Laplacian

    write(0,*) 'oass 4'

   !(5) Array size definition
   npad=npad_in; ztapco=ztapco_in
   nz=size(v,1);
   nnz=nz-2*npad
   nx=size(v,2);
   nnx=nx-2*npad

   !(6) Setup damping boundary
   call init_taperCos(npad)
   
    write(0,*) 'oass 5'   
   
   ! (7) square the velocity
   allocate(v2(nz,nx))
   do ix=1,nx
      do iz=1,nz
         v2(iz,ix) = v(iz,ix)*v(iz,ix)
!	 if (mod(iz,50) .eq. 0) write(0,*) 'iz, v2', iz, v2(iz)
      end do
   end do


  end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #2: INITIALIZE THE SOURCE
!! This subroutine define the parameters necessary to propagate source wavefields 
!! 1) Allocate src_loc(1,1): the size of this array is 1 row by 1 column
!!    Allocate src_func(nt_data*jst,1) - store the source-wavelet in time
!! 2) Set src_loc location
!! 3) Interpolate source to src_func to fit the number of steps we want to propagate
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  subroutine sou_rec_init(source, zsrc_in, xsrc_in, zrec_in, nrx_in, orx_in, drx_in, tdelay)
    real, allocatable, dimension(:) :: source !Source function 
    real :: zsrc_in, xsrc_in !Source location 
    real :: zrec_in          ! receiver depth
    integer :: nrx_in        ! receiver x position param
    real    :: orx_in, drx_in  ! receiver x position param
    integer :: i1,i0, ir, irx, it, ix, iz !number of sources
    real :: fract, tdelay, xr


    write(0,*) 'Running: source_init'

   !(0) Itdelay
    itdelay = (tdelay)/dt + 0.5


    !(1) Allocations:
    allocate(src_loc(2,1), src_func(nt_data*jt+itdelay,1))

    !(2) Set src_loc
    zsrc=zsrc_in
    src_loc(1,1)=(zsrc_in-oz)/dz+.5 +max(nop,npad)
    src_loc(2,1)=(xsrc_in-ox)/dx+.5 +max(nop,npad)

    ! error check to make sure inside grid
    if ((src_loc(1,1) .gt. nz) .and. (src_loc(1,1) .lt. 1)) call erexit('src z location out of bound')
    if ((src_loc(2,1) .gt. nx) .and. (src_loc(2,1) .lt. 1)) call erexit('src x location out of bound')


    !(3) Interpolate source to src_func to fit the number of steps in propagation   
    src_func=0
    ! linear interpolation
    do i0=1,size(source)
       do i1=1,jst
          fract=real(i1-1)/real(jst)
          src_func(i1+(i0-1)*jst,1)=(1.-fract)*source(i0)+fract*source(min(i0+1,size(source)))
       end do
    end do

    ! (4) Propagate and save source wavefield
    allocate(source_array(nnz,nnx,nt_prop))  
    do it=1,nt_data
     do ix=1,nnx
     do iz=1,nnz
       source_array(iz,ix,it) = 0.
     end do
     end do
    end do
    call create_source_array(source_array)

    ! (5) Set receiver position
    zrec=zrec_in    
    nrx = nrx_in
    orx = orx_in
    drx = drx_in

   write(0,*) 'receiver_init'
   
   !(5.1) Allocation
   allocate (rec_loc(2,nrx))   

   !(5.2) Set rec_loc 
   do irx=1,nrx
     xr = (irx-1)*drx+orx
     rec_loc(2,irx)=(xr-ox)/dx+0.5+max(nop,npad)
   end do   
   rec_loc(1,:)=(zrec-oz)/dz+0.5 + max(nop,npad)
   
   
  ! write(0,*) 'irz', rec_loc(1,1), zrec
   !check to make sure inside grid
   if ( (rec_loc(1,1) .lt. 1) .or. (rec_loc(1,1) .gt. nz)) call erexit('receiver z location out of bound')
   do irx=1,nrx
      ir = rec_loc(2,irx)
      if ( (ir .lt. 1) .or. (ir .gt. nx)) call erexit('receiver x location out of bound')
   end do   

  end subroutine



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #3: WAVE-FIELD PROPAGATION
!! Input: prev,cur,next =  u(t+1), u(t), u(t-1)
!! Input: rec_f = time-index marker on where to inject data
!! Input: nt = number of time step to propagate
!! output: next (represent the wavefield after propagating nt time steps)
!! (1) Apply Laplacian and time-marching (in reverse time)
!!         u(t+/-1) = - u(t-/+1) + 2*u(t) + (v^2)*(dt^2)*(Laplacian)u(t)
!! Note that propagate forward and reverse in time is symmetric
!! (2) data/source injection       
!! (3) Time stepping (update the 3 time slices)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


 subroutine propagate_wavefield(prev,cur,next,time_marker,nt)
    real, pointer, dimension(:,:)    :: prev,cur,next !Previous,current, and next wavefields
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt,ir,irx, j, sh
    real*8 :: val,val2

 ! if (mod(time_marker,200) == 0) write(0,*) 'Running: propagate_wavefield', time_marker ,'out of', nt_data*jt , sum(prev), sum(cur), sum(next)
    
    !if (time_marker .eq. 6000) write(0,*) '*****', size(prev), size(cur), nz, nop, nnz, nz

    
    do it=1,nt
       val = 0.
       
!       write(0,*) 'pro-wave 1', time_marker, sum(prev),  sum(cur), sum(next)
       
        do ix=1+nop,nx-nop
           do iz=1+nop,nz-nop
             val = DBLE(v2(iz,ix))*(bz(0)*DBLE(cur(iz,ix)) + & 
                   bz(1)*(DBLE(cur(iz+1,ix))+DBLE(cur(iz-1,ix))) + &
                   bz(2)*(DBLE(cur(iz+2,ix))+DBLE(cur(iz-2,ix))) + &
                   bz(3)*(DBLE(cur(iz+3,ix))+DBLE(cur(iz-3,ix))) + &
                   bz(4)*(DBLE(cur(iz+4,ix))+DBLE(cur(iz-4,ix))) + &
                   bz(5)*(DBLE(cur(iz+5,ix))+DBLE(cur(iz-5,ix))) + &
                   bz(6)*(DBLE(cur(iz+6,ix))+DBLE(cur(iz-6,ix))) + &
                   bz(7)*(DBLE(cur(iz+7,ix))+DBLE(cur(iz-7,ix))) + &
                   bz(8)*(DBLE(cur(iz+8,ix))+DBLE(cur(iz-8,ix))) + &                                                                                                                                     
		   bx(1)*(DBLE(cur(iz,ix+1))+DBLE(cur(iz,ix-1))) + &
		   bx(2)*(DBLE(cur(iz,ix+2))+DBLE(cur(iz,ix-2))) + &
		   bx(3)*(DBLE(cur(iz,ix+3))+DBLE(cur(iz,ix-3))) + &
		   bx(4)*(DBLE(cur(iz,ix+4))+DBLE(cur(iz,ix-4))) + &
		   bx(5)*(DBLE(cur(iz,ix+5))+DBLE(cur(iz,ix-5))) + &
		   bx(6)*(DBLE(cur(iz,ix+6))+DBLE(cur(iz,ix-6))) + &
		   bx(7)*(DBLE(cur(iz,ix+7))+DBLE(cur(iz,ix-7))) + &
		   bx(8)*(DBLE(cur(iz,ix+8))+DBLE(cur(iz,ix-8))))		   		   		   		   		   		   		   
	     next(iz,ix) = REAL(val)
           end do
       end do       
       
       
       do ix=1,nx
         do iz=1,nz
!            next(iz,ix) = next(iz,ix) - prev(iz,ix) + 2*cur(iz,ix)
            next(iz,ix) = next(iz,ix) + 0.02*cur(iz,ix)            
         end do
       end do     

!       write(0,*) 'pro-wave 2', time_marker, sum(prev),  sum(cur), sum(next)
     
       ! (1) Apply Laplacian and time-marching
!         do ix=nop+1,nx-nop
!	  do iz=1,nop
!	        val=0.; val2=0.;
!	     do j=1,iz
!	        sh = nop-j+1
!	        val = val + bz(sh)*DBLE(cur(iz+sh,ix))
!	        val2 = val2 + bz(sh)*DBLE(cur(nz-iz+1-sh,ix))
!	        next(iz,ix) = REAL(val)*DBLE(v2(iz,ix))
!	        next(nz-iz+1,ix) = REAL(val2)*DBLE(v2(nz-iz+1,ix))
!	     end do
!	  end do
!	end do
	
!         do iz=nop+1,nz-nop
!	  do ix=1,nop
!	        val=0.; val2=0.;
!	     do j=1,ix
!	        sh = nop-j+1
!	        val = val + bx(sh)*DBLE(cur(iz,ix+sh))
!	        val2 = val2 + bx(sh)*DBLE(cur(iz,nx-ix+1-sh))
!	        next(iz,ix) = REAL(val)*DBLE(v2(iz,ix))
!	        next(iz,nx-ix+1) = REAL(val2)*DBLE(v2(iz,nx-ix+1))
!	     end do
!	  end do
!	end do
	
	         

	! Time-stepping
        do ix=1,nx
           do iz=1,nz
		prev(iz,ix) = cur(iz,ix)
		cur(iz,ix) = next(iz,ix)
	   end do
	end do

!       write(0,*) 'pro-wave 3', time_marker, sum(prev),  sum(cur), sum(next)	
	
        ! Applying damping boundary
!        call taperingCos(prev,cur,next)
        
        
   end do   
end subroutine



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #4: WAVE-FIELD PROPAGATION Adjoint
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine propagate_wavefield_adj(prev,cur,next,time_marker,nt)
    real, dimension(:,:)    :: prev,cur,next !Previous,current, and next wavefields
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt, j, sh
    real*8 :: val, val2

! write(0,*) 'Running: propagate_wavefield', time_marker ,'out of', nt_data*jt

!  if (mod(time_marker,500) == 0) write(0,*) 'Running: propagate_wavefield_adj', time_marker ,'out of', nt_data*jt

    do it=1,nt
       val = 0.
     
	!(1) Tapering
!	call taperingCos(prev,cur,next)

!       write(0,*) 'pro-wave adj 1', time_marker, sum(prev),  sum(cur), sum(next)

        do ix=1+nop,nx-nop
           do iz=1+nop,nz-nop
             val = bz(0)*DBLE(v2(iz,ix))*DBLE(cur(iz,ix)) + &
		bz(1)*(DBLE(v2(iz+1,ix))*DBLE(cur(iz+1,ix))+ DBLE(v2(iz-1,ix))*DBLE(cur(iz-1,ix))) + &
		bz(2)*(DBLE(v2(iz+2,ix))*DBLE(cur(iz+2,ix))+ DBLE(v2(iz-2,ix))*DBLE(cur(iz-2,ix))) + &
		bz(3)*(DBLE(v2(iz+3,ix))*DBLE(cur(iz+3,ix))+ DBLE(v2(iz-3,ix))*DBLE(cur(iz-3,ix))) + &
		bz(4)*(DBLE(v2(iz+4,ix))*DBLE(cur(iz+4,ix))+ DBLE(v2(iz-4,ix))*DBLE(cur(iz-4,ix))) + &
		bz(5)*(DBLE(v2(iz+5,ix))*DBLE(cur(iz+5,ix))+ DBLE(v2(iz-5,ix))*DBLE(cur(iz-5,ix))) + &
		bz(6)*(DBLE(v2(iz+6,ix))*DBLE(cur(iz+6,ix))+ DBLE(v2(iz-6,ix))*DBLE(cur(iz-6,ix))) + &
		bz(7)*(DBLE(v2(iz+7,ix))*DBLE(cur(iz+7,ix))+ DBLE(v2(iz-7,ix))*DBLE(cur(iz-7,ix))) + &
	        bz(8)*(DBLE(v2(iz+8,ix))*DBLE(cur(iz+8,ix))+ DBLE(v2(iz-8,ix))*DBLE(cur(iz-8,ix))) + &
	        bx(1)*(DBLE(v2(iz,ix+1))*DBLE(cur(iz,ix+1))+ DBLE(v2(iz,ix-1))*DBLE(cur(iz,ix-1))) + &
	        bx(2)*(DBLE(v2(iz,ix+2))*DBLE(cur(iz,ix+2))+ DBLE(v2(iz,ix-2))*DBLE(cur(iz,ix-2))) + &
	        bx(3)*(DBLE(v2(iz,ix+3))*DBLE(cur(iz,ix+3))+ DBLE(v2(iz,ix-3))*DBLE(cur(iz,ix-3))) + &
	        bx(4)*(DBLE(v2(iz,ix+4))*DBLE(cur(iz,ix+4))+ DBLE(v2(iz,ix-4))*DBLE(cur(iz,ix-4))) + &
	        bx(5)*(DBLE(v2(iz,ix+5))*DBLE(cur(iz,ix+5))+ DBLE(v2(iz,ix-5))*DBLE(cur(iz,ix-5))) + &
	        bx(6)*(DBLE(v2(iz,ix+6))*DBLE(cur(iz,ix+6))+ DBLE(v2(iz,ix-6))*DBLE(cur(iz,ix-6))) + &
	        bx(7)*(DBLE(v2(iz,ix+7))*DBLE(cur(iz,ix+7))+ DBLE(v2(iz,ix-7))*DBLE(cur(iz,ix-7))) + &
	        bx(8)*(DBLE(v2(iz,ix+8))*DBLE(cur(iz,ix+8))+ DBLE(v2(iz,ix-8))*DBLE(cur(iz,ix-8)))        	        	        	      	     
	     next(iz,ix) = REAL(val)             	
           end do
       end do


       do ix=1,nx
         do iz=1,nz
!            next(iz,ix) = next(iz,ix) - prev(iz,ix) + 2*cur(iz,ix)
           next(iz,ix) = next(iz,ix) + 0.02*cur(iz,ix)                        
         end do
       end do

!       write(0,*) 'pro-wave adj 2', time_marker, sum(prev),  sum(cur), sum(next)
  
       ! (2) Apply Laplacian and time-marching
!	do ix=nop+1,nx-nop
!	  do iz=1,nop
!		val=0.; val2=0.;
!	     do j=1,iz
!	        sh = nop-j+1
!	        val = val + bz(sh)*(v2(iz+sh,ix)*cur(iz+sh,ix))
!	        val2 = val2 + bz(sh)*(v2(nz-iz+1-sh,ix)*cur(nz-iz+1-sh,ix))
!	        next(iz,ix) = REAL(val)
!	        next(nz-iz+1,ix) = REAL(val2)
!	     end do
!	  end do
!	end do
	
!	do iz=nop+1,nz-nop
!	  do ix=1,nop
!		val=0.; val2=0.;
!	     do j=1,ix
!	        sh = nop-j+1
!	        val = val + bx(sh)*(v2(iz,ix+sh)*cur(iz,ix+sh))
!	        val2 = val2 + bx(sh)*(v2(iz,nx-ix+1-sh)*cur(iz,nx-ix+1-sh))
!	        next(iz,ix) = REAL(val)
!	        next(iz,nx-ix+1) = REAL(val2)
!	     end do
!	  end do
!	end do
		
	  


        !(3) Time stepping
        do ix=1,nx
           do iz=1,nz
		prev(iz,ix) = cur(iz,ix)
		cur(iz,ix) = next(iz,ix)
	   end do
	end do
        
 !      write(0,*) 'pro-wave adj 3', time_marker, sum(prev),  sum(cur), sum(next)        
        
   end do   
end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #6: IMAGING
!! (1) apply zero-lag cross correlation imaging condition
!! (2) Output to wavefield movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 


 subroutine simple_imaging_condition(source0, receiver,image)
    real, dimension(:,:) :: image,  source0, receiver
    integer              ::  iz,ix

    !(1) Apply imaging condition
    do ix=1,nnx
       do iz=1,nnz
          image(iz,ix) = image(iz,ix) + source0(iz,ix)*receiver(iz+npad,ix+npad)
       end do
    end do
    

 end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #7: Forward Modeling Born Injection
!! (1) Apply Born source injection to receiver wavefield
!! (2) Output wavefield movies  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 


 subroutine simple_born_inject(source0,receiver,image)

    real, dimension(:,:) :: image,  source0, receiver
    integer              :: iz, ix

    !(1) Apply Born source injection to receiver wavefield
    do ix=1,nnx
       do iz=1,nnz
          receiver(iz+npad,ix+npad) = receiver(iz+npad,ix+npad) + source0(iz,ix)*image(iz,ix)
       end do
    end do

 end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #8: Propagate and save source wavefield
!! (1) Allocation
!! (2) propagate source wavefield
!! (3) Save to array at every jt time-step
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 


  subroutine create_source_array(array)
  real, allocatable, dimension(:,:,:)  :: array !Array of nz,nt_data
  integer :: i, iz, it_cur, id, is, ix, iid
  real, pointer, dimension(:,:) :: prev,cur,next

!  write(0,*) 'Running: create_source_array',  itdelay, nt_prop, size(array,2)
  ! (1) Allocation  
  allocate(prev(nz,nx),cur(nz,nx),next(nz,nx))
  do ix=1,nx
   do iz=1,nz
       prev(iz,ix) = 0.; 
       cur(iz,ix) = 0.;
       next(iz,ix) = 0.;
    end do
  end do

  
  do i=1,nt_prop+itdelay
      it_cur = i
      !(1) source/receiver injection
      cur(src_loc(1,1),src_loc(2,1))=cur(src_loc(1,1),src_loc(2,1))+src_func(it_cur,1)

      !(2) Apply Laplacian
       call propagate_wavefield(prev,cur,next,it_cur,1)

     ! (3) Save a time-slice of sou wavefield to the array at every jt time-step
       id = (i-itdelay)
       if ( (id .ge. 1)) then 
         if ( (mod(id,jt) .eq. 0) .or. (jt .eq. 1)) then
            iid = id/jt
            if ((iid .ge. 1) .and. (iid .le. nt_data)) then
!            write(0,*) 'i, id, jt, iid', i, id, jt, iid
            do ix=1,nnx  
            do iz=1,nnz
              array(iz,ix,iid)=next(iz+npad,ix+npad)
            end do
            end do
            end if
         end if
       end if

  end do
        
           
  deallocate(prev,cur,next)

end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #10: Propagate the Born receiver wavefield
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine propagate_born(image,dat,source_array)
  real, dimension(:,:,:) :: source_array  
  real, dimension(:,:) :: image, dat
  integer :: i, id, ix, it,iz, irx, io, iid  !Internal time variable
  real, allocatable,dimension(:,:) :: prev2,cur2,next2, next2v2

!  write(0,*) 'Running: propagate_born'  

  ! (1) Allocation
  allocate(prev2(nz,nx),cur2(nz,nx),next2(nz,nx))


  ! (2) Initialization
    do ix=1,nx
      do iz=1,nz
        prev2(iz,ix)=0; cur2(iz,ix)=0; next2(iz,ix)=0;
     end do
    end do
  it=0

  do i=1,nt_prop

     ! (2) propagate rec wavefield 
         call propagate_wavefield_adj(prev2,cur2,next2,i,1)



     ! (3) Inject Born Source at every time-step    
      if ((mod(i,jt) .eq. 0) .or. (jt .eq. 1)) then
       iid = i/jt
      ! io=0
       if ((iid .ge. 1) .and. (iid .le. nt_data)) then
       	  call simple_born_inject(source_array(:,:,iid),cur2,image)
       end if       	  

      end if


     !(4) Output wavefield movies  
       if ( (wfldout) .and. (mod(i,jt) .eq. 0)) then
         call sep_write(image,wfldimg_)
         call sep_write(source_array(:,:,id),wfldsou_)
         call sep_write(cur2,wfldrec_)
       end if


    ! (5) collect data
    
!     write(0,*) 'size data', size(dat,1), size(dat,2), jt
     if ((mod(i,jt) .eq. 0) .or. (jt .eq. 1)) then
       iid = i/jt
       if ((iid .ge. 1) .and. (iid .le. nt_data)) then
          write(0,*) 'collect data', i, iid
          do irx=1,size(dat,2)
            dat(iid,irx) = dat(iid,irx) +  cur2(rec_loc(1,irx),rec_loc(2,irx))  
          end do
       end if       	  
    end if      


  end do


  call erexit('exit now')
 ! Check size of dat 
 !  write(0,*) '@@dat', size(dat,1), size(dat,2), size(rec_loc,1),size(rec_loc,2)

  deallocate(prev2,cur2,next2)

  write(0,*) 'propagate_born ends'

end subroutine 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #9: Propagate receiver wavefield and apply imaging condition
!! (1) Allocation
!! (2) propagate receiver wavefield
!! (3) apply imaging condition
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine propagate_receivers(image,source_array,recs)
  real, dimension(:,:,:) :: source_array  
  real, dimension(:,:) :: image,recs
  integer :: i,io,id,irx, ir,ix,iz,it,iid
  real, pointer,dimension(:,:) :: prev1,cur1,next1,next1v2

!  write(0,*) 'Running: propagate_receivers'  

  ! (1) Allocation
  allocate(prev1(nz,nx),cur1(nz,nx),next1(nz,nx))


  ! (2) Initialization
    do ix=1,nx
      do iz=1,nz
        prev1(iz,ix)=0; cur1(iz,ix)=0; next1(iz,ix)=0;
     end do
    end do

  do i=nt_prop,1,-1


   ! (3) propagate rec wavefield for jt time-step (note: it is reverse in time)
        call propagate_wavefield(prev1,cur1,next1,i,1)
 
 !     if (mod(i,2) .eq. 0) write(0,*) 'pass 2', i, sum(prev1), sum(cur1), sum(next1)


    !(4) source/receiver injection
     do ir=1,size(rec_loc,2)
        cur1(rec_loc(1,ir),rec_loc(2,ir))=cur1(rec_loc(1,ir),rec_loc(2,ir))+recs(i,ir)
     end do  
     
     if ((mod(i,jt) .eq. 0) .or. (jt .eq. 1)) then
       iid = i/jt
       if ((iid .ge. 1) .and. (iid .le. nt_data))    call simple_imaging_condition(source_array(:,:,iid),cur1,image)

      end if
     

    !(6) Output wavefield movies  
    if ((wfldout) .and. (mod(i,jt) .eq. 0) ) then
       call sep_write(image,wfldimg_)
       call sep_write(source_array(:,:,nt_prop-i),wfldsou_)
       call sep_write(prev1,wfldrec_)
    end if


  end do

  deallocate(prev1,cur1,next1)


end subroutine


   
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #11: MIGRATE A SHOT
!! (1) Allocation
!! (2) Calculate source wavefield
!! (3) Calculate rec wavefield and apply imaging condition
!! (4) Write wavefield movie header (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine migrate_shot(add,image,dat)

  logical              :: add
  real, dimension(:,:) :: image,dat !Image and data

  write(0,*) 'migrate_shot 1'

  ! (1) Calculate receiver wavefield and apply imaging condition 
    call propagate_receivers(image,source_array,dat)

  write(0,*) 'migrate_shot 2'

  ! (2) Write wavefield movies header 
  if (wfldout) then
       call write_movie_header1(wfldimg_)
       call write_movie_header1(wfldsou_)
       call write_movie_header2(wfldrec_)
   end if

  write(0,*) 'migrate_shot 3'

end subroutine 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #13: BORN Model
!! (1) Allocation
!! (2) Propagate and save source wavefield
!! (3) Collect data
!! (4) Output source movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine born_model(add,image,dat) 

  logical              :: add
  real, dimension(:,:) :: image, dat

  write(0,*) 'born_model 1'

  ! (1) born modeling
  call propagate_born(image,dat,source_array)

  write(0,*) 'born_model 2'

  ! (2) Write wavefield movies header 
  if (wfldout) then
       call write_movie_header1(wfldsou_)
       call write_movie_header2(wfldrec_)
       call write_movie_header1(wfldimg_)
   end if
 
   write(0,*) 'born_model 3'  

end subroutine  


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #12: MODEL A SHOT
!! (1) Allocation
!! (2) Propagate and save source wavefield
!! (3) Collect data
!! (4) Output source movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine model_shot(dat,order,outname_) 

  integer              :: itrec, iz,ix, it, icur,order, irx
  real                 :: rz,trec
  real, dimension(:,:) :: dat !Shot gather
  real, allocatable, dimension(:,:,:) :: source_array
  character(len=128)   :: outname_

  write(0,*) 'Running: model_shot'
  
  ! (1) Allocation
  write(0,*) 'allocate', nz,nx, nt_data
  allocate(source_array(nz,nx,nt_data))

  ! (2) Propagate and save source wavefield
      call create_source_array(source_array)


  if (wfldout) then
    call sep_write(source_array,wfldsou_)
    call auxputch("n1",'i',nz,wfldsou_)
    call auxputch("n2",'i',nx,wfldsou_)
    call auxputch("n3",'i',nt_data,wfldsou_)
    call auxputch("d1",'f',dz,wfldsou_)
    call auxputch("d2",'f',dx,wfldsou_)
    call auxputch("d3",'f',dt_data,wfldsou_)
    call auxputch("o1",'f',oz,wfldsou_)
    call auxputch("o2",'f',ox,wfldsou_)
    call auxputch("o3",'f',ot_data,wfldsou_)
    call auxputch("label1",'s','z(m)',wfldsou_) 
    call auxputch("label2",'s','x(m)',wfldsou_) 
    call auxputch("label3",'s','t(s)',wfldsou_) 
  end if 

 !  write(0,*) 'zsrc/vel', zsrc/velconst, tdelay      

  ! (3) collect data
  do irx=1,size(dat,2)
     dat(:,irx) = source_array(rec_loc(1,irx),rec_loc(2,irx),:)  
  end do

  if (order .eq. 4) then
  ! Output data
    call auxputch("n1",'i',nt_data,outname_)
    call auxputch("n2",'i',nrx,outname_)
    call auxputch("d1",'f',dt_data,outname_)
    call auxputch("d2",'f',drx,outname_)
    call auxputch("o1",'f',ot_data,outname_)
    call auxputch("o2",'f',orx,outname_)
    call auxputch("label1",'s','t(s)',outname_) 
    call auxputch("label2",'s','x(m)',outname_) 
  call sep_write(dat,outname_)
    write(0,*) 'order', order, outname_
  end if

  write(0,*) 'sum dat, source_array', sum(dat), sum(source_array)

 ! deallocate(source_array) 
!  write(0,*) 'deallocate source_array'
!  call clean_up_array()

!  write(0,*) 'after dallocate'

end subroutine  


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #14: CLEANUP! 
!! Deallocate memory specific to a shot
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

 subroutine clean_up_array()
     write(0,*) 'dealloc src_loc, src_func, rec_loc'
     deallocate (src_loc,src_func,rec_loc) 
     write(0,*) 'dealloc zxtaper ctaper'
     deallocate(zxtaper,ctaper)	
 end subroutine
 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #15: Write wavefield movie header file
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
  subroutine write_movie_header1(filename)

    character(len=128)  :: filename

    call auxputch("n1",'i',nnz,filename)
    call auxputch("n2",'i',nnx,filename)
    call auxputch("n3",'i',nt_data-1,filename)
    call auxputch("d1",'f',dz,filename)
    call auxputch("d2",'f',dx,filename)    
    call auxputch("d3",'f',dt*jt,filename)
    call auxputch("o1",'f',oz,filename)
    call auxputch("o2",'f',ox,filename)    
    call auxputch("o3",'f',0.,filename)
    call auxputch("label1",'s','z(m)',filename) 
    call auxputch("label2",'s','x(m)',filename)     
    call auxputch("label3",'s','t(s)',filename) 

   end subroutine

  subroutine write_movie_header2(filename)

    character(len=128)  :: filename

    call auxputch("n1",'i',nz,filename)
    call auxputch("n2",'i',nx,filename)    
    call auxputch("n3",'i',nt_data-1,filename)
    call auxputch("d1",'f',dz,filename)
    call auxputch("d2",'f',dx,filename)    
    call auxputch("d3",'f',dt*jt,filename)
    call auxputch("o1",'f',oz,filename)
    call auxputch("o2",'f',ox,filename)    
    call auxputch("o3",'f',0.,filename)
    call auxputch("label1",'s','z(m)',filename) 
    call auxputch("label2",'s','x(m)',filename)     
    call auxputch("label3",'s','t(s)',filename) 

   end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE # 16: Apply tapering
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine taperingCos(prev,cur,next)

    integer  :: iz, ix
    real, pointer, dimension(:,:) :: prev,cur,next !Previous,current, and next wavefields

       ! Corner
       do ix=1,npad
          do iz=1,npad

              prev(iz,ix) = prev(iz,ix)*ctaper(iz,ix);                   ! top left
              prev(iz,nx-ix+1) = prev(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
              prev(nz-iz+1,nx-ix+1) = prev(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              prev(nz-iz+1,ix) = prev(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left

              cur(iz,ix) = cur(iz,ix)*ctaper(iz,ix);                   ! top left
              cur(iz,nx-ix+1) = cur(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
              cur(nz-iz+1,nx-ix+1) = cur(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              cur(nz-iz+1,ix) = cur(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left

              next(iz,ix) = next(iz,ix)*ctaper(iz,ix);                   ! top left
              next(iz,nx-ix+1) = next(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
              next(nz-iz+1,nx-ix+1) = next(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              next(nz-iz+1,ix) = next(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left
          end do
       end do

       ! Left Right
       do iz=npad+1,nz-npad
          prev(iz, 1:npad) = prev(iz, 1:npad) * zxtaper(1:npad:+1); 
          prev(iz,nx-npad+1:nx) = prev(iz,nx-npad+1:nx) * zxtaper(npad:1:-1)
          cur(iz, 1:npad) = cur(iz, 1:npad) * zxtaper(1:npad:+1);
          cur(iz,nx-npad+1:nx) = cur(iz,nx-npad+1:nx) * zxtaper(npad:1:-1)
          next(iz, 1:npad) = next(iz, 1:npad) * zxtaper(1:npad:+1);
          next(iz,nx-npad+1:nx) = next(iz,nx-npad+1:nx) * zxtaper(npad:1:-1)
       end do

       ! Top and Bottom
       do ix=npad+1,nx-npad
          prev(1:npad,ix) = prev(1:npad,ix) * zxtaper(1:npad:+1);
          prev(nz-npad+1:nz,ix) = prev(nz-npad+1:nz,ix) * zxtaper(npad:1:-1)
          cur(1:npad,ix) = cur(1:npad,ix) * zxtaper(1:npad:+1);
          cur(nz-npad+1:nz,ix) = cur(nz-npad+1:nz,ix) * zxtaper(npad:1:-1)
          next(1:npad,ix) = next(1:npad,ix) * zxtaper(1:npad:+1);
          next(nz-npad+1:nz,ix) = next(nz-npad+1:nz,ix) * zxtaper(npad:1:-1)
       end do



  end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 17: Tapering Setup
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
    subroutine init_taperCos(npad)

    integer                     :: iz, ix,ii, npad
    real                       :: dist
    real                        :: pi=3.14159265359


    allocate(zxtaper(npad), ctaper(npad,npad))

       ! all four edges
       do ii=1,npad
          zxtaper(ii) = 0.985 + 0.015*cos(pi*(npad-ii)/npad)
       end do

       ! Corner
       do ix=1,npad
          do iz=1,npad
              dist = sqrt(1.001*(npad-ix)*(npad-ix)/npad/npad + 0.999999*(npad-iz)*(npad-iz)/npad/npad)
              ctaper(iz,ix) =  0.985 + 0.015*cos(pi*dist)
          end do
       end do

  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 21: Take derivative for 1D array
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

  subroutine FirstDer(adj,souIn,souOut,nt,dt)

    logical :: adj
    integer :: i, nt
    real    :: dt, rdt
    real, dimension(:) :: souIn, souOut
    real, dimension(:), allocatable :: temp

   allocate(temp(nt))
   temp(1:nt) = souIn(1:nt)
    
    souOut(1:nt) = 0.
    
    
    if (.not. adj) then
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1)+ temp(i)
         souOut(i)   = souOut(i) - temp(i)
      end do
      
    else
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1) + temp(i+1)-temp(i)
      end do
         
    end if    

    deallocate(temp)
  end subroutine 



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 23: Take Vertical Derivative of image
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
  subroutine Ver2ndDer(adj,m,nz)
  ! Take the second derivative vertically and its adjoint

  logical                   :: adj
  integer                   :: nz,iz
  real, dimension(:), allocatable :: temp
  real, dimension(:)      :: m


  allocate(temp(nz))

      do iz=2,nz-1
         if (adj) then
            m(iz) = m(iz) + (-2.*temp(iz) + temp(iz+1) + temp(iz-1))
         else
            m(iz) = m(iz) + -2*temp(iz)
            m(iz) = m(iz) + temp(iz+1)
            m(iz) = m(iz) + temp(iz-1)
         end if
       end do


  end subroutine



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE # XX: Apply Laplacian 
!! 1) Fourier transform
!! 2) multiply by -kz**2
!! 3) Inverse fourier transform
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine ApplyLaplac2D(inArr,outArr)

  integer               :: iz, ix
  real, dimension(:,:)    :: inArr, outArr
    real*8             :: val
  !write(0,*) 'Inside Apply Laplac'
  !write(0,*) 'nz, dz', nz, dz


       ! (1) Apply Laplacian 
       do ix=1+nop,nx-nop   
          do iz=nop+1,nz-nop
             val = a0*inArr(iz,ix)
             val = val + a1x*DBLE(inArr(iz,ix+1))
	     val = val + a1x*DBLE(inArr(iz,ix-1))
             val = val + a2x*DBLE(inArr(iz,ix+2))
	     val = val + a2x*DBLE(inArr(iz,ix-2))
             val = val + a3x*DBLE(inArr(iz,ix+3))
	     val = val + a3x*DBLE(inArr(iz,ix-3))
             val = val + a4x*DBLE(inArr(iz,ix+4))
	     val = val + a4x*DBLE(inArr(iz,ix-4))
             val = val + a1z*DBLE(inArr(iz+1,ix))
	     val = val + a1z*DBLE(inArr(iz-1,ix))
             val = val + a2z*DBLE(inArr(iz+2,ix))
	     val = val + a2z*DBLE(inArr(iz-2,ix))
             val = val + a3z*DBLE(inArr(iz+3,ix))
	     val = val + a3z*DBLE(inArr(iz-3,ix))
             val = val + a4z*DBLE(inArr(iz+4,ix))
	     val = val + a4z*DBLE(inArr(iz-4,ix))
	     outArr(iz,ix) = REAL(val)
           end do
        end do



end subroutine


end module


