program testHybrid_2D

  use sep
  use prop_mod2DTX

  use hycdsolver_smp_mod
  use hycdsolver_reg_mod
  use hycdstep_mod
  use identity_mod
  use omp_lib

  implicit none

  integer                                      ::  niter         ! number of iterations
  real                                         :: qntd, qntm     ! quantile
  real                                         :: thresh_d,thresh_m  ! threshold

  logical(4) :: success
  logical                               :: wfldmovie, DW, MW, sder
  real                                  :: dz, oz, ost, dst, ot_data, dt_data, dt_prop
  real                                  :: zsrc, zrec,vmax,tdelay,tapeco,eps
  real                                  :: drx, orx, dsx, osx
  real                                  :: ox, dx, tswmax
  integer                               :: nrx, nsx, nnx, ix, nx
  integer                               :: nnz, nz, nst, nt_data, nop, npad
  integer            			:: iz , it, bout, stat, record_sx
  real, allocatable, dimension(:,:)     :: vel, mwarr
  real, allocatable, dimension(:)       ::dat, DWarray
  real, allocatable, dimension(:)     :: image, resdArr, shotposArr
  real, allocatable, dimension(:)       ::wavelet, mwarr1D, waveletVD
  real, allocatable, dimension(:,:)    :: mmov,rmov
  character(len=128)                    :: souWfile_, imgWfile_, recWfile_
  character(len=128)                    :: outfile_

   ! OMP stuff 
  integer :: Totalnodes

  call sep_init()


  ! read in velocity model
  call from_history("n1", nz);    call from_history("n2", nx)
  call from_history("o1", oz);    call from_history("o2", ox)
  call from_history("d1", dz);    call from_history("d2", dx)

  ! read source wavelet information
  call from_aux("aux", "n1", nst)
  call from_aux("aux", "o1", ost)
  call from_aux("aux", "d1", dst)
  call from_param("tswmax",tswmax,(nt_data-1)*dt_data+ot_data)

  ! read shot gather (data) file info
  call from_aux("datafile", "n1", nt_data);   call from_aux("datafile", "n2", nrx);  call from_aux("datafile","n3",nsx);
  call from_aux("datafile", "o1", ot_data);   call from_aux("datafile", "o2", orx);  call from_aux("datafile","o3",osx);
  call from_aux("datafile", "d1", dt_data);   call from_aux("datafile", "d2", drx);  call from_aux("datafile","d3",dsx);

  write(0,*) 'nt_data, ot, dt', nt_data, ot_data, dt_data
  write(0,*) 'dt_prop tswmax', dt_prop , tswmax

  ! read source location
  call from_param("zsrc",zsrc)
  call from_param("npad",npad,30)
  call from_param("zrec",zrec,zsrc)
  
  ! Read wavefield movie switch
  call from_param("wfldout",wfldmovie,.false.)
  call from_param("DW",DW,.false.);   call from_param("MW",MW,.false.)
  call from_param("souwfld",souWfile_,"souwfld.H")
  call from_param("imgwfld",imgWfile_,"imgwfld.H")  
  call from_param("recwfld",recWfile_,"recwfld.H")
  call from_param("outfile",outfile_)
  call from_param("tdelay",tdelay)
  call from_param("dt_prop",dt_prop,0.001)
  call from_param("tapeco",tapeco,0.00002);
  call from_param("sder", sder, .false.);
  
   ! Parallel Programming
   ! Parallel Programming
   call from_param("nodes" ,Totalnodes, 8)
   !Totalnodes =min(omp_get_num_procs(), 16)
   Totalnodes =min(Totalnodes, 16)
   call omp_set_num_threads(Totalnodes); 
    write(0,*) 'OMP: using', Totalnodes, 'out of', omp_get_num_procs() 

  if (dst .ne. dt_data) call erexit('dst should equal to dt_data')
  if (dt_prop .ne. dt_data) call erexit('expect dt_prop .eq. dt_data for now')


  nnz=nz
  nz = nnz+2*npad
  nnx=nx
  nx = nnx+2*npad
  record_sx=1.

  ! Allocate memory
  allocate(vel(nz,nx),image(nnz*nnx),wavelet(nt_data),waveletVD(nt_data))
  allocate(dat(nt_data*nrx*nsx))
  allocate(DWArray(nt_data*nrx*nsx))
  allocate(shotposArr(nsx))


  vel=0.
  image=0.
  wavelet=0.
  waveletVD=0.
  dat=0.
  resdArr=0.

  ! Read Velocity Model
  call sep_read(vel(npad+1:nz-npad,npad+1:nx-npad));  call from_param("vmax",vmax);

  ! Padding 
  do iz=1,npad
     vel(iz,:) = vel(npad+1,:)
     vel(nz-iz+1,:) = vel(nz-npad-1,:)
  end do
  do ix=1,npad
     vel(:,ix) = vel(:,npad+1)
     vel(:,nx-ix+1) = vel(:,nx-npad-1)
  end do


  ! Read source wavelet
  call sep_read(wavelet,"aux")
  call Ver2ndDer(.true.,.false.,wavelet,waveletVD,nt_data)
  if (DW) call sep_read(DWArray,'DWfile')  
  
  ! Read data
  call sep_read(dat,'datafile')
  call sep_read(shotposArr,'shotposfile')

  call from_param("outfile",outfile_,'outfile.H')

  call from_param ("niter", niter, 2)
  call from_param ('thresh_d',thresh_d, 1.)
  call from_param ('thresh_m',thresh_m, 1.)
  call from_param ('eps',eps, 1.)
  write(0,*) "niter   =",niter
  write(0,*) "thresh_d=",thresh_d
  write(0,*) "thresh_m=",thresh_m
  write(0,*) "eps=",eps

  allocate(mmov(nnz*nnx,niter))
  allocate(rmov(size(dat),niter))
 
  ! Initiate propagation
  call init_propagate(vel,oz,dz,ox,dx,dst,dt_prop,nt_data,dt_data,ot_data, tdelay,vmax,npad,tapeco, wfldmovie,record_sx,souWfile_, recWfile_, imgWfile_,Totalnodes)
  ! Initiate source parameters
  if (sder) then
     call sou_rec_init(waveletVD,zsrc,nsx,dsx,osx,zrec,nrx,orx,drx,tdelay,tswmax,shotposArr)
  else
    call sou_rec_init(wavelet,zsrc,nsx,dsx,osx,zrec,nrx,orx,drx,tdelay,tswmax,shotposArr)
  end if
  
  
  ! Set threshold values for hycd data weighting
  call identity_init(thresh_d,thresh_m)

  if (DW) then
    ! Initiate data weighting
    call init_datwgt(DWArray=DWArray,thresh_d=thresh_d)  
    call hycdsolver_reg( m=image,d=dat, Fop=born2way_op , Aop=identity_lop, Wop=datawgt_op , Wmop=identitym_lop, stepper=hycdstep, nAop=nnz*nnx, niter=niter, eps=eps, resd=resdArr, mmov=mmov, rmov=rmov, verb=.true.)
  else
    call hycdsolver_reg( m=image,d=dat, Fop=born2way_op , Aop=identity_lop, Wop=identityd_lop , Wmop=identitym_lop, stepper=hycdstep, nAop=nnz*nnx, niter=niter, eps=eps, resd=resdArr, mmov=mmov, verb=.true.)   
  end if
  
   call close_propagate()
   call sou_rec_close()




     call sep_write(image,outfile_)

    call auxputch("n1",'i',nnz,outfile_)
    call auxputch("n2",'i',nnx,outfile_)
    call auxputch("d1",'f',dz,outfile_)
    call auxputch("d2",'f',dx,outfile_)
    call auxputch("o1",'f',oz,outfile_)
    call auxputch("o2",'f',ox,outfile_)
    call auxputch("label1",'s','z(m)',outfile_)
    call auxputch("label2",'s','x(m)',outfile_)

    call auxputch("n1",'i',nnz,'mmov')
    call auxputch("n2",'i',nnx,'mmov')
    call auxputch("d1",'f',dz,'mmov')
    call auxputch("d2",'f',dx,'mmov')
    call auxputch("o1",'f',oz,'mmov')
    call auxputch("o2",'f',ox,'mmov')
    call auxputch("label1",'s','z(m)','mmov')
    call auxputch("label2",'s','x(m)','mmov')
    call auxputch("label3",'s','niter','mmov')
    call auxputch("n3",'i',niter,'mmov')
    call auxputch("n1",'i',nt_data,'rmov')
    call auxputch("n2",'i',nrx,'rmov')
    call auxputch("n3",'i',nsx,'rmov')
    call auxputch("n4",'i',niter,'rmov')
    call srite('mmov',mmov,4*nnz*nnx*niter)
    call srite('rmov',rmov,4*size(dat)*niter)

end program
