module waveprop_mod

  use sep
  use type_mod
	

  implicit none

  
contains

!!--------------------------------------------------------------

  subroutine Gaussian1D(adj,m,nz,nx)
  ! Take a 1D Gaussian and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, ix, iz, nop,wd, i, j
  real                      :: dx, dz, width, val, norm 
  real, dimension(:), allocatable :: gauss
  real, dimension(:,:)      :: m
  real, dimension(:,:), allocatable  :: d


  wd = 2
  norm = 0.0
  width = 0.0
  val = 0.0

  allocate(d(nz,nx))
  allocate(gauss(2*wd+1))  

  do ix=1,nx
     do iz=1,nz
        d=0.
     end do
  end do
 
  write(0,*) 'Constructing filters'

  do i=1,2*wd+1
       val = real(i-(wd+1))
       width = real(wd+1)
       gauss(i) = exp(-(3.0*val/width)*(3.0*val/width)/2.0)
       norm = norm + gauss(i)
  end do

  gauss = gauss/norm
        write(0,*) 'gauss filter',  gauss
        write(0,*) 'sum of filter', sum(gauss)
        
      

  write(0,*) 'Applying filter'

  if (.not. adj) then
    
    do iz=1+wd, nz-wd
       do j=-wd,wd
          d(iz,:) = d(iz,:) + gauss(j+wd+1)*m(iz+j,:)       
       end do

 
    enddo


  else

     do iz=1+wd, nz-wd
       do j=-wd,wd
          d(iz+j,:) = d(iz+j,:) + gauss(j+wd+1)*m(iz,:)   
       end do
    enddo

  end if

    write(0,*) 'm, d', m(iz/6,ix/2), d(iz/6,ix/2)


    do ix=1,nx
     do iz=1,nz
        m(iz,ix)=m(iz,ix) - d(iz,ix)
     end do
    end do

  write(0,*) 'outputting Gaussian', sum(m)

  end subroutine

!!--------------------------------------------------------------
!!--------------------------------------------------------------

  subroutine Gaussian2D(adj,m,nz,nx)
  ! Take a 1D Gaussian and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, ix, iz, nop,wd, i, j, k, l
  real                      :: dx, dz, width2, val, val2, norm, a, kk 
  real, dimension(:,:), allocatable :: gauss
  real, dimension(:,:)      :: m
  real, dimension(:,:), allocatable  :: d


  wd = 2
  norm = 0.0
  width2 = 0.0
  val = 0.0

  allocate(d(nz,nx))
  allocate(gauss(wd+1,wd+1))  

  do ix=1,nx
     do iz=1,nz
        d=0.
     end do
  end do
 
  write(0,*) 'Constructing filters'
  ! filter is only a quarter of the entire coverage because it is symmetric about zero
  ! the center of the filter is gauss(1,1) (largest amp), and the smallest value is gauss(wd+1,wd+1)
  a = 5.
  width2 = 2.0*real(wd+1)**2
  kk = a*a/width2

  do j=1,wd+1
    do i=1,wd+1
       val = real(i-1)
       val2 = real(j-1)
       gauss(i,j) = exp(-kk*val*val-kk*val2*val2)
       norm = norm + gauss(i,j)
    end do
  end do

  gauss = gauss/norm/4.0
        write(0,*) 'gauss filter',  gauss


  write(0,*) 'Applying filter'

  if (.not. adj) then
   do ix=1+wd, nx-wd  
     do iz=1+wd, nz-wd
       do k=0,wd
         do l=0,wd
          d(iz,ix) = d(iz,ix) + gauss(k,l)*(m(iz+k,ix+l) + m(iz-k,ix+l) + m(iz+k,ix-l) + m(iz-k,ix-l))      
         end do
       end do
     enddo
   end do


  else

   do ix=1+wd, nx-wd  
     do iz=1+wd, nz-wd
       do k=0,wd
         do l=0,wd
           d(iz+k,ix+l) = d(iz+k,ix+l) + gauss(k,l)*m(iz,ix)
           d(iz-k,ix+l) = d(iz-k,ix+l) + gauss(k,l)*m(iz,ix)
           d(iz+k,ix-l) = d(iz+k,ix-l) + gauss(k,l)*m(iz,ix)
           d(iz-k,ix-l) = d(iz-k,ix-l) + gauss(k,l)*m(iz,ix)     
         end do
       end do
     enddo
   end do
  end if

    write(0,*) 'm, d', m(iz/6,ix/2), d(iz/6,ix/2)


    do ix=1,nx
     do iz=1,nz
        m(iz,ix)=m(iz,ix) - d(iz,ix)
     end do
    end do

  write(0,*) 'outputting Gaussian', sum(m)

  end subroutine
!!--------------------------------------------------------------


!!--------------------------------------------------------------

  subroutine Laplacian2D(adj,m,nz,nx)
  ! Take a 2D laplacian and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, ix, iz, nop
  real, dimension(:,:)      :: m
  real, dimension(:,:), allocatable  :: d


  allocate(d(nz,nx))
  
  do ix=1,nx
     do iz=1,nz
        d=0.
     end do
  end do
 
  nop=1 

  if (.not. adj) then
    do ix=1+nop,nx-nop
       do iz=1+nop, nz-nop

          d(iz,ix) = d(iz,ix) + 8.*m(iz,ix) - m(iz+1,ix) - m(iz-1,ix) - &
                     m(iz,ix+1) - m(iz,ix-1) - m(iz+1,ix+1) -&
                     m(iz+1,ix-1) - m(iz-1,ix+1) -m(iz-1,ix-1)

       enddo
    enddo


  else


    do ix=1+nop,nx-nop
       do iz=1+nop, nz-nop
          d(iz,ix) = d(iz,ix) + 8.*m(iz,ix)
          d(iz+1,ix) = d(iz+1,ix) + -1.*m(iz,ix)
	  d(iz-1,ix) = d(iz-1,ix) + -1.*m(iz,ix)
          d(iz,ix+1) = d(iz,ix+1) + -1.*m(iz,ix)
          d(iz,ix-1) = d(iz,ix-1) + -1.*m(iz,ix)
          d(iz+1,ix+1) = d(iz+1,ix+1) + -1.*m(iz,ix)
          d(iz+1,ix-1) = d(iz+1,ix-1) + -1.*m(iz,ix)
          d(iz-1,ix+1) = d(iz-1,ix+1) + -1.*m(iz,ix)
          d(iz-1,ix-1) = d(iz-1,ix-1) + -1.*m(iz,ix)
       enddo
    enddo


  end if

    do ix=1,nx
     do iz=1,nz
        m(iz,ix)=d(iz,ix)
     end do
    end do

  write(0,*) 'Laplace v2', adj

  end subroutine

!!--------------------------------------------------------------
!! Weighting matrix setup for Wmodel

  subroutine WmodeltaperSetup(tapebd,wd,Wup,nz,nx, cent, A)
  ! tapebd = center of tapering zone
  ! wd = width of taper zone
  
    integer                     :: iz, ix, nz, nx, dist, iztp, wd
    integer                     :: izb, ize
    integer, dimension(:)          :: tapebd
    real, dimension(:,:)        :: Wup
    real                        :: cent,A,pi=3.14159265359
    real                        :: ztapco, amp

    ztapco = 1.

   write(0,*) 'nx,nz',nx,nz
   ! Make wd even integer
   if (mod(wd,2) .ne. 0) then
      wd = wd+1
   end if
  
  
   write(0,*) 'hello world 1', wd, tapebd(nx/2)    

    
     do ix=1,nx

        iztp = tapebd(ix)
        izb = iztp-wd/2
        ize = izb + wd
!        if (izb .lt. 1) then
!            izb=1
!            ize= izb+wd
!        else if (ize .gt. nz) then
!            ize= nz
!            izb= ize-wd
!        end if

!        write(0,*) 'ix, izb, ize', ix, izb, ize
               
        do iz=max(izb,1),min(ize,nz)
            dist = wd-(izb-iz)
           ! Cosine Taper
            Wup(iz,ix) = (cent + A*cos(pi*(dist)/wd))
!            write(0,*) Wup(iz,ix), iz, ix
 
        end do
           amp = Wup(izb,ix)
           Wup(1:izb-1,ix) = amp
           amp = Wup(ize,ix)
           Wup(ize+1:nz,ix) = amp
!	   Wup(ize+1:nz,ix) = 0.5
     end do
           


  end subroutine

!! ---------------------------------------------------------
!! Random seed Initialize

   subroutine init_random_seed(seed_in)
      
     integer               :: i, n, clock, seed_in
     integer, dimension(:), allocatable    :: seed

     n=1
     call RANDOM_SEED(size=n)
     allocate(seed(n))

     call SYSTEM_CLOCK(COUNT=clock)
 
     seed = seed_in + 37*(/(i-1,i=1,n)/)
     call RANDOM_SEED(PUT = seed)

     DEALLOCATE(seed)


   end subroutine


!!----------------------------------------------------------
!! Random Boundary Condition Setup
!! Put random number around the outer boundary region
!! bndx, bndz = width of outer boundary in x & z direction
!! nx,nz = total length of the grid (include outer boundary)
!! vp2 = velocity grid (size nz by nx)
!! maxv = bound on max velocity to ensure stable calculation

   subroutine rbcsetup(ishot, v2,Ran1, Ran2, bndx,bndz,nx,nz,seed, maxv)

      logical                :: found
      integer                :: ix, iz, ishot
      integer                :: bndx, bndz, nx, nz, seed, seed_it(2)
      real, dimension(:,:)   :: v2
      real, dimension(:,:,:) :: Ran1, Ran2
      real                   :: maxv, bx, bz, r, d, tempV, minV


      bx = (real(bndx))
      bz = (real(bndz))

      write(0,*) 'Random Boundary Setup'

      call init_random_seed(seed)

      write(0,*) 'after seed'



      do ix=1,nx

 
         do iz=1,nz
            found=.false.
! **************************
            minV = 2000.
! **************************
            do while(.not. found)

!if (.not. found) write(0,*) 'first thing', ishot, iz, ix

               ! Draw a random number
               call random_number(r)
               r = r-.5
               r=r
               
               ! Calculate distance to the inner boundary
               if (ix .le. bndx) then
                  if (iz .lt. bndz) then   ! Corner position
                     d = sqrt(1.*((ix-bndx)*(ix-bndx)/bx + (iz-bndz)*(iz-bndz)/bz ) )
	          else if (iz .gt. (nz-bndx)) then  ! corner position
                     d = sqrt(1.*((ix-bndx)*(ix-bndx)/bx + (iz-(nz-bndz))*(iz-(nz-bndz))/bz))
                  else ! edge
                     d = (real(bndx-ix)/bndx)
                  end if


                  tempV = v2(iz,ix) + r*min(minV**2*d,maxv*maxv)
                  Ran1(iz,ix,ishot) = tempV

! write(0,*) 'RAN 1', ishot, iz, ix

               else if (ix .gt. (nx-bndx)) then
                  if (iz .lt. bndz) then   ! Corner position
                     d = sqrt(1.*((ix-(nx-bndx))*(ix-(ix-bndx))/bx + (iz-bndz)*(iz-bndz)/bz ) )
	          else if (iz .gt. (nz-bndx)) then  ! corner position
                     d = sqrt(1.*((ix-(nx-bndx))*(ix-(nx-bndx))/bx + (iz-(nz-bndz))*(iz-(nz-bndz))/bz))
                  else ! edge
                     d = (real((nx-bndx)-ix)/bndx)
                  end if

                  tempV = v2(iz,ix) + r*min(minV*2*d,maxv*maxv)
		  Ran1(iz, ix-nx+2*bndx,ishot) = tempV

! write(0,*) 'RAN 2', ishot, iz, ix

               else if (iz .le. bndz) then ! edge

                     d = real(bndz-iz)/bndz

                  tempV = v2(iz,ix) + r*min(minV**2*d,maxv*maxv)

		  Ran2(iz, ix-bndx, ishot) = tempV

! write(0,*) 'RAN 3', ishot, iz, ix

               else if (iz .gt. (nz-bndz)) then  ! edge

                     d = real(iz-nz+bndz)/bndz

                  tempV = v2(iz,ix) + r*min(minV**2*d,maxv*maxv)

		  Ran2(iz-nz+2*bndz, ix-bndx, ishot) = tempV


! write(0,*) 'RAN 4', ishot, iz, ix

               else

		  tempV = 18.e6

! write(0,*) 'RAN 5', ishot, iz, ix
               end if
               

              if ((tempV .lt. 40.) .or.  (tempV .gt. 36.e6)) then
	         found=.false. 
                 write(0,*) 'v2, tmpV', ishot, iz, ix
                 write(0,*) 'found false', v2(iz,ix), tempV
              else
                 found=.true.
              end if

              
if (.not. found) write(0,*) 'last thing before enddo'
            enddo
         end do
       end do   
           

write(0,*) 'end of RAN setup'

   end subroutine

!!----------------------------------------------------------
!! take 2nd derivative of the wavelet

  subroutine SecDer(souIn,souOut,nt,dt)

    integer :: it, nt
    real    :: dt
    real, dimension(:) :: souIn, souOut
    real, dimension(:), allocatable :: temp

    allocate(temp(nt))

    temp = souIn

    do it=1+1,nt-1
!       write(0,*) 'it', it
       souOut(it) = (-2.*temp(it) + temp(it+1) + temp(it-1))
    end do

    souOut= souOut/dt**2
    
    ! clip the edge to be zero
    souOut(1:2) = 0.;
    souOUt(nt-2:nt) = 0.; 

    deallocate(temp)


  end subroutine 


!!----------------------------------------------------------
!! take 1st derivative of the  wavelet

  subroutine FirstDer(adj,souIn,souOut,nt,dt)

    logical :: adj
    integer :: i, nt
    real    :: dt, rdt
    real, dimension(:) :: souIn, souOut
    real, dimension(:), allocatable :: temp

   allocate(temp(nt))
   temp(1:nt) = souIn(1:nt)
    
    souOut(1:nt) = 0.
    
    
    if (.not. adj) then
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1)+ temp(i)
         souOut(i)   = souOut(i) - temp(i)
      end do
      
    else
       
      do i=1,nt-1
         souOut(i) = souOut(i) + temp(i+1)-temp(i)
      end do
         
    end if    

    deallocate(temp)


  end subroutine 
  
!!---------------------------------------------------------- 
!! one-way absorbing BC setup

  subroutine abcsetup(nx,nz,vp,dt,dx, dz, nop, bndz0,bndz1,bndx0,bndx1,sqed)

    logical  :: sqed
    integer  :: ix, iz, nx, nz, nop
    real, dimension(:) :: bndz0,bndz1,bndx0,bndx1
    real     :: delp, dt,dx, dz
    real,  dimension(:,:) :: vp


    if (sqed) then
       do ix=1,nx
        
          delp = sqrt(vp(1+nop,ix)) *dt/dz

         ! Want absorbtive top boundary
          bndz0(ix) = (1.-delp)/(1.+delp)

          delp = sqrt(vp(nz-nop,ix)) *dt/dz
          bndz1(ix) = (1.-delp)/(1.+delp)
       end do

       do iz=1,nz
          delp = sqrt(vp(iz,1+nop)) *dt/dx
          bndx0(iz) = (1.-delp)/(1.+delp)

          delp = sqrt(vp(iz,nx-nop)) *dt/dx
          bndx1(iz) = (1.-delp)/(1.+delp)
       end do
    else
       do ix=1,nx
          delp = vp(1+nop,ix) *dt/dz

         ! Want absorbtive top boundary
          bndz0(ix) = (1.-delp)/(1.+delp)

          delp = vp(nz-nop,ix) *dt/dz
          bndz1(ix) = (1.-delp)/(1.+delp)
       end do

       do iz=1,nz
          delp = vp(iz,1+nop) *dt/dx
          bndx0(iz) = (1.-delp)/(1.+delp)

          delp = vp(iz,nx-nop) *dt/dx
          bndx1(iz) = (1.-delp)/(1.+delp)
       end do

    end if

  end subroutine 

!!---------------------------------------------------------- 
!! one-way absorbing BC setup
!! Constant velocity

  subroutine abcsetupV2(nx,nz,wvel,dt,dx, dz, nop, bndz0,bndz1,bndx0,bndx1)

    integer  :: ix, iz, nx, nz, nop
    real, dimension(:) :: bndz0,bndz1,bndx0,bndx1
    real     :: delp, dt,dx, dz, wvel   

       delp = wvel*dt/dz
       do ix=1,nx
         ! Want absorbtive top boundary
          bndz0(ix) = (1.-delp)/(1.+delp)
          bndz1(ix) = (1.-delp)/(1.+delp)
       end do

       delp = wvel*dt/dx
       do iz=1,nz
          bndx0(iz) = (1.-delp)/(1.+delp)
          bndx1(iz) = (1.-delp)/(1.+delp)
       end do


  end subroutine 

!!---------------------------------------------------------- 

!! Absorbing BC 1 - absorbtive top boundary (used to propagate the source wavefield)

  subroutine absorbingBC1(nz,nx,nop,bndz0,bndz1,bndx0,bndx1,uo,um)

    integer  :: iop, nx, nz, nop
    real, dimension(:) :: bndz0,bndz1,bndx0,bndx1
    real, dimension(:,:) :: um,uo

       !! absorbing BC
       do iop=1,nop
          uo( 1+nop-iop  ,:) = &
               um( 1+nop-iop+1,:) + bndz0 * (um( 1+nop-iop,:)-uo( 1+nop-iop+1,:))

          uo(nz-nop+iop  ,:) = &
               um(nz-nop+iop-1,:) + bndz1 * (um(nz-nop+iop,:)-uo(nz-nop+iop-1,:))

          uo(:, 1+nop-iop  ) = &
               um(:, 1+nop-iop+1) + bndx0 * (um(:, 1+nop-iop)-uo(:, 1+nop-iop+1))

          uo(:,nx-nop+iop  ) = &
               um(:,nx-nop+iop-1) + bndx1 * (um(:,nx-nop+iop)-uo(:,nx-nop+iop-1))
       end do


  end subroutine


!!---------------------------------------------------------- 
!! Absorbing BC 2 - Reflective top boundary (used to image the mirror signal)
!!                - assumes padding at the top

  subroutine absorbingBC2(nz,nx,nop,bndz0,bndz1,bndx0,bndx1,uo,um)

    integer  :: iop, nx, nz, nop
    real, dimension(:) :: bndz0,bndz1,bndx0,bndx1
    real, dimension(:,:) :: um,uo

  !  write(0,*) 'Calling BC2'

       !! absorbing BC
       do iop=1,nop
          uo( 1+nop-iop  ,:) = &
               um( 1+nop-iop+1,:) + 1. * (um( 1+nop-iop,:)-uo( 1+nop-iop+1,:))

          uo(nz-nop+iop  ,:) = &
               um(nz-nop+iop-1,:) + bndz1 * (um(nz-nop+iop,:)-uo(nz-nop+iop-1,:))

          uo(:, 1+nop-iop  ) = &
               um(:, 1+nop-iop+1) + bndx0 * (um(:, 1+nop-iop)-uo(:, 1+nop-iop+1))

          uo(:,nx-nop+iop  ) = &
               um(:,nx-nop+iop-1) + bndx1 * (um(:,nx-nop+iop)-uo(:,nx-nop+iop-1))
       end do

  
  end subroutine

!!-------------------------------------------------------------
!!---------------------------------------------------------- 
!! Absorbing BC  3 - Reflective top boundary (used to image the mirror signal)
!!                 - assumes no padding on the top                  

  subroutine absorbingBC3(nz,nx,nop,bndz0,bndz1,bndx0,bndx1,uo,um)

    integer  :: iop, nx, nz, nop
    real, dimension(:) :: bndz0,bndz1,bndx0,bndx1
    real, dimension(:,:) :: um,uo

  !  write(0,*) 'Calling BC2'

       !! absorbing BC
       do iop=1,nop
!          uo( 1+nop-iop  ,:) = &
!               um( 1+nop-iop+1,:) + 1. * (um( 1+nop-iop,:)-uo( 1+nop-iop+1,:))

          uo(nz-nop+iop  ,:) = &
               um(nz-nop+iop-1,:) + bndz1 * (um(nz-nop+iop,:)-uo(nz-nop+iop-1,:))

          uo(:, 1+nop-iop  ) = &
               um(:, 1+nop-iop+1) + bndx0 * (um(:, 1+nop-iop)-uo(:, 1+nop-iop+1))

          uo(:,nx-nop+iop  ) = &
               um(:,nx-nop+iop-1) + bndx1 * (um(:,nx-nop+iop)-uo(:,nx-nop+iop-1))
       end do

  
  end subroutine


!!--------------------------------------------------------------

  subroutine absorbingBCHale(adj, vel_grid,dt, vp2, um,uo,up)
  
     type(velo)    :: vel_grid
     logical           :: adj
     integer           :: nx, nz, iz, ix
     real              :: dt, dx, dz
     real              :: ovs, ov, dpdx, dpdxs, dpdt, dpdts, cosa, beta, gamma
     real, dimension(:,:)   :: vp2, uo, um, up
     
     nx = vel_grid%x%n
     nz = vel_grid%z%n
     dx = vel_grid%x%d
     dz = vel_grid%z%d


    !! Top boundary     
    iz=3
    do ix=1,nx
    
       ovs = 1.0/vp2(iz,ix);
       ov  = sqrt(ovs);
       
       if (ix == 1) then
       
          dpdx = (uo(iz,2) - uo(iz, 1))/dx
       else if (ix == nx) then
          dpdx = (uo(iz,nx) - uo(iz,nx-1))/dx
       else 
          dpdx = (uo(iz,ix+1) - uo(iz,ix-1))/(2.*dx)
       end if
       
       dpdt = (up(iz,ix) - um(iz,ix))/(2.*dt)
       dpdxs = dpdx*dpdx
       dpdts = dpdt*dpdt
       
       if (ovs*dpdts .gt. dpdxs) then
          cosa=sqrt(1.0-dpdxs/(ovs*dpdts));
       else
          cosa = 0.0;
       end if
       
       beta = ov*cosa*dz/dt
       gamma = (1.0-beta)/(1.0+beta);
      
       up(iz,ix) = gamma*(up(iz+1,ix)-uo(iz,ix))+uo(iz+1,ix)
       
       if (ix .eq. nx/2) then
          write(0,*) ovs, dpdt, dpdx, sum(uo)  
      !   write(0,*) 'gamma', cosa, gamma       
       end if
       
        
    end do 

      
    !! Bottom boundary
    iz=nz-1
    do ix=1,nx
    
       ovs = 1.0/vp2(iz,ix);
       ov  = sqrt(ovs);
       
       if (ix == 1) then
       
          dpdx = (uo(iz,ix+1) - uo(iz, ix))/dx
       else if (ix == nx) then
          dpdx = (uo(iz,ix) - uo(iz,ix-1))/dx
       else 
          dpdx = (uo(iz,ix+1) - uo(iz,ix-1))/(2.*dx)
       end if
       
       dpdt = (up(iz,ix) - um(iz,ix))/(2.*dt)
       dpdxs = dpdx*dpdx
       dpdts = dpdt*dpdt
       
       if (ovs*dpdts .gt. dpdxs) then
          cosa=sqrt(1.0-dpdxs/(ovs*dpdts));
       else
          cosa = 0.0;
       end if
       
       beta = ov*cosa*dz/dt
       gamma = (1.0-beta)/(1.0+beta);
       
       up(iz+1,ix) = gamma*(up(iz,ix)-uo(iz+1,ix))+uo(iz,ix)
       
    end do   
     
  
  
  end subroutine
  
!!--------------------------------------------------------------

  subroutine ashotsetup(izsou,shotdir,nx,vp,dt,dz,nop,bndz2)
    integer  :: ix, izsou, nx, nop, shotdir
    real, dimension(:) :: bndz2
    real     :: delp, dt, dz
    real,  dimension(:,:) :: vp

	write(0,*) 'ashot setup'

      
    do ix=1,nx
       delp = vp(izsou+1+nop,ix) *dt/dz
       bndz2(ix) = (1.-delp)/(1.+delp)
       
    end do

  end subroutine 

!!--------------------------------------------------------------

  subroutine ashotcondition(izsou,shotdir,nop,bndz2,uo,um)

    integer  :: izsou, iop, nop, shotdir
    real, dimension(:) :: bndz2
    real, dimension(:,:) :: um,uo

	

       !! absorbing shot
    do iop=1,nop
       if (shotdir .eq. 1) then
          uo( izsou+1 +nop-iop  ,:) = um( izsou+1+nop-iop+1,:) + &
               bndz2 * (um( izsou+1+nop-iop,:)-uo( izsou+1+nop-iop+1,:))
       else
          uo(izsou-nop+iop  ,:) = &
               um(izsou-nop+iop-1,:) + bndz2 * (um(izsou-nop+iop,:)-uo(izsou-nop+iop-1,:))

       end if

    end do
  end subroutine

!!--------------------------------------------------------------       
!! Taper Setup Gaussian

  subroutine tapersetup(bndz,bndx,ztapco,xtapco,ztaper,xtaper)

    integer  :: iz, ix, bndz, bndx
    real, dimension(:) :: xtaper,ztaper
    real                        :: xtapco,ztapco


       do iz=1,bndz
          ztaper(iz) = exp(-ztapco * iz*iz)
       end do

       do ix=1,bndx
          xtaper(ix) = exp(-xtapco * ix*ix)
       end do

  end subroutine

!!--------------------------------------------------------------
!! Taper Setup Cosine

  subroutine tapersetupCos(bndz,bndx,ztaper,xtaper, ctaper)

    integer                     :: iz, ix, bndz, bndx
    real, dimension(:)          :: xtaper,ztaper
    real, dimension(:,:)        :: ctaper
    real                       :: dmax, dist
    real                        :: xtapco,ztapco
    real                        :: pi=3.14159265359


     dmax = max(bndx, bndz)
!    dmax = sqrt(pi*bndx*bndx + bndz*bndx)
    


       ! Left and right edge
       do iz=1,bndz
          ztaper(iz) = 0.98 + 0.02*cos(pi*(bndz-iz)/bndz)
       end do

       ! top and bottom edge
       do ix=1,bndx
          xtaper(ix) = 0.98 + 0.02*cos(pi*(bndx-ix)/bndx)
       end do


       ! Corner
       do ix=1,bndx
          do iz=1,bndz
              dist = sqrt(1.001*(bndx-ix)*(bndx-ix)/bndx/bndx + 0.999999*(bndz-iz)*(bndz-iz)/bndz/bndz)
              ctaper(iz,ix) =  0.97 + 0.03*cos(pi*dist)
!	 	ctaper(iz,ix) = xtaper(ix)*ztaper(iz)
          end do
       end do

  end subroutine

!!----------------------------------------------------------------
!!--------------------------------------------------------------
!! Taper Setup Cosine

  subroutine tapersetupCos2(bndz,bndx,ztaper,xtaper, ctaper)

    integer                     :: iz, ix, bndz, bndx
    real, dimension(:)          :: xtaper,ztaper
    real, dimension(:,:)        :: ctaper
    real                       :: dmax, dist
    real                        :: xtapco,ztapco
    real                        :: pi=3.14159265359


     dmax = max(bndx, bndz)
!    dmax = sqrt(pi*bndx*bndx + bndz*bndx)
    


       ! Left and right edge
       do iz=1,bndz
          ztaper(iz) = 0.6 + 0.4*cos(pi*(bndz-iz)/bndz)
       end do

       ! top and bottom edge
       do ix=1,bndx
          xtaper(ix) = 0.6 + 0.4*cos(pi*(bndx-ix)/bndx)
       end do


       ! Corner
       do ix=1,bndx
          do iz=1,bndz
              dist = sqrt(1.001*(bndx-ix)*(bndx-ix)/bndx/bndx + 0.999999*(bndz-iz)*(bndz-iz)/bndz/bndz)
              ctaper(iz,ix) =  0.6 + 0.4*cos(pi*dist)
!	 	ctaper(iz,ix) = xtaper(ix)*ztaper(iz)
          end do
       end do

  end subroutine

!!----------------------------------------------------------------
!! Tapering
!! Version 2 has reflective top boundary 
!! no padding is applied to the top boundary
!! nz = nnz+1*bndz + nop

  subroutine taperingCosV2(nz,nx,bndz,bndx,ztaper,xtaper,ctaper,uo)

    integer  :: iz, ix, nz,nx,bndz,bndx
    real, dimension(:) :: xtaper,ztaper
    real, dimension(:,:) :: uo, ctaper

       ! Corner
       do ix=1,bndx
          do iz=1,bndz
              uo(nz-iz+1,nx-ix+1) = uo(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              uo(nz-iz+1,ix) = uo(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left
          end do
       end do

       do iz=1,nz-bndz
          uo(iz, 1:bndx) = uo(iz, 1:bndx) * xtaper(1:bndx:+1);          ! left
	  uo(iz,nx-bndx+1:nx) = uo(iz,nx-bndx+1:nx) * xtaper(bndx:1:-1) ! right
       end do

       do ix=1,nx-bndx
          uo(nz-bndz+1:nz,ix) = uo(nz-bndz+1:nz,ix) * ztaper(bndz:1:-1) ! bottom
       end do

  end subroutine

!!----------------------------------------------------------------
!! Tapering


  subroutine taperingCos(nz,nx,bndz,bndx,ztaper,xtaper,ctaper,uo)

    integer  :: iz, ix, nz,nx,bndz,bndx
    real, dimension(:) :: xtaper,ztaper
    real, dimension(:,:) :: uo, ctaper

       ! Corner
       do ix=1,bndx
          do iz=1,bndz

              uo(iz,ix) = uo(iz,ix)*ctaper(iz,ix);         ! top left
              uo(iz,nx-ix+1) = uo(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
              uo(nz-iz+1,nx-ix+1) = uo(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              uo(nz-iz+1,ix) = uo(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left
          end do
       end do

       do iz=bndz+1,nz-bndz

          uo(iz, 1:bndx) = uo(iz, 1:bndx) * xtaper(1:bndx:+1); 
	  uo(iz,nx-bndx+1:nx) = uo(iz,nx-bndx+1:nx) * xtaper(bndx:1:-1)
       end do

       do ix=bndx+1,nx-bndx

          uo(1:bndz,ix) = uo(1:bndz,ix) * ztaper(1:bndz:+1);             ! top
          uo(nz-bndz+1:nz,ix) = uo(nz-bndz+1:nz,ix) * ztaper(bndz:1:-1)  ! bottom
       end do

  end subroutine


!!--------------------------------------------------------------------------
!! 2nd Order Laplacian

  subroutine laplac2(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz
    real, dimension(:,:) :: uo,up, um, vp2
    real       :: idx, idz, dt, dt2, sum, xscale, zscale

    dt2 = dt**2
    xscale = dt2*idx
    zscale = dt2*idz
    sum = 2.*(xscale+zscale)

    if (adj) then    !! RTM
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=2,nx-1       
          do iz=2,nz-1
          
             up(iz,ix) = 2.*uo(iz,ix) - um(iz,ix) + &
                  vp2(iz,ix)*((uo(iz  ,ix-1) + uo(iz  ,ix+1))*xscale  + &                 
                   (uo(iz-1,ix  ) + uo(iz+1,ix  ))*zscale -uo(iz,ix)*sum) 

          end do
       end do
    !$OMP end parallel do   
    else 
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=2,nx-1
          do iz=2,nz-1         

             up(iz,ix) = 2*uo(iz,ix) - um(iz,ix) + &
                         vp2(iz,ix)*((uo(iz,ix-1)+uo(iz,ix+1)-2.*uo(iz,ix))*xscale + &
                                     (uo(iz-1,ix)+uo(iz+1,ix)-2.*uo(iz,ix))*zscale )
                                     
!            up(iz,ix) = 2*uo(iz,ix) - um(iz,ix) + &
!               ((vp2(iz,ix-1)*uo(iz  ,ix-1) + vp2(iz,ix+1)*uo(iz  ,ix+1) - 2.*vp2(iz,ix)*uo(iz,ix))*xscale + &
!                (vp2(iz-1,ix)*uo(iz-1,ix  ) + vp2(iz+1,ix)*uo(iz+1,ix  ) - 2.*vp2(iz,ix)*uo(iz,ix))*zscale)

                
          end do
        end do
   !$OMP end parallel do
    end if

  end subroutine


!!--------------------------------------------------------------------------

!!--------------------------------------------------------------------------
!! 4th Order Laplacian
!! source has been scaled by density 'ro'


  subroutine laplac4(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz
    real, dimension(:,:) :: uo,up, um, vp2
    real       :: c0, c1, c2, idx, idz, dt, dt2
    real       :: xscale, zscale, sum

    c0 = -30./12.
    c1 = +16./12.
    c2 = - 1./12.
    dt2 = dt**2
    xscale = dt2*idx
    zscale = dt2*idz
    sum    = xscale+zscale

    

    if (adj) then    !! RTM
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=3,nx-2       
          do iz=3,nz-2
          

             up(iz,ix) = vp2(iz,ix)* &
                  (c1*(uo(iz  ,ix-1) + uo(iz  ,ix+1))*xscale + &
                  c2*(uo(iz  ,ix-2) + uo(iz  ,ix+2))*xscale + &
                  c0* uo(iz,  ix  ) * sum                  + &
                  c1*(uo(iz-1,ix  ) + uo(iz+1,ix  ))*zscale + &
                  c2*(uo(iz-2,ix  ) + uo(iz+2,ix  ))*zscale) + &
   	          2*uo(iz,ix) - um(iz,ix)
          end do
       end do
    !$OMP end parallel do   
    else 
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=3,nx-2
          do iz=3,nz-2         
            up(iz,ix) = &
               (c1*(vp2(iz,ix-1)*uo(iz  ,ix-1) + vp2(iz,ix+1)*uo(iz  ,ix+1))*xscale + &
                c2*(vp2(iz,ix-2)*uo(iz  ,ix-2) + vp2(iz,ix+2)*uo(iz  ,ix+2))*xscale + &
                c0* vp2(iz,ix)*uo(iz,  ix  ) * sum                                  + &
                c1*(vp2(iz-1,ix)*uo(iz-1,ix  ) + vp2(iz+1,ix)*uo(iz+1,ix  ))*zscale + &
                c2*(vp2(iz-2,ix)*uo(iz-2,ix  ) + vp2(iz+2,ix)*uo(iz+2,ix  ))*zscale)+ & 
                2*uo(iz,ix) - um(iz,ix)

          end do
        end do
   !$OMP end parallel do
    end if

  end subroutine

!!--------------------------------------------------------------------------

!!--------------------------------------------------------------------------
!! 4th Order Laplacian
!! source has been scaled by density 'ro'
!! A faster version of laplac4 when velocity is constant


  subroutine laplac4V2(adj, um, uo,up, wvel, dt, nx, nz, idx, idz, nop)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz
    real, dimension(:,:) :: uo,up, um
    real       :: c0, c1, c2, idx, idz, dt, dt2, wvel
    real       :: xscale, zscale, total

    c0 = -30./12.
    c1 = +16./12.
    c2 = - 1./12.
    dt2 = dt**2
    xscale = wvel*dt2*idx
    zscale = wvel*dt2*idz
    total    = xscale+zscale

    

    if (adj) then    !! RTM
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=3,nx-2       
          do iz=3,nz-2
          

             up(iz,ix) =  &
                  (c1*(uo(iz  ,ix-1) + uo(iz  ,ix+1))*xscale + &
                  c2*(uo(iz  ,ix-2) + uo(iz  ,ix+2))*xscale + &
                  c0* uo(iz,  ix  ) * total                 + &
                  c1*(uo(iz-1,ix  ) + uo(iz+1,ix  ))*zscale + &
                  c2*(uo(iz-2,ix  ) + uo(iz+2,ix  ))*zscale) + &
   	          2*uo(iz,ix) - um(iz,ix)
          end do
       end do
    !$OMP end parallel do   
    else 
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=3,nx-2
          do iz=3,nz-2         
            up(iz,ix) = &
               (c1*(uo(iz  ,ix-1) + uo(iz  ,ix+1))*xscale + &
                c2*(uo(iz  ,ix-2) + uo(iz  ,ix+2))*xscale + &
                c0* uo(iz,  ix  ) * total                 + &
                c1*(uo(iz-1,ix  ) + uo(iz+1,ix  ))*zscale + &
                c2*(uo(iz-2,ix  ) + uo(iz+2,ix  ))*zscale)+ & 
                2*uo(iz,ix) - um(iz,ix)

          end do
        end do
   !$OMP end parallel do
    end if

  end subroutine

!!--------------------------------------------------------------------------

!! 4th Order Laplacian with Random Boundary

subroutine laplac4RAN(ishot, um, uo,up, vp2, Ran1, Ran2, dt, nx, nz, bndx, bndz, idx, idz, nop)
!! Perform laplac4 with random boundary
!! Ran1(nz,2*bndx,nshot) and Ran2(nx-2*bndx,2*bndz,nshot) contains the random velocity value in the boundary zone


    logical    :: adj, flag, topflag
    integer    :: iz, ix, nop, nx, nz, ishot, bndx, bndz
    real, dimension(:,:) :: uo,up, um, vp2
    real, dimension(:,:,:) :: Ran1, Ran2
    real       :: c0, c1, c2, idx, idz, dt, dt2, v

    c0 = -30./12.
    c1 = +16./12.
    c2 = - 1./12.
    dt2 = dt**2
    flag = .true. 
    topflag = .false.
       
!         write(0,*) 'Inside Laplac4 Random', nop
          

          do ix=1+nop,nx-nop

             do iz=1+nop,nz-nop

! write(0,*) '*** iz,ix,v', iz, ix, v
                 ! initialize
                 flag = .true. 

                 if (ix .le. bndx) then
                     v = Ran1(iz,ix,ishot);
                     flag = .false.;
                 else if (ix .gt. nx-bndx) then
                     v = Ran1(iz,ix-nx, ishot);
                     flag = .false.;
                 end if
                
                 if (flag) then
                   if ((iz .le. bndz) .and. topflag ) then
                       v = Ran2(iz,ix,ishot);
                   else if ((iz .gt. nz-bndz)) then
                       v = Ran2(iz+bndz, ix,ishot);
                   else
                       v = vp2(iz,ix);
                   end if
                 end if

! write(0,*) 'iz,ix,v', iz, ix, v

              up(iz,ix) = dt2*v* &
                  (c1*(uo(iz  ,ix-1) + uo(iz  ,ix+1))*idx + &
                  c2*(uo(iz  ,ix-2) + uo(iz  ,ix+2))*idx + &
                  c0* uo(iz,  ix  ) * (idx + idz)        + &
                  c1*(uo(iz-1,ix  ) + uo(iz+1,ix  ))*idz + &
                  c2*(uo(iz-2,ix  ) + uo(iz+2,ix  ))*idz) + &
   	          2*uo(iz,ix) - um(iz,ix)
          end do
       end do

  end subroutine
!!--------------------------------------------------------------------------
!! 6th Order approximation to the Laplacian
!! Coefficients are obtained from 
!! Journal of Computational Physics (Vol. 227, Issue 1, year 2007, pp.513-536)


  subroutine laplac6(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz
    real, dimension(:,:) :: uo,up, um, vp2
    real       :: c0, c1, c2, c3, c4, idx, idz, dt, dt2
    real       :: xscale, zscale, total

    c0 = -252./60.
    c1 = +52./60.
    c2 = +16./60.
    c3 = -1./60.
    c4 = -2./60


    dt2 = dt**2
    xscale = dt2*idx
    zscale = dt2*idz
    total    = xscale+zscale

 !   write(0,*) 'xscale, zscale, sum, vp2', xscale, zscale, sum, vp2(nz/2,nx/2) 


    if (adj) then    !! RTM
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=1+nop,nx-nop       
          do iz=1+nop,nz-nop
          

	      up(iz,ix) =  2*uo(iz,ix) - um(iz,ix) + vp2(iz,ix)*xscale*( c0*uo(iz,ix-1) + &
                          c1*(uo(iz+1,ix) + uo(iz-1,ix) + uo(iz,ix+1) + uo(iz,ix-1)) + &
                          c2*(uo(iz+2,ix) + uo(iz-2,ix) + uo(iz,ix+2) + uo(iz,ix-2)) ) 
!                          c3*(uo(iz+2,ix+1) + uo(iz+1,ix+2) + uo(iz-2,ix+1) + uo(iz-1,ix+2)) + &
!                          c3*(uo(iz-2,ix-1) + uo(iz-1,ix-2) + uo(iz+2,ix-1) + uo(iz+1,ix-2)) + &
                           
			 


!    write(0,*) 'c1*xscale*vp, vp', c1*xscale*vp2(iz,ix), vp2(iz,ix)
!     write(0,*) 'uo(ix+1) uo(ix-1)', uo(iz,ix-1)+uo(iz,ix+1)
          end do
       end do
!!    !$OMP end parallel do   
    else 
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=1+nop,nx-nop
          do iz=1+nop,nz-nop   
	    up(iz,ix) = 2*uo(iz,ix) - um(iz,ix) + xscale*( c0*vp2(iz,ix)*uo(iz,ix)  + &
                        c1*(uo(iz+1,ix)*vp2(iz+1,ix) + uo(iz-1,ix)*vp2(iz-1,ix) + uo(iz,ix+1)*vp2(iz,ix+1) + uo(iz,ix-1)*vp2(iz,ix-1)) + &
                        c2*(uo(iz+2,ix)*vp2(iz+2,ix) + uo(iz-2,ix)*vp2(iz-2,ix) + uo(iz,ix+2)*vp2(iz,ix+2) + uo(iz,ix-2)*vp2(iz,ix-2)) ) 
!         c3*(uo(iz+2,ix+1)*vp2(iz+2,ix+1) + uo(iz+1,ix+2)*vp2(iz+1,ix+2) + uo(iz-2,ix+1)*vp2(iz-2,ix+1) + uo(iz-1,ix+2)*vp2(iz-1,ix+2)) + &
!         c3*(uo(iz-2,ix-1)*vp2(iz-2,ix-1) + uo(iz-1,ix-2)*vp2(iz-1,ix-2) + uo(iz+2,ix-1)*vp2(iz+2,ix-1) + uo(iz+1,ix-2)*vp2(iz+1,ix-2)) + &
                         
      

          end do
        end do
!!   !$OMP end parallel do
    end if

             

  end subroutine
!!--------------------------------------------------------------------------
!! 6th Order approximation to the Laplacian
!! source has been scaled by density 'ro'


  subroutine laplac6v2(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz
    real, dimension(:,:) :: uo,up, um, vp2
    real       :: c0, c1, c2, c3, idx, idz, dt, dt2
    real       :: xscale, zscale, total

    c0 = -49./18.
    c1 = +3.
    c2 = -3./10.
    c3 = +1./45.


    dt2 = dt**2
    xscale = dt2*idx
    zscale = dt2*idz
    total    = xscale+zscale

 !   write(0,*) 'xscale, zscale, sum, vp2', xscale, zscale, sum, vp2(nz/2,nx/2) 


    if (adj) then    !! RTM
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=1+nop,nx-nop       
          do iz=1+nop,nz-nop
          

	      up(iz,ix) = 2*uo(iz,ix) - um(iz,ix) + vp2(iz,ix)*(uo(iz, ix-1) + uo(iz, ix+1))*c1*xscale
			
			 
!	      up(iz,ix) = vp2(iz,ix)*((uo(iz, ix-1) + uo(iz, ix+1))*c1*xscale + &
!                          (uo(iz, ix-2) + uo(iz, ix+2))*c2*xscale + &
!                          (uo(iz, ix-3) + uo(iz, ix+3))*c3*xscale + &
!                           uo(iz, ix  )*c0*total + &
!                          (uo(iz-1,ix) + uo(iz+1,ix))*c1*zscale + &
!                          (uo(iz-2,ix) + uo(iz+2,ix))*c2*zscale + &
!                          (uo(iz-3,ix) + uo(iz+3,ix))*c3*zscale) 
!                          (2*uo(iz,ix) - um(iz,ix))



!    write(0,*) 'c1*xscale*vp, vp', c1*xscale*vp2(iz,ix), vp2(iz,ix)
!     write(0,*) 'uo(ix+1) uo(ix-1)', uo(iz,ix-1)+uo(iz,ix+1)
          end do
       end do
!!    !$OMP end parallel do   
    else 
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=3,nx-2
          do iz=3,nz-2         
            up(iz,ix) = &
                c1*(vp2(iz,ix-1)*uo(iz  ,ix-1) + vp2(iz,ix+1)*uo(iz  ,ix+1))*xscale + &
                c2*(vp2(iz,ix-2)*uo(iz  ,ix-2) + vp2(iz,ix+2)*uo(iz  ,ix+2))*xscale + &
		c3*(vp2(iz,ix-3)*uo(iz  ,ix-3) + vp2(iz,ix+3)*uo(iz  ,ix+3))*xscale + &
                c0* vp2(iz,ix)*uo(iz,  ix  ) * total                                 + &
                c1*(vp2(iz-1,ix)*uo(iz-1,ix  ) + vp2(iz+1,ix)*uo(iz+1,ix  ))*zscale + &
                c2*(vp2(iz-2,ix)*uo(iz-2,ix  ) + vp2(iz+2,ix)*uo(iz+2,ix  ))*zscale + & 
                c3*(vp2(iz-3,ix)*uo(iz-3,ix  ) + vp2(iz+3,ix)*uo(iz+3,ix  ))*zscale + & 
                2*uo(iz,ix) - um(iz,ix)

          end do
        end do
!!   !$OMP end parallel do
    end if

             

  end subroutine
!!--------------------------------------------------------------------------

    subroutine data_init(nz,nx,nt,nshot, fname_)
      
      complex, allocatable, dimension(:) :: ctmp
      integer                            :: i, nloop
      integer                            :: nx, nz, nt,  nshot
      character(len=128)                 :: fname_

      write(0,*) 'Initializaing data ...'
      allocate(ctmp(nz*nx*nt))
      ctmp=0
      nloop= nshot
      do i=1, nloop

         write(0,*) 'Initializing:', i, 'of', nloop
         call sep_write(ctmp, fname_, esize=4)

      end do

      deallocate(ctmp)
      
    end subroutine data_init

!!--------------------------------------------------------------------------

!!--------------------------------------------------------------------------
    subroutine win_out(us, nz, nx, nt, nshot,ishot,fname_)
      
      real, dimension(:,:,:)      :: us
      integer                   :: nz, nx, nt, nshot, istore, ishot
      integer, dimension(4)     :: n_cube, j_wind,n_wind, f_wind
      integer                   :: n_dim_cube, err, srite_window
      character :: y
      character (len=128)       :: fname_
      

      n_dim_cube = 4
      n_cube =(/nz, nx, nt, nshot/)
      j_wind =(/1, 1, 1, 1/)
      n_wind = (/nz, nx, nt, 1/)
      f_wind = (/0, 0, 0, ishot-1/)

      write(0,*) 'writing sou wfld shot#:', ishot
     write(0,*) 'fname_', fname_

      err = srite_window(fname_, n_dim_cube, n_cube, n_wind, f_wind, j_wind, 4 , us)
      
      if (err /= 0) call seperr("problem writing data");

    end subroutine 
!! ___________________________________________________________________________

!!--------------------------------------------------------------------------
    subroutine win_in(us, nz, nx, nt, nshot,ishot,fname_)
      
      real, dimension(:,:,:)      :: us
      integer                   :: nz, nx, nt, nshot, ishot
      integer, dimension(4)     :: n_cube, j_wind,n_wind, f_wind
      integer                   :: n_dim_cube, err, sreed_window
      character :: y
      character (len=128)       :: fname_
      

      n_dim_cube = 4
      n_cube =(/nz, nx, nt, nshot/)
      j_wind =(/1, 1, 1, 1/)
      n_wind = (/nz, nx, nt, 1/)
      f_wind = (/0, 0, 0, ishot-1/)

      write(0,*) 'reeding sou wfld shot#:', ishot



      call sep_read(us, fname_, dim1=nz, dim2=nx, dim3=nt, esize=4)
!      err = sreed_window("souSave2.H", n_dim_cube, n_cube, n_wind, f_wind, j_wind, 4 , us)
      
      if (err /= 0) call seperr("problem reeding data");

    end subroutine 
!! ___________________________________________________________________________


!!--------------------------------------------------------------------------
    subroutine test_sreed()
      
      real, dimension(:,:), allocatable      :: array
      real, dimension(:,:,:,:,:), allocatable      :: array2
      integer                   :: nz, nx, nt, nshot, ishot
      integer, dimension(5)     :: n_cube, j_wind,n_wind, f_wind
      integer                   :: n_dim_cube, err, sreed_window
      

      n_dim_cube = 5
      n_cube =(/5, 1400, 4000, 1, 1/)
      j_wind =(/1, 1, 1, 1, 1/)
      n_wind = (/5, 1400, 1, 1, 1/)
      f_wind = (/0, 0, 0, 0, 0/)
	

      allocate(array(5,1400));
      allocate(array2(5,1400,1,1,1));
	write(0,*)n_cube;write(0,*)n_wind;write(0,*)j_wind;write(0,*)f_wind
	write(0,*)size(array2,1),size(array2,2),size(array2,3),size(array2,4),size(array2,5)

      write(0,*) 'testing sreed_window:'

   !   call sep_read(array,"5shots.H")

      call sreed("5shots.H",array2,0)
     write(0,*) "SIZE",size(array2)
write(0,*) "WHAT",n_dim_cube
write(0,*) "XX",n_cube
write(0,*) "XX",f_wind
write(0,*) "XX",j_wind
write(0,*) "XX",n_wind

      err = sreed_window('shots.H', n_dim_cube, n_cube, n_wind, f_wind, j_wind, 4 , array2)
      
      if (err /= 0) call seperr("problem reeding data");

      write(0,*) 'array 3,700', array(3,700)
      write(0,*) 'array sum', sum(array)

      deallocate(array,array2)

    end subroutine 
!! ___________________________________________________________________________


end module
