module hycdstep_mod  
	
    use hpf_mod

contains

    integer function hycdstep(sd,alfa_keep,beta_keep,ss,rd,gg,rm,eps) 
	real, dimension (:)::                        ss,rd,gg
        real, dimension (:), optional ::                      rm          
        double precision   ::   alfa_keep, beta_keep
	real, optional     ::                                    eps
        logical            :: sd         ! steepest descent or not
 
	real, dimension (:), allocatable :: hp,hs,tmpd,tmpm

        double precision     :: alfa, beta, h, hm, hsave
        double precision     :: sds, gdg, gds, determ, gdr, sdr 
        integer              :: nsteps,k,ntries,l,nd,nm
	real                 :: cst
	logical              :: cont, icont, reg

	reg=present(rm)
	nsteps=5
	ntries=5
        nm=0
	nd=size(rd)
	if (reg) nm=size(rm)

	allocate (hp(nd+nm))
        allocate (hs(nd+nm))
        allocate (tmpd(nd))
        if (reg) allocate (tmpm(nm))
    
        k=0
        alfa_keep = 0.d0    ! gradient direction scaling
        beta_keep = 0.d0    ! previous direction scaling
        alfa = 0.d0
        beta = 0.d0
	cst  = 0.001
        hm  = hpf(rd)
	if (reg) hm=hm+eps*hpf(rm)
	h   = 0.
	cont=.true.

	if (sd) then !! Line search
          do while(cont)          
              hp(1:nd) = hpfp(rd)
	      if (reg) hp(nd+1:)= eps*hpfp(rm)
              hs(1:nd) = hpfs(rd)
	      if (reg) hs(nd+1:)= eps*hpfs(rm)
 
              if( dot_product(hs*gg, gg) == 0 ) call erexit('hycdstep: grad vanishes identically')
	      alfa = - sum( dprod( gg, hp)) / sum( dprod( hs*gg, gg))
	      tmpd=rd+alfa*gg(1:nd)
	      if (reg) tmpm=rm+alfa*gg(nd+1:)
	      h  = hpf(tmpd)
	      if (reg) h=h+eps*hpf(tmpm)
              if (abs(alfa).lt.epsilon(abs(alfa))) exit         
            
              hsave = h
              icont=.false.
              if (h.gt.hm) icont=.true.
	      do while (icont) 
           	   alfa=alfa/2
  	   	   tmpd=rd+alfa*gg(1:nd)
	   	   if (reg) tmpm=rm+alfa*gg(nd+1:)

	   	   h  = hpf(tmpd)
           	   if (reg) h=h+eps*hpf(tmpm)
		
           	   if (hsave.eq.h) icont=.false.
           	   if (h.le.hm) icont=.false.
           	   hsave = h
              end do          
               
              write(0,*) 'h',h,'hm',hm
              rd = rd + alfa*gg(1:nd)
              if (reg)rm = rm + alfa*gg(nd+1:)
              alfa_keep=alfa_keep+alfa 

              k=k+1

	      if (h.gt.hm) cont=.false.
	      if (k.ge.nsteps) cont=.false.

              hm = h

	    end do
         
        else         
          do while(cont)     
              hp(1:nd) = hpfp(rd)
	      if (reg) hp(nd+1:)= eps*hpfp(rm)
              hs(1:nd) = hpfs(rd)
	      if (reg) hs(nd+1:)= eps*hpfs(rm)

              gdg =   sum( dprod( hs*gg, gg))  !  ---                              --- 
              sds =   sum( dprod( hs*ss, ss))  !  \  H_i''|(gi) (gi si)|| alfa |= -\  H_i'|gi|
              gds =   sum( dprod( hs*gg, ss))  !  /__     |(si)        || beta |   /__    |si|
              gdr = - sum( dprod(    gg, hp))  !   i                                i
              sdr = - sum( dprod(    ss, hp))

              if( gdg==0. .or. sds==0.)   hycdstep = 1; return 
              determ = gdg * sds * max (1.d0 - (gds/gdg)*(gds/sds), 1.d-12)
              
              alfa = ( sds * gdr - gds * sdr ) / determ
              beta = (-gds * gdr + gdg * sdr ) / determ
         
              tmpd=rd+alfa*gg(1:nd)+beta*ss(1:nd)
	      tmpm=rm+alfa*gg(nd+1:)+beta*ss(nd+1:)
	      h  = hpf(tmpd)
              if (reg) h=h+eps*hpf(tmpm)
	      
              hsave = h
              icont=.false.
              if (h.gt.hm) icont=.true.
	      do while (icont) 
                 alfa=alfa/2; beta=beta/2
                 tmpd=rd+alfa*gg(1:nd)+beta*ss(1:nd)
	         if (reg) tmpm=rm+alfa*gg(nd+1:)+beta*ss(nd+1:)
	         h  = hpf(tmpd)
                 if (reg) h=h+eps*hpf(tmpm)
             
                 if (hsave.eq.h) icont=.false.
                 if (h.le.hm) icont=.false.
                 hsave = h
              end do

              rd = rd+alfa*gg(1:nd)+beta*ss(1:nd)
              if (reg) rm = rm+alfa*gg(nd+1:)+beta*ss(nd+1:)   

              alfa_keep=alfa_keep+alfa          
              beta_keep=beta_keep+beta 

              k=k+1
 
	      if (h.gt.hm) cont=.false.
	      if (k.gt.nsteps) cont=.false.
              hm = h
          end do         	
        end if
        hycdstep=0
	deallocate(hp,hs,tmpd)
        if (reg) deallocate(tmpm)

    end function

end module
