!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! Wavefield propagation module
!! Written by Noha in 09-12, modified by Mandy in 11-12 for BEI
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

module prop_mod1D
  use sep
  use fourier_modOMP1D
  use smooth_mod
  implicit none 


  logical, private       :: wfldout , VDer                                ! Switch for output wavefield movie
  real, private :: dt, dt_data, dt_prop,ot_data, velconst,tdelay                                   ! Propagation time step size
  real, private :: oz, dz, zsrc,zrec, okz, dkz, pi=4.*atan(1.)
  real          :: ztapco
  integer,private :: orderT
  integer,private :: jt, jst                                    ! Ratio between dt and dt_out
  integer,private :: itdelay                               ! itdelay
  integer,private :: nt_data, nt_prop, ntwin, nz, nnz,nkz             ! Size of data
  integer,private :: nop, npad                             ! order of Laplacian operator
  real, private,pointer :: v(:), v2(:)                          ! Velocity 
  real*8, private         :: a1z, a2z, a3z, a4z, a0        ! FD coef 8th order
  real*8, private         :: b1z, b2z, b3z, b4z, b5z, b6z, b7z, b8z, b0        ! FD coef 12th order
  real, private         ::d2t  ,d2t12                         ! coef
  real, private, allocatable :: src_func(:,:),rec_func(:,:)                 ! src wavelet, receiver data
  integer,private, allocatable  :: src_loc(:,:), rec_loc(:,:)               ! source and receiver location
  real, private, pointer :: ztaper(:)                     ! Taper
  real :: ovel(2),dvel(2)                                  ! Sampling of velocity model
  complex, dimension(:), pointer    :: in1(:), out1(:), cdatapad(:)
  character(len=128), private    :: wfldfile_
  character(len=128)              :: wfldsou_, wfldimg_, wfldrec_
  contains


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
!! Wrapper function:  born2way_op
!! input:    add ([not] erase output)
!!           adj  [T] model=Ladj data; [F] data=L model
!!           model = 1D array model space vector
!!           data = 1D array data space vector
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
  integer function born2way_op(adj,add,model_in,data_in)

    logical, intent(in)          :: add, adj
    real, dimension(:)    :: model_in, data_in
    real, dimension(:), allocatable   :: tArr

    if (adj) then
       if (.not. add) model_in=0.
    else 
       if (.not. add) data_in=0.
    end if

    allocate(tArr(size(model_in)))
    tArr = 0.

    if (adj) then
       write(0,*) 'Beg Migration', adj, add, sum(model_in), sum(data_in)
       call receiver_init(adj,data_in)
	if (VDer) then
          call migrate_shot(tArr)
          call Ver2ndDer(adj,tArr,model_in,nnz)
       else 
	  call migrate_shot(model_in)
       end if
       write(0,*) 'End Migration',  sum(model_in), sum(data_in)
       call receiver_close()
    else
       write(0,*) 'Beg born', adj, add, sum(model_in), sum(data_in)
       call receiver_init(adj,data_in)
       if (VDer)  then 
          call Ver2ndDer(adj,model_in,tArr,nnz)
          call born_model(tArr,data_in)
       else
	  call born_model(model_in,data_in)
       end if
       call receiver_close()
       write(0,*) 'End born', adj, add, sum(model_in), sum(data_in)
    end if

    deallocate(tArr)
 
	born2way_op = 0

  end function born2way_op

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
!! SUBROUTINE #1: INITIALIZE PROPAGATION
!! This subroutine is to prepare for propagation by calculating necessary parameters and coeffients
!! 1) Figure out a dt that is an integer ratio of dt_out for propagation and passes stability
!! 2) Calculate coefficients 
!! 3) Set velocity sampling, it will just be the same as in the input velocity model. 
!! 4) Other definitions
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
  
  subroutine init_propagate(vel,oz_in,dz_in,dst,dt_prop_in,nt_data_in,dt_data_in,ot_data_in,tdelay_in,vmax,npad_in,ztapco_in,orderT_in,VDer_in,wfldout_in,wfldsou_in,wfldrec_in,wfldimg_in) 
    logical          :: wfldout_in, VDer_in
    real, target     :: vel(:) !Input velocity model
    real             :: dz_in, ztapco_in
    real*8             :: idz    
    real             :: dst, dt_data_in,ot_data_in, dt_prop_in,dt_stable, val, tdelay_in  !Sampling of the data
    real             :: oz_in   !Origin of velocity model
    real             :: vmax    !the maximum velocity contained in the velocity model
    integer          :: iz,nt_data_in,k,p, npad_in, orderT_in !integral ratios between dt_data and dt, origins
  character(len=128)    :: wfldfile_in,wfldsou_in,wfldrec_in,wfldimg_in
    
    write(0,*) 'Running: init_propagate'

    ! (1) The stability condition to calculate the new dt from dt_data:
 !   vmax=MAXVAL(vel)
    nt_data=nt_data_in
    dt_data=dt_data_in 
    ot_data=ot_data_in
    dt_prop=dt_prop_in
    nt_prop = ((nt_data-1)*dt_data)/dt_prop+2.5
    tdelay=tdelay_in
    dz=dz_in; oz=oz_in
    wfldout=wfldout_in
    VDer = VDer_in
    wfldfile_=wfldfile_in
   wfldsou_ = wfldsou_in
   wfldrec_ = wfldrec_in
   wfldimg_ = wfldimg_in

	k=ceiling(dt_data/(0.8*dz/vmax))+1	!using ceiling to get the largest possible integer,
	!!since we need vmax*dt to be less than 0.707 (2D Courant number)
	dt_stable=dt_data/k
        if (dt_prop .gt. dt_stable) write(0,*) 'Warning, dt_prop is greater than dt_stable', dt_stable, dt_prop
        dt=dt_prop
	jt=dt_data/dt_prop
        jst=dst/dt
        write(0,*) 'init_', vmax, nt_data, k, dt_data, dt,jt, dz


     write(0,*) 'nt_data nt_prop', nt_data, nt_prop
     write(0,*) 'dt_data dt_prop', dt_data, dt_prop
     write(0,*) 'tdelay', tdelay

    !(2) Coefficient calculation (8th order finite diff)
    idz=DBLE(dt*dt/(dz*dz))
!    d2t=dt*dt	;   d2t12 = dt*dt/12.
    d2t=1.      ;   d2t12 = 1./12.	
    a0  = DBLE(-2.84722222222*idz)
    a1z = DBLE(1.6000000*idz)       	
    a2z = DBLE(-0.2000000*idz)       	
    a3z = DBLE(0.0253968254*idz)       	
    a4z = DBLE(-0.00178571428*idz)       	

    !(2b) Coefficient calculation (16th order finite diff)
    b0  = DBLE(-3.054844104292959 *idz)  
    b1z = DBLE( 1.777777777756123 *idz)
    b2z = DBLE( -0.311111111087384 *idz)
    b3z = DBLE( 0.075420875408331 *idz)
    b4z = DBLE(-0.017676767673411*idz)
    b5z = DBLE( 0.003480963480268 *idz)
    b6z = DBLE( -5.180005178943935e-04*idz)
    b7z = DBLE( 5.074290787522322e-05 *idz)
    b8z = DBLE(  -2.428127427618616e-06*idz)

    !(3) Find the velocity parameters
    v=>vel      

   !(4) Other defn
   nop=4;   ! Using 8th order Laplacian
   orderT = orderT_in  ! 2=2nd or 4=4th order in time

   npad=npad_in; ztapco=ztapco_in
   nz=size(v,1);
   nnz=nz-2*npad


  !(5) Setup damping boundary
   call init_taperCos(npad)

   ! (7) square the velocity
   allocate(v2(nz))

      do iz=1,nz
         v2(iz) = v(iz)*v(iz)
!	  v2(iz) = 3000.*3000.
!	 if (mod(iz,50) .eq. 0) write(0,*) 'iz, v2', iz, v2(iz)
      end do

  !! Initialize Fourier Transform parameter
  allocate(in1(2*nz),out1(2*nz),cdatapad(2*nz))
  nkz = 2*nz
  okz = -1.*pi/(dz)
  dkz = 2.*pi/(2*nz*dz)
  call ft_init1D(2*nz,in1,out1)

!  write(0,*) 'nz,oz,dz',nz,oz,dz  
!  write(0,*) 'nkz,okz,dkz',nkz,okz,dkz
!  call erexit('exit noe to check kz')

  end subroutine

  subroutine close_propagate()

     deallocate(v2) 
     deallocate(ztaper)
     deallocate(in1,out1, cdatapad)


  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #2: INITIALIZE THE SOURCE
!! This subroutine define the parameters necessary to propagate source wavefields 
!! 1) Allocate src_loc(1,1): the size of this array is 1 row by 1 column
!!    Allocate src_func(nt_data*jst,1) - store the source-wavelet in time
!! 2) Set src_loc location
!! 3) Interpolate source to src_func to fit the number of steps we want to propagate
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  subroutine source_init(source,zsrc_in,zrec_in,tdelay)
    real, allocatable, dimension(:) :: source !Source function 
    real :: zsrc_in, zrec_in !Source location in initial velocity model coordinate system !!
    integer :: i1,i0, iit !number of sources
    real :: fract, tdelay


    write(0,*) 'Running: source_init'

   !(0) Itdelay
    itdelay = (tdelay)/dt + 0.5
    zsrc=zsrc_in
    zrec=zrec_in
    !(1) Allocations:
    allocate(src_loc(1,1), src_func(nt_data*jt+itdelay,1))

    !(2) Set src_loc
    src_loc(1,1)=(zsrc-oz)/dz+.5 + max(nop,npad)
    ! error check to make sure inside grid
    if ((src_loc(1,1) .gt. nnz+npad) .and. (src_loc(1,1) .lt. 1)) call erexit('src z location out of bound')

    write(0,*) 'src_loc, npad', src_loc(1,1), npad, dz

    !(3) Interpolate source to src_func to fit the number of steps in propagation   
    src_func=0
    ! linear interpolation
    do i0=1,size(source)
       do i1=1,jst
          fract=real(i1-1)/real(jst)
          src_func(i1+(i0-1)*jst,1)=(1.-fract)*source(i0)+fract*source(min(i0+1,size(source)))
       end do
    end do

!   write(0,*) 'output src_func and see if it makes sense'
!   call sep_write(src_func,'srcfunc.H')
!    call auxputch("n1",'i',size(src_func,1),"srcfunc.H")
!    call auxputch("d1",'f',dt_prop,"srcfunc.H")
!    call auxputch("o1",'f',tdelay,"srcfunc.H")
!    call auxputch("label1",'s','t(s)',"srcfunc.H") 
!  call erexit('exit now')


  end subroutine

  subroutine source_close()

     deallocate(src_loc,src_func)

  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!! SUBROUTINE #3: RECEIVER INITIALIZATION 
!! Initialize the parameter necessary for receivers wavefield or location
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine receiver_init(direction,recs)
  logical :: direction   ! [0] modeling [1] migration
  integer :: io,il,iit
  real                   :: fract
  real, dimension(:)    :: recs !Recorded data
    
 !  write(0,*) 'Running: receiver_init'
   allocate(rec_loc(1,1))
   allocate (rec_func(size(recs,1)*jt,1))


 !  write(0,*) 'receiver_init', size(recs,1), size(rec_func,1), size(rec_func,2)

   !(2) Set rec_loc
    rec_loc(1,1)=(zrec-oz)/dz+.5 + max(nop,npad)
    ! error check to make sure inside grid
    if ((rec_loc(1,1) .gt. nnz+npad) .and. (rec_loc(1,1) .lt. 1)) call erexit('rec z location out of bound')
  
   !(3) Interpolate the data in time from sampling dt_data to dt (store in reverse time)
   rec_func=0.
   if (direction) then
      ! write(0,*) 'interpolate rec func'
         do io=size(recs,1),1,-1
            do il=1,jt
               fract=real(il-1)/real(jt)
               iit = nt_data*jt-1*(-il+io*jt+1)+1
               rec_func(iit,1)=(fract)*recs(io)+(1.-fract)*recs(min(io+1,size(recs),1))
            end do
         end do
   end if
   
   
!   write(0,*) 'sum(recs), sum(rec_func)', sum(recs), sum(rec_func)
!
!	   call sep_write(rec_func, 'recfunc.H')
!   	    call auxputch("n1",'i',size(rec_func),'recfunc.H')
!   	    call auxputch("d1",'f',1.,'recfunc.H')
!   	    call auxputch("o1",'f',0.,'recfunc.H')
!   	    call auxputch("label1",'s','z(m)','recfunc.H') 
!	   call erexit('exit now')

 end subroutine

  subroutine receiver_close()

     if (allocated(rec_loc)) deallocate(rec_loc)
     if (allocated(rec_func)) deallocate(rec_func)

  end subroutine



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #4: WAVE-FIELD PROPAGATION
!! Input: prev,cur,next =  u(t+1), u(t), u(t-1)
!! Input: rec_f = time-index marker on where to inject data
!! Input: nt = number of time step to propagate
!! output: next (represent the wavefield after propagating nt time steps)
!! (1) Apply Laplacian and time-marching (in reverse time)
!!         u(t+/-1) = - u(t-/+1) + 2*u(t) + (v^2)*(dt^2)*(Laplacian)u(t)
!! Note that propagate forward and reverse in time is symmetric
!! (2) data/source injection       
!! (3) Time stepping (update the 3 time slices)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine propagate_wavefield(prev,cur,next,time_marker,nt)
    real, pointer, dimension(:)    :: prev,cur,next !Previous,current, and next wavefields
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt,ir,irx
    real*8 :: val

 ! if (mod(time_marker,200) == 0) write(0,*) 'Running: propagate_wavefield', time_marker ,'out of', nt_data*jt , sum(prev), sum(cur), sum(next)


    do it=1,nt
       val = 0.
     
       
       ! (1) Apply Laplacian and time-marching
          do iz=nop+1,nz-nop
             val = b0*DBLE(cur(iz))
             val = val + b1z*DBLE(cur(iz+1))
	     val = val + b1z*DBLE(cur(iz-1))
	     val = val + b2z*DBLE(cur(iz+2))
	     val = val + b2z*DBLE(cur(iz-2))
	     val = val + b3z*DBLE(cur(iz+3))
	     val = val + b3z*DBLE(cur(iz-3))
	     val = val + b4z*DBLE(cur(iz+4))
	     val = val + b4z*DBLE(cur(iz-4))
	     val = val + b5z*DBLE(cur(iz+5))
	     val = val + b5z*DBLE(cur(iz-5))
	     val = val + b6z*DBLE(cur(iz+6))
	     val = val + b6z*DBLE(cur(iz-6))
	     val = val + b7z*DBLE(cur(iz+7))
	     val = val + b7z*DBLE(cur(iz-7))
	     val = val + b8z*DBLE(cur(iz+8))
	     val = val + b8z*DBLE(cur(iz-8))
 	     val = val*DBLE(v2(iz))
	     val = val + DBLE(-prev(iz) + 2*cur(iz))
	     next(iz) = REAL(val)
        end do

        ! Applying damping boundary
        call taperingCos(prev,cur,next)

        !(3) Time stepping
        prev=cur
        cur=next
        
        
   end do   
end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #5: WAVE-FIELD PROPAGATION Adjoint
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine propagate_wavefield_adj(prev,cur,next,time_marker,nt)
    real, pointer, dimension(:)    :: prev,cur,next !Previous,current, and next wavefields
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt
    real :: val

! write(0,*) 'Running: propagate_wavefield', time_marker ,'out of', nt_data*jt

!  if (mod(time_marker,500) == 0) write(0,*) 'Running: propagate_wavefield_adj', time_marker ,'out of', nt_data*jt

    do it=1,nt
       val = 0.
       
!       ! (1) Apply Laplacian and time-marching
          do iz=nop+1,nz-nop
             val = b0*v2(iz)*cur(iz)
             val = val + b1z*v2(iz+1)*cur(iz+1)
             val = val + b1z*v2(iz-1)*cur(iz-1)
             val = val + b2z*v2(iz+2)*cur(iz+2)
             val = val + b2z*v2(iz-2)*cur(iz-2)
             val = val + b3z*v2(iz+3)*cur(iz+3)
             val = val + b3z*v2(iz-3)*cur(iz-3)
             val = val + b4z*v2(iz+4)*cur(iz+4)
             val = val + b4z*v2(iz-4)*cur(iz-4)	 
             val = val + b5z*v2(iz+5)*cur(iz+5)
             val = val + b5z*v2(iz-5)*cur(iz-5)
             val = val + b6z*v2(iz+6)*cur(iz+6)
             val = val + b6z*v2(iz-6)*cur(iz-6)
             val = val + b7z*v2(iz+7)*cur(iz+7)
             val = val + b7z*v2(iz-7)*cur(iz-7)
             val = val + b8z*v2(iz+8)*cur(iz+8)
             val = val + b8z*v2(iz-8)*cur(iz-8)	 
	     val = val + -prev(iz) + 2.*cur(iz)
	     next(iz) = val
           end do
     


	!(2) Tapering
	call taperingCos(prev,cur,next)

        !(3) Time stepping
        prev=cur
        cur=next
        
        
   end do   
end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #5: WAVE-FIELD PROPAGATION 4th order in time
!! Input: prev,cur,next =  u(t+1), u(t), u(t-1)
!! Input: rec_f = time-index marker on where to inject data
!! Input: nt = number of time step to propagate
!! output: next (represent the wavefield after propagating nt time steps)
!! (1) Apply Laplacian and time-marching (in reverse time)
!!         u(t+/-1) = - u(t-/+1) + 2*u(t) + (v^2)*(dt^2)*(Laplacian)u(t) - (dt^2)(v^2)/12 (Laplacian)(v^2 Laplacian u(t))
!! Note that propagate forward and reverse in time is symmetric
!! (2) data/source injection       
!! (3) Time stepping (update the 3 time slices)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine propagate_wavefield4th(prev,cur,next,next2,time_marker,nt)
    real, pointer, dimension(:)    :: prev,cur,next,next2 !Previous,current, and next wavefields
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt,ir
    real :: val

!   if (time_marker .gt. 600) write(0,*) 'Running: propagate_wavefield 4th', time_marker ,sum(prev), sum(cur), sum(next)
!   if (time_marker .gt. 1000) call erexit('exit nnoww')
  if (mod(time_marker,200) == 0) write(0,*) 'Running: propagate_wavefield 4th', time_marker ,'out of', nt_prop, sum(prev), sum(cur), sum(next)

    do it=1,nt
       val = 0.
       

       ! (1) Apply Laplacian;  next = Laplacian cur  
       call ApplyLaplac1D(cur,next)

       ! (2) Calculate (Laplacian) (dt2 v2 Laplacian u(t) )
        do iz=1,nz
             val = next(iz)
             next(iz) = d2t*v2(iz)*val
             next2(iz) = 0.
        end do 
        call ApplyLaplac1D(next,next2)

       ! (3) Time-marching 4th order
         do iz=1,nz
              val = next2(iz)
              next(iz) = -prev(iz) + 2*cur(iz) + next(iz) !+ d2t12*v2(iz)*val
!              next(iz) = -prev(iz) + 2*cur(iz) + next(iz) + d2t12*v2(iz)*val
         end do

      ! (4) Applying damping boundary
 !       call taperingCos(prev,cur,next)

        !(5) Time stepping
        prev=cur
        cur=next
        
        
   end do   
end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #10: Propagate the Born receiver wavefield
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine propagate_born(image,dat,source_array)
  real, dimension(:,:) :: source_array  
  real, dimension(:) :: image, dat
  integer :: i, id, ix, it,iz, irx, io  !Internal time variable
  real, pointer,dimension(:) :: prev2,cur2,next2, next2v2

!  write(0,*) 'Running: propagate_born'  

  ! (1) Allocation
  allocate(prev2(nz),cur2(nz),next2(nz))
  if (orderT .eq. 4) allocate(next2v2(nz))

  ! (2) Initialization
     do iz=1,nz
        prev2(iz)=0; cur2(iz)=0; next2(iz)=0;
     end do
     if (orderT .eq. 4) next2v2=0.
  it=0
  

  do i=1,nt_data*jt

     ! (2) propagate rec wavefield 
     if (orderT .eq. 2) then
         call propagate_wavefield_adj(prev2,cur2,next2,i,1)
     else if (orderT .eq. 4) then
         call propagate_wavefield4th(prev2,cur2,next2,next2v2,i,1)
     end if


     ! (3) Inject Born Source at every time-step    
       id = i
       io=0
       call forward_born_inject(source_array(:,id),source_array(:,id+1),cur2,image,io,jt)


     !(4) Output wavefield movies  
       if ( (wfldout) .and. (mod(i,jt) .eq. 0)) then
         call sep_write(image,wfldimg_)
         call sep_write(source_array(:,id),wfldsou_)
         call sep_write(cur2,wfldrec_)
       end if


    ! (4) Collect data without Ghost
     if (jt .gt. 1) then
	id = (i-1)/jt+1
       if (mod(i,jt)==1) then      
            dat(id) = cur2(rec_loc(1,1)) 
       end if
     else
            dat(i) = cur2(rec_loc(1,1)) 
     end if


  end do

  deallocate(prev2,cur2,next2)
  if (orderT .eq. 4) deallocate(next2v2)  


end subroutine 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #9: Propagate receiver wavefield and apply imaging condition
!! (1) Allocation
!! (2) propagate receiver wavefield
!! (3) apply imaging condition
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine propagate_receivers(image,source_array)
  real, dimension(:,:) :: source_array  
  real, dimension(:) :: image
  integer :: i,io,id,irx, ir
  real, pointer,dimension(:) :: prev1,cur1,next1,next1v2

!  write(0,*) 'Running: propagate_receivers'  

  ! (1) Allocation
  allocate(prev1(nz),cur1(nz),next1(nz))
  if (orderT .eq. 4) allocate(next1v2(nz))
  prev1=0; cur1=0; next1=0;
 
  do i=1,nt_prop

        !(1) source/receiver injection
        do ir=1,size(rec_loc,2)
 !          if ((mod(it+time_marker,100).eq.1) .and. (it .lt. 2000)) write(0,*) 'wave_func', it, time_marker, wave_func(it+time_marker,ir)
           cur1(rec_loc(1,ir))=cur1(rec_loc(1,ir))+rec_func(i,ir)
        end do  

     ! (2) propagate rec wavefield for jt time-step (note: it is reverse in time)
       if (orderT .eq. 2) then
          call propagate_wavefield(prev1,cur1,next1,i,1)
       else if (orderT .eq. 4) then
          call propagate_wavefield4th(prev1,cur1,next1,next1v2,i,1)
       end if

       id = i
       io = 0
       if (id .lt. nt_prop) then
        ! (3) Apply imaging condition at every jt time-step
        call imaging_condition(source_array(:,nt_prop-id+1),source_array(:,nt_prop-id),prev1,image,io,jt)
       end if
     

    !(2) Output wavefield movies  
    if ((wfldout) .and. (mod(i,jt) .eq. 0) ) then
       call sep_write(image,wfldimg_)
       call sep_write(source_array(:,nt_prop-i),wfldsou_)
       call sep_write(prev1,wfldrec_)
    end if


  end do

  deallocate(prev1,cur1,next1)
  if (orderT .eq. 4) deallocate(next1v2)  

end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #6: IMAGING
!! (1) apply zero-lag cross correlation imaging condition
!! (2) Output to wavefield movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

 subroutine imaging_condition(source0,source1, receiver,image,io,jt)
    real, dimension(:) :: image,  source0, source1, receiver
    integer              ::  iz,io,jt

    !(1) Apply imaging condition
       do iz=1,nnz
  !        image(iz) = image(iz) + (source0(iz)*(jt-io)+io*source1(iz))*receiver(iz+npad)
          image(iz) = image(iz) + source1(iz)*receiver(iz+npad)
       end do

 end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #7: Forward Modeling Born Injection
!! (1) Apply Born source injection to receiver wavefield
!! (2) Output wavefield movies  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
 subroutine forward_born_inject(source0,source1,receiver,image,io,jt)

    real, dimension(:) :: image,  source0, source1, receiver
    integer              :: iz,io,jt

    !(1) Apply Born source injection to receiver wavefield
       do iz=1,nnz
          receiver(iz+npad) = receiver(iz+npad) + (source0(iz)*(jt-io)+source1(iz)*io)*image(iz)
       end do

 end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #8: Propagate and save source wavefield
!! (1) Allocation
!! (2) propagate source wavefield
!! (3) Save to array at every jt time-step
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

  subroutine create_source_array(array)
  real, allocatable, dimension(:,:)  :: array !Array of nz,nt_data
  integer :: i, iz, it_cur, id, is
  real, pointer, dimension(:) :: prev,cur,next,next2

!  write(0,*) 'Running: create_source_array',  itdelay, nt_prop, size(array,2)
  ! (1) Allocation  
  allocate(prev(nz),cur(nz),next(nz))

  if (orderT .eq. 4) allocate(next2(nz))

  do i=1,nt_prop+itdelay
      it_cur = i

        !(1) source/receiver injection
        do is=1,size(src_loc,2)
 !          if ((mod(it+time_marker,100).eq.1) .and. (it .lt. 2000)) write(0,*) 'wave_func', it, time_marker, wave_func(it+time_marker,ir)
           cur(src_loc(1,is))=cur(src_loc(1,is))+src_func(it_cur,is)
        end do  

     if (orderT .eq. 2) then
         call propagate_wavefield(prev,cur,next,it_cur,1)
     else if (orderT .eq. 4) then
         call propagate_wavefield4th(prev,cur,next,next2,it_cur,1)
     end if


     ! (3) Save a time-slice of sou wavefield to the array at every jt time-step
       id = (i-itdelay)
       if ( (id .ge. 1) .and. (id .lt. nt_prop)) then      
         do iz=1,nnz
           array(iz,id)=next(iz+1*npad)
        end do
       end if


  end do
    
  deallocate(prev,cur,next)
  if (orderT .eq. 4) deallocate(next2)

end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #9: Propagate receiver wavefield and apply imaging condition
!! (1) Allocation
!! (2) propagate receiver wavefield
!! (3) apply imaging condition
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #10: Propagate the Born receiver wavefield
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


   
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #11: MIGRATE A SHOT
!! (1) Allocation
!! (2) Calculate source wavefield
!! (3) Calculate rec wavefield and apply imaging condition
!! (4) Write wavefield movie header (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine migrate_shot(image)

  real, dimension(:) :: image !Imagef
  real, allocatable, dimension(:,:) :: source_array
  
!  write(0,*) 'Running: migrate_shot'

  ! (1) Allocation
  allocate(source_array(nnz,nt_prop))
  source_array = 0.



  ! (2) Calculate source wavefield
    call create_source_array(source_array)


  ! (3) Calculate receiver wavefield and apply imaging condition 
    call propagate_receivers(image,source_array)

  ! (4) Write wavefield movies header 
  if (wfldout) then
       call write_movie_header1(wfldimg_)
       call write_movie_header1(wfldsou_)
       call write_movie_header2(wfldrec_)
   end if

  deallocate(source_array)

end subroutine 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #12: MODEL A SHOT
!! (1) Allocation
!! (2) Propagate and save source wavefield
!! (3) Collect data
!! (4) Output source movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine model_shot(dat) 

  integer              :: itout, iz, itb, ite, it, icur
  real                 :: rz,tout
  real, dimension(:,:) :: dat !Shot gather
  real, allocatable, dimension(:,:) :: source_array

!  write(0,*) 'Running: model_shot'
  
  ! (1) Allocation
  write(0,*) 'allocate', nnz, nt_data
  allocate(source_array(nnz,nt_prop))
  source_array = 0.

  ! (2) Propagate and save source wavefield
  call create_source_array(source_array)


  if (wfldout) then
    call sep_write(source_array,wfldfile_)
    call auxputch("n1",'i',nz,wfldfile_)
    call auxputch("n2",'i',nt_prop,wfldfile_)
    call auxputch("d1",'f',dz,wfldfile_)
    call auxputch("d2",'f',dt,wfldfile_)
    call auxputch("o1",'f',oz,wfldfile_)
    call auxputch("o2",'f',ot_data-tdelay,wfldfile_)
    call auxputch("label1",'s','z(m)',wfldfile_) 
    call auxputch("label2",'s','t(s)',wfldfile_) 
  end if 


  ! (3) collect data


  write(0,*) 'sum dat, source_array', sum(dat), sum(source_array)

  deallocate(source_array) 


end subroutine  

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #13: BORN Model
!! (1) Allocation
!! (2) Propagate and save source wavefield
!! (3) Collect data
!! (4) Output source movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine born_model(image,dat) 

  integer              :: it, irx
  real, dimension(:) :: image, dat
  real, allocatable, dimension(:,:) :: source_array
  character(len=60)   :: ddfile_

!  write(0,*) 'Running: born_model'
  

  ! (1) Allocation
  allocate(source_array(nnz,nt_prop))
  source_array = 0.

  ! (2) Propagate and save source wavefield
     call create_source_array(source_array)

  ! (3) Calculate receiver wavefield and inject born source
    call propagate_born(image,dat,source_array)

  ! (4) Write wavefield movies header 
  if (wfldout) then
       call write_movie_header1(wfldsou_)
       call write_movie_header2(wfldrec_)
       call write_movie_header1(wfldimg_)
   end if

   deallocate(source_array)


end subroutine  


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #15: Write wavefield movie header file
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
  subroutine write_movie_header1(filename)

    character(len=128)  :: filename

    call auxputch("n1",'i',nnz,filename)
    call auxputch("n2",'i',nt_data-1,filename)
    call auxputch("d1",'f',dz,filename)
    call auxputch("d2",'f',dt*jt,filename)
    call auxputch("o1",'f',oz,filename)
    call auxputch("o2",'f',0.,filename)
    call auxputch("label1",'s','z(m)',filename) 
    call auxputch("label2",'s','t(s)',filename) 

   end subroutine

  subroutine write_movie_header2(filename)

    character(len=128)  :: filename

    call auxputch("n1",'i',nz,filename)
    call auxputch("n2",'i',nt_data-1,filename)
    call auxputch("d1",'f',dz,filename)
    call auxputch("d2",'f',dt*jt,filename)
    call auxputch("o1",'f',oz,filename)
    call auxputch("o2",'f',0.,filename)
    call auxputch("label1",'s','z(m)',filename) 
    call auxputch("label2",'s','t(s)',filename) 

   end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 22: write header for image file
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

 subroutine  writeimageheader(filename,n1)
   
    character(len=128)   :: filename
    integer              :: n1, n2

    call auxputch("n1",'i',n1,filename)
    call auxputch("d1",'f',dz,filename)
    call auxputch("o1",'f',oz,filename)
    call auxputch("label1",'s','z(m)',filename) 

  end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE # 16: Apply tapering
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine taperingCos(prev,cur,next)

    integer  :: iz
    real, pointer, dimension(:) :: prev,cur,next !Previous,current, and next wavefields
    
  !  write(0,*) 'ztaper', ztaper
  !  write(0,*) 'ztapco', ztapco, size(ztaper)
  !  call erexit('exit now')  

          prev(1:npad) = prev(1:npad) * ztaper(1:npad:+1);
          prev(nz-npad+1:nz) = prev(nz-npad+1:nz) * ztaper(npad:1:-1)

          cur(1:npad) = cur(1:npad) * ztaper(1:npad:+1);
          cur(nz-npad+1:nz) = cur(nz-npad+1:nz) * ztaper(npad:1:-1)
          next(1:npad) = next(1:npad) * ztaper(1:npad:+1);
          next(nz-npad+1:nz) = next(nz-npad+1:nz) * ztaper(npad:1:-1)




  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 17: Tapering Setup
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
    subroutine init_taperCos(npad)

    integer                     :: iz, ix,ii, npad
    real                       :: dist
    real                        :: pi=3.14159265359


    allocate(ztaper(npad))

       ! all four edges
       do ii=1,npad
          !!ztaper(ii) = 0.99 + 0.01*cos(pi*(npad-ii)/npad)
          ztaper(npad-ii+1) = exp(-ztapco*real(ii)**2)
          
       end do

  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 21: Take derivative for 1D array
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

  subroutine FirstDer(adj,souIn,souOut,nt,dt)

    logical :: adj
    integer :: i, nt
    real    :: dt, rdt
    real, dimension(:) :: souIn, souOut
    real, dimension(:), allocatable :: temp

   allocate(temp(nt))
   temp(1:nt) = souIn(1:nt)
    
    souOut(1:nt) = 0.
    
    
    if (.not. adj) then
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1)+ temp(i)
         souOut(i)   = souOut(i) - temp(i)
      end do
      
    else
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1) + temp(i+1)-temp(i)
      end do
         
    end if    

    deallocate(temp)
  end subroutine 



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 23: Take Vertical Derivative of image
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
  subroutine Ver2ndDer(adj,m,m_out,nz)
  ! Take the second derivative vertically and its adjoint

  logical                   :: adj
  integer                   :: nz,iz
  real                      :: fac
  real, dimension(:), allocatable :: temp
  real, dimension(:)      :: m, m_out


  allocate(temp(nz))
  do iz=1,nz
     temp(iz) = m(iz)
     m_out(iz) = 0.
  end do
 
   fac = 0.5

      do iz=2,nz-1
         if (adj) then
            m_out(iz) = m_out(iz) + fac*(2.*temp(iz) - temp(iz+1) - temp(iz-1))
         else
            m_out(iz) = m_out(iz) + fac*2*temp(iz)
            m_out(iz) = m_out(iz) - fac*temp(iz+1)
            m_out(iz) = m_out(iz) - fac*temp(iz-1)
         end if
       end do

   m_out(1:5) = 0.
   m_out(nz-5:nz) = 0.


  end subroutine

  subroutine Time_2ndDer(adj,inArray)

   logical                :: adj
   real, dimension(:,:)   :: inArray
   integer                :: nz_in, nt_in, iz
   real, dimension(:), allocatable :: tmpArr, tmpArr2

   nz_in = size(inArray,1)
   nt_in = size(inArray,2)
   allocate(tmpArr(nt_in))
   allocate(tmpArr2(nt_in)) 

   do iz=1,nz_in
	tmpArr(1:nt_in) = inArray(iz,1:nt_in)      
        call Ver2ndDer(adj,tmpArr,tmpArr2, nt_in)
   tmpArr2(1:5) = 0.
   tmpArr2(nt_in-5:nt_in) = 0.
   inArray(iz,1:nt_in) = tmpArr2(1:nt_in)
   end do


  end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 20: create an omega^(power) wavelet with bandpass
!! nt,ot,dt, tdelay
!! power
!! Phase: zero?
!! Bandpass; flo, fhi
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

  subroutine makewavelet(fs,nt,ot,dt,tdelay,power,phase,flo,fhi)

  integer                    :: nt, phase, it, iw
  integer*8                   :: plan1f
  real                       :: pi, T, ot, dt, tdelay, power
  real                       :: flo, fhi, omega, eps
  real, dimension(:)         :: fs
  real, allocatable, dimension(:)    :: inAmp
  complex, allocatable, dimension(:) :: in, out

  allocate(in(nt), inAmp(nt), out(nt))
  

  pi=3.14159265
  T = nt*dt
  eps = 0.0001

  write(0,*) 'pass 2', T, nt, dt
  ! create amplitude spectrum
  do iw=1,nt
     omega = (2*pi/nt/T)*(iw-1)
     if (((omega .gt. flo) .or. (omega .lt. fhi)) .and. (omega .gt. eps))then
       inAmp(iw)  = omega**power  ! zero phase
     else
       inAmp(iw) = 0.
     end if
  end do


  write(0,*) 'pass 3',sum(inAmp)
  ! Smooth the bandpass
  call triangle( 5, nt, inAmp, inAmp)

  write(0,*) 'pass 4',sum(inAmp)
  do iw=1,nt
     in(iw) = cmplx(inAmp(iw),0.)
  end do
    write(0,*) 'pass 5', sum(in)
   !! Forward FFT
   call sfftw_plan_dft_1d(plan1f, nt, in, out, FFTW_BACKWARD, FFTW_ESTIMATE)
   call sfftw_execute_dft(plan1f, in, out)
   
   do it=1,nt
      fs(it)=out(it)/sqrt(real(nt)) ! divided by sqrt(n1) instead of n1. this helps when computing correlation in FFT domain
   end do
   call sfftw_destroy_plan(plan1f)

  write(0,*) 'pass 6', sum(in), sum(out), sum(fs)

  end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE # XX: Apply Laplacian 
!! 1) Fourier transform
!! 2) multiply by -kz**2
!! 3) Inverse fourier transform
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine ApplyLaplac1D(inArr,outArr)

  integer               :: iz
  real*8 :: val
  real, dimension(:)    :: inArr, outArr

    
  ! (1) Apply Laplacian and time-marching (12th order)
  do iz=nop+1,nz-nop
     val = b0*inArr(iz)
     val = val + b1z*DBLE(inArr(iz+1))
     val = val + b1z*DBLE(inArr(iz-1))
     val = val + b2z*DBLE(inArr(iz+2))
     val = val + b2z*DBLE(inArr(iz-2))
     val = val + b3z*DBLE(inArr(iz+3))
     val = val + b3z*DBLE(inArr(iz-3))
     val = val + b4z*DBLE(inArr(iz+4))
     val = val + b4z*DBLE(inArr(iz-4))
     val = val + b5z*DBLE(inArr(iz+5))
     val = val + b5z*DBLE(inArr(iz-5))
     val = val + b6z*DBLE(inArr(iz+6))
     val = val + b6z*DBLE(inArr(iz-6))
     val = val + b7z*DBLE(inArr(iz+7))
     val = val + b7z*DBLE(inArr(iz-7))
     val = val + b8z*DBLE(inArr(iz+8))
     val = val + b8z*DBLE(inArr(iz-8))
     outArr(iz)=real(val)
  end do


end subroutine




end module


