program testMain

  use sep
  use prop_mod1D

  use hycdsolver_smp_mod
  use hycdstep_mod
  use identity_mod

  implicit none

  integer                                      ::  niter         ! number of iterations
  real                                         :: qntd, qntm     ! quantile
  real                                         :: thresh_d,thresh_m  ! threshold

  logical(4) :: success
  logical                               :: wfldmovie, VDer, DW, MW
  real                                  :: dx, dz, oz, ost, dst, ot_data, dt_data, dt_prop
  real                                  :: zsrc, zrec,vmax,tdelay,tapeco
  real                                  :: dot1, dot2
  integer                               :: nnz, nz, nst, nt_data, nop, npad
  integer            			:: iz , it, orderT, bout
  real, allocatable, dimension(:)     :: vel, mwarr, dat, DWarray
  real, allocatable, dimension(:)     :: image, resdArr
  real, allocatable, dimension(:)       ::wavelet, mwarr1D
  real, allocatable, dimension(:)       ::x1, x2, y1, y2
  character(len=128)                    :: souWfile_, imgWfile_, recWfile_
  character(len=128)                    :: outfile_

  call sep_init()

  ! read in velocity model
  call from_history("n1", nz)
  call from_history("o1", oz)
  call from_history("d1", dz)

  ! read source wavelet information
  call from_aux("aux", "n1", nst)
  call from_aux("aux", "o1", ost)
  call from_aux("aux", "d1", dst)

  ! read shot gather (data) file info
  call from_aux("datafile", "n1", nt_data)
  call from_aux("datafile", "o1", ot_data)
  call from_aux("datafile", "d1", dt_data)

  write(0,*) 'nt_data, ot, dt', nt_data, ot_data, dt_data

  ! read source location
  call from_param("zsrc",zsrc)
  call from_param("npad",npad,30)
  call from_param("zrec",zrec,zsrc)
  
  ! Read wavefield movie switch
  call from_param("wfldout",wfldmovie,.false.)
  call from_param("VDer",VDer,.false.)
  call from_param("DW",DW,.false.);   call from_param("MW",MW,.false.)
  call from_param("souwfld",souWfile_,"souwfld.H")
  call from_param("imgwfld",imgWfile_,"imgwfld.H")  
  call from_param("recwfld",recWfile_,"recwfld.H")
  call from_param("outfile",outfile_)
  call from_param("tdelay",tdelay)
  call from_param("dt_prop",dt_prop,0.001)
  call from_param("tapeco",tapeco,0.00002);
  call from_param("orderT",orderT,2);



  if (dst .ne. dt_data) call erexit('dst should equal to dt_data')

  nnz=nz
  nz = nnz+2*npad

  ! Allocate memory
  allocate(vel(nz),image(nnz),wavelet(nt_data))
  allocate(dat(nt_data),resdArr(nt_data))

  ! Read velocity model
  call sep_read(vel(npad+1:nz-npad));  call from_param("vmax",vmax);
  ! padding
  do iz=1,npad
     vel(iz) = vel(npad+1)
     vel(nz-iz+1) = vel(nz-npad-1)
  end do

  vel= 2500. 

  ! Read source wavelet
  call sep_read(wavelet,"aux")
  
  ! Read data
  call sep_read(dat,'datafile')


  call from_param("outfile",outfile_,'outfile.H')

  call from_param ("niter", niter, 2)
  call from_param ('thresh_d',thresh_d, 1.)
  call from_param ('thresh_m',thresh_m, 1.)
  write(0,*) "niter   =",niter
  write(0,*) "thresh_d=",thresh_d
  write(0,*) "thresh_m=",thresh_m
  write(0,*) 'VDer', VDer
  write(0,*) "orderT", orderT

  ! Initiate propagation
  call init_propagate(vel,oz,dz,dst,dt_prop,nt_data,dt_data,ot_data, tdelay,vmax,npad,tapeco,orderT, VDer, wfldmovie,souWfile_, recWfile_, imgWfile_) 
  ! Initiate source parameters
  call source_init(wavelet,zsrc,zrec,tdelay)


! Dot product test
!   y1 = L x1
!   x2 = Ladj y2
!   x1*x2 = y1*y2

  allocate(x1(nnz), x2(nnz), y1(nt_data), y2(nt_data))
  do iz=1,nz 
      x1(iz) = cos(real(iz))
      x2(iz) = 0.
  end do
  do it=1,nt_data
       y2(it) = sin(real(it))
       y1(it) = 0.
  end do

  bout = born2way_op(.false.,.false.,x1,y1)
  bout = born2way_op(.true.,.false.,x2,y2)

  do iz=1,nnz
     dot1 = dot1 + x1(iz)*x2(iz)
  end do
  do it=1,nt_data
     dot2 = dot2 + y1(it)*y2(it)
  end do

  write(0,*) 'dot test result', dot1, dot2
  call erexit('dot test finished')

!  bout = born2way_op(.true.,.false.,image,dat)
!	   call sep_write(image, 'image.H')
!   	    call auxputch("n1",'i',size(image),'image.H')
!   	    call auxputch("d1",'f',dz,'image.H')
!   	    call auxputch("o1",'f',0.,'image.H')
!   	    call auxputch("label1",'s','z(m)','image.H') 
!  bout = born2way_op(.false.,.true.,image,dat)
!	   call sep_write(dat, 'gg.H')
!   	    call auxputch("n1",'i',nt_data,'gg.H')
!   	    call auxputch("d1",'f',dt_data,'gg.H')
!   	    call auxputch("o1",'f',0.,'gg.H')
!   	    call auxputch("label1",'s','z(m)','gg.H') 

  ! Set threshold values for hycd data weighting
  call identity_init(thresh_d,thresh_m)
  ! Run the inversion over niter-ations
  call hycdsolver_smp( m=image,d=dat, Fop=born2way_op , Wop=identityd_lop , stepper=hycdstep,niter=niter, resd=resdArr, verb=.true.)

   call close_propagate()
   call source_close()

   call to_history("n1",nnz,outfile_)
    call to_history("o1",oz,outfile_)
    call to_history("d1",dz,outfile_)
    call to_history  ("label1","z(m)",outfile_)
    call sep_write(image,outfile_)


  deallocate (image, dat)




end program
